#include <stddef.h>
#include <errormanager.h>
#include <fsmanager.h>
#include <intermanager.h>
#include <textvideo.h>
#include <usermodmanager.h>
#include <lib/stringman.h>

void Start(int argc, char *argv[], char *envp[]) __attribute__ ((section (".init")));

int (*TV_Write)(const char*,int);

void StartLibs(void) {
    unsigned int i;
    int (*start)(void);

    i = 0;
    while ((start = UserModManager_UGetStartFunctionI(i++))) {
        if ((unsigned int)start != 0xffffffff && start()) {
            TV_Write("Error starting lib.\n", sizeof("Error starting lib.\n")-1);
            UserModManager_UExit(2);
        }
    }
}

void PrintError(int error) {
    char buf[32];

    TV_Write("ls: ", sizeof("ls: ")-1);
    switch(error) {
        case E_BAD_VALUE:
            TV_Write("Bad argument.\n", sizeof("Bad argument.\n")-1);
            break;
        case E_TOO_MANY_OPEN_FILES:
            TV_Write("Too many openn files.\n", sizeof("Too many openn files.\n")-1);
            break;
        case E_NOT_FILE_DIR:
            TV_Write("Invalid file type.\n", sizeof("Invalid file type.\n")-1);
            break;
        case E_FAULT:
            TV_Write("Invalid memory address.\n", sizeof("Invalid memory address.\n")-1);
            break;
        case E_TOO_BIG:
            TV_Write("Path too big.\n", sizeof("Path too big.\n")-1);
            break;
        case E_PERMISSION_DENIED:
            TV_Write("Permission denied.\n", sizeof("Permission denied.\n")-1);
            break;
        case E_NOT_FOUND:
            TV_Write("Path not found.\n", sizeof("Path not found.\n")-1);
            break;
        case E_NOT_A_DIRECTORY:
            TV_Write("Path is not a dir.\n", sizeof("Path is not a dir.\n")-1);
            break;
        case E_IO_ERROR:
            TV_Write("IO error.\n", sizeof("IO error.\n")-1);
            break;
        case E_NO_MEMORY:
            TV_Write("Not enough memory.\n", sizeof("Not enough memory.\n")-1);
            break;
        case E_LOOP:
            TV_Write("Too many symbolic links encountered while traversing path.\n", sizeof("Too many symbolic links encountered while traversing path.\n")-1);
            break;
        default:
            TV_Write("Undefined error: ", sizeof("Undefined error: ")-1);
            User_StringMan_BufPrintF(buf, "0x%x\n", error);
            TV_Write(buf, (int)User_StringMan_Length(buf));
            break;
    }
}

void Start(int argc, char *argv[], char *envp[]) {
    int dir, ret;
    char outBuf[256];
    InterManager_OutputInfo *outInfo;

    /* get TextVideo functions */
    outInfo = (InterManager_OutputInfo*)outBuf;
    ret = InterManager_UGetOutputInfo(outInfo, sizeof(outBuf));
    if (ret) {
        UserModManager_UExit(1);
    }
    TV_Write = UserModManager_UGetFunction(outInfo->names, outInfo->names+outInfo->interSize, TextVideo__UWRITE, UserModManager_SYS_FUNCTION);
    if (IS_PTR_ERROR(TV_Write)) {
        UserModManager_UExit(1);
    }
    StartLibs();
    if (argc > 1)
        dir = FSManager_UOpen(argv[1], FSManager_READ+FSManager_DIRECTORY);
    else
        dir = FSManager_UOpen(".", FSManager_READ+FSManager_DIRECTORY);
    if (dir >= 0) {
        char buf[4096];
        int read;
        while ((read = FSManager_URead(dir, (FSManager_DirEntry*)buf, 4096)) >= 0) {
            int curr;
            curr = 0;
            while (curr < read) {
                TV_Write(buf+curr+2, (int)User_StringMan_Length(buf+curr+2));
                TV_Write("\n", sizeof("\n")-1);
                curr = curr + *(unsigned short*)(buf+curr);
            }
        }
        if (read != -E_END_REACHED)
            ret = -read;
        else
            ret = 0;
        FSManager_UClose(dir);
    } else
        ret = -dir;    /* main loop */
    if (ret)
        PrintError(ret);
    UserModManager_UExit(ret);
}

