#ifndef ELF_FILE_H
#define ELF_FILE_H

#include <fstream>
#include <elf.h>
#include <string>
#include <string.h>
#include "exceptions_file.h"

class ELFFile {
    protected:
        std::ifstream fileIn;
        Elf32_Ehdr header;
        Elf32_Shdr section;
        Elf32_Sym *symtab;
        Elf32_Rel *roDataReloc, *dataReloc, *codeReloc, *initReloc;
        char *strtab, *shstrtab;
        unsigned int status, nSymb, sizeStr, nDataReloc, nRODataReloc, nCodeReloc, nInitReloc;

    public:
        ELFFile();
        ELFFile(char *) throw (EFileError);
        ~ELFFile();
        void open(char *) throw (EFileError);
        unsigned char getClass();
        unsigned char getDataEnc();
        unsigned char getVersion();
        Elf32_Half getType();
        Elf32_Half getMachine();
        Elf32_Sym *getSymbol(unsigned int n);
        Elf32_Sym *getSymbol(unsigned int *n, int section);
        Elf32_Sym *getSymbol(const char *name);
        char *getString(unsigned int start);
        int getSectionNumber(const char *name);
        int getSectionStart(int sNumb);
        int getSectionSize(int sNumb);
        void read(char *buffer, int start, int size) throw (EFileError);
        Elf32_Rel *getDataReloc(unsigned int n);
        Elf32_Rel *getRODataReloc(unsigned int n);
        Elf32_Rel *getTextReloc(unsigned int n);
        Elf32_Rel *getInitReloc(unsigned int n);
};

#endif /* ELF_FILE_H */
