#ifndef MODULE_FILE_H
#define MODULE_FILE_H

#include <fstream>
#include <string>
#include <vector>
#include <algorithm>
#include <functional>
#include "common.h"
#include "exceptions_file.h"

class ModuleFile {
    protected:
        struct InternalHeader {
            char md5[16];
            char sign[4];
            unsigned int stackSize;
            int code;
            int codeSize;
            int roData;
            int roDataSize;
            int data;
            int dataSize;
            int uDataSize;
            int usedFuncs;
            int usedFuncsSize;
            int usedFuncReloc;
            int usedFuncRelocSize;
            int strings;
            short int stringsSize;
            short int comment;
        };

        struct InternalString {
            std::string str;
            short int iStr;
        };

        struct InternalSection {
            char *buffer;
            unsigned int size;
        };

    public:
        enum SectionType {ROData, Data, BSS, Code};

        enum RelocationType {Relative, Abs};

        struct UsedFuncRelocation {
            unsigned int offset;
            ModuleFile::RelocationType type;
            std::string name;
        };

    protected:
        std::fstream fileOut;
        ModuleFile::InternalHeader header;
        std::vector<ModuleFile::InternalString*> strings;
        std::vector<ModuleFile::UsedFuncRelocation*> *usedFuncRelocations;
        std::vector<Interface*> *usedFunctions;
        ModuleFile::InternalSection roData, data, code;
        int nextString;
        friend bool findStr(ModuleFile::InternalString *i);
        void DoMD5() throw (EFileError);
        bool GetUsedFuncNumber(unsigned int*,std::string&);

    public:
        struct Header {
            std::string *comment;
            unsigned int stackSize;
        };

        ModuleFile();
        ~ModuleFile();
        void open(const char *fileName) throw (EFileError);
        void close() throw (EFileError);
        void closeFD() throw (EFileError);
        void defHeader(ModuleFile::Header &header);
        void writeSection(char *buffer, unsigned int size, ModuleFile::SectionType type);
        // relocations must be sorted by offset
        void writeUsedFuncRelocations(std::vector<ModuleFile::UsedFuncRelocation*> *relocations);
        void writeUsedFunctions(std::vector<Interface*> *usedFunctions);
};

#endif /*MODULE_FILE_H*/
