%union {
   string* str;
   unsigned int uint;
}

%token <str> TOK_HEADER
%token <str> TOK_STRING
%token <uint> TOK_HEXNUMBER
%token <str> TOK_INTERFACE
%token <str> TOK_IMPLEMENTATION
%token <str> TOK_USEDFUNCTIONS
%token <str> TOK_OPEN
%token <str> TOK_CLOSE

%type <str> def header interface more_interfaces usedfunctions more_functions more_implementations more_usedfunctions

%start def

%%

def: header interface usedfunctions
     ;

header: TOK_HEADER TOK_OPEN TOK_STRING TOK_HEXNUMBER TOK_HEXNUMBER TOK_STRING TOK_STRING TOK_CLOSE
        { this->comment = $3;
          this->version = $4;
          this->prop = $5;
          this->start = $6;
          this->shutdown = $7;
        }
     ;

interface: TOK_INTERFACE TOK_OPEN TOK_STRING TOK_IMPLEMENTATION TOK_OPEN TOK_STRING TOK_STRING TOK_HEXNUMBER
           { if (!$3)
                throw ESyntaxeError(string("Interface must have a name."));
             if ($3->length() > 31)
                throw ESyntaxeError(string("Interface name cannot be bigger than 31 chars."));
             this->newInt = new Interface;
             this->newInt->name = $3;
             this->interfaces.push_back(this->newInt);
             if (!$6)
                throw ESyntaxeError(string("Implementation must have a name."));
             if ($6->length() > 31)
                throw ESyntaxeError(string("Implementation name cannot be bigger than 31 chars."));
             this->newImp = new Implementation;
             this->newImp->name = $6;
             this->newInt->implementations.push_back(this->newImp);
             if (!$7)
                throw ESyntaxeError(string("Function must have a name."));
             Function *newF = new Function;
             newF->name = $7;
             newF->prop = $8;
             this->newImp->functions.push_back(newF);
           }
           more_functions TOK_CLOSE more_implementations TOK_CLOSE more_interfaces
     ;

more_interfaces: /* empty */  { }
                 | interface
     ;

usedfunctions: /* empty */  { }
               | TOK_USEDFUNCTIONS TOK_OPEN TOK_STRING TOK_HEXNUMBER TOK_OPEN TOK_STRING TOK_OPEN TOK_STRING TOK_HEXNUMBER
               { if (!$3)
                     throw ESyntaxeError(string("Used interface must have a name."));
                 if ($3->length() > 31)
                     throw ESyntaxeError(string("Used interface name cannot be bigger than 31 chars."));
                 this->newUsedInt = new UsedInterface;
                 this->newUsedInt->name = $3;
                 this->newUsedInt->prop = $4;
                 this->usedInterfaces.push_back(this->newUsedInt);
                 if ($6 && $6->length() > 31)
                     throw ESyntaxeError(string("Used implementation name cannot be bigger than 31 chars."));
                 this->newUsedImp = new UsedImplementation;
                 this->newUsedImp->name = $6;
                 this->newUsedInt->imps.push_back(this->newUsedImp);
                 if (!$8)
                     throw ESyntaxeError(string("Used function must have a name."));
                 UsedFunction *newUF = new UsedFunction;
                 newUF->name = $8;
                 newUF->number = $9;
                 this->newUsedImp->functions.push_back(newUF);
               }
               more_usedfunctions TOK_CLOSE more_usedimplementations TOK_CLOSE more_usedinterfaces TOK_CLOSE
     ;

more_functions: /* empty */ { }
                | TOK_STRING TOK_HEXNUMBER
                { if (!$1)
                     throw ESyntaxeError(string("Function must have a name."));
                  Function *newF = new Function;
                  newF->name = $1;
                  newF->prop = $2;
                  this->newImp->functions.push_back(newF);
                }
                more_functions
     ;

more_implementations: /* empty */ { }
                      | TOK_IMPLEMENTATION TOK_OPEN TOK_STRING TOK_STRING TOK_HEXNUMBER
                      { if (!$3)
                           throw ESyntaxeError(string("Implementation must have a name."));
                        if ($3->length() > 31)
                           throw ESyntaxeError(string("Implementation name cannot be bigger than 31 chars."));
                        this->newImp = new Implementation;
                        this->newImp->name = $3;
                        this->newInt->implementations.push_back(this->newImp);
                        if (!$4)
                           throw ESyntaxeError(string("Function must have a name."));
                        Function *newF = new Function;
                        newF->name = $4;
                        newF->prop = $5;
                        this->newImp->functions.push_back(newF);
                      }
                      more_functions TOK_CLOSE more_implementations
     ;

more_usedfunctions: /* empty */ { }
                    | TOK_STRING TOK_HEXNUMBER
                    { if (!$1)
                         throw ESyntaxeError(string("Used function must have a name."));
                      UsedFunction *newUF = new UsedFunction;
                      newUF->name = $1;
                      newUF->number = $2;
                      this->newUsedImp->functions.push_back(newUF);
                    }
                    more_usedfunctions
     ;

more_usedimplementations: /* empty */ { }
                    | TOK_STRING TOK_OPEN TOK_STRING TOK_HEXNUMBER
                    { if ($1 && $1->length() > 31)
                          throw ESyntaxeError(string("Used implementation name cannot be bigger than 31 chars."));
                      this->newUsedImp = new UsedImplementation;
                      this->newUsedImp->name = $1;
                      this->newUsedInt->imps.push_back(this->newUsedImp);
                      if (!$3)
                          throw ESyntaxeError(string("Used function must have a name."));
                      UsedFunction *newUF = new UsedFunction;
                      newUF->name = $3;
                      newUF->number = $4;
                      this->newUsedImp->functions.push_back(newUF);
                    }
                    more_usedfunctions TOK_CLOSE more_usedimplementations
     ;

more_usedinterfaces: /* empty */  { }
               | TOK_STRING TOK_HEXNUMBER TOK_OPEN TOK_STRING TOK_OPEN TOK_STRING TOK_HEXNUMBER
               { if (!$1)
                     throw ESyntaxeError(string("Used interface must have a name."));
                 if ($1->length() > 31)
                     throw ESyntaxeError(string("Used interface name cannot be bigger than 31 chars."));
                 this->newUsedInt = new UsedInterface;
                 this->newUsedInt->name = $1;
                 this->newUsedInt->prop = $2;
                 this->usedInterfaces.push_back(this->newUsedInt);
                 if ($4 && $4->length() > 31)
                     throw ESyntaxeError(string("Used implementation name cannot be bigger than 31 chars."));
                 this->newUsedImp = new UsedImplementation;
                 this->newUsedImp->name = $4;
                 this->newUsedInt->imps.push_back(this->newUsedImp);
                 if (!$6)
                     throw ESyntaxeError(string("Used function must have a name."));
                 UsedFunction *newUF = new UsedFunction;
                 newUF->name = $6;
                 newUF->number = $7;
                 this->newUsedImp->functions.push_back(newUF);
               }
               more_usedfunctions TOK_CLOSE more_usedimplementations TOK_CLOSE more_usedinterfaces
     ;

%%

void yyerror (char *s) {
    cout << s << endl;
}

