#ifndef MODULE_FILE_H
#define MODULE_FILE_H

#include <fstream>
#include <string>
#include <vector>
#include <algorithm>
#include <functional>
#include "common.h"
#include "exceptions_file.h"

class ModuleFile {
    protected:
        struct InternalHeader {
            char md5[16];
            char sign[4];
            int code;
            int codeSize;
            int roData;
            int roDataSize;
            int data;
            int dataSize;
            int uDataSize;
            int usedFuncs;
            int usedFuncsSize;
            int usedFuncReloc;
            int usedFuncRelocSize;
            int interfaces;
            int interfacesSize;
            int roDataReloc;
            int roDataRelocSize;
            int dataReloc;
            int dataRelocSize;
            int codeReloc;
            int codeRelocSize;
            int strings;
            short int stringsSize;
            short int version;
            short int prop;
            short int comment;
            int startF;
            int shutdownF;
        };

        struct InternalString {
            std::string str;
            short int iStr;
        };

        struct InternalSection {
            char *buffer;
            unsigned int size;
        };

    public:
        enum SectionType {ROData, Data, BSS, Code};

        enum RelocationType {Relative, Abs};

        struct Relocation {
            ModuleFile::SectionType what;
            unsigned int offset;
            int add;
        };

        struct UsedFuncRelocation {
            unsigned int offset;
            ModuleFile::RelocationType type;
            std::string name;
        };

    protected:
        std::fstream fileOut;
        ModuleFile::InternalHeader header;
        std::vector<ModuleFile::InternalString*> strings;
        std::vector<Interface*> *interfaces;
        std::vector<ModuleFile::Relocation*> *roDataRelocs, *dataRelocs, *codeRelocs;
        std::vector<ModuleFile::UsedFuncRelocation*> *usedFuncRelocations;
        std::vector<UsedInterface*> *usedFunctions;
        ModuleFile::InternalSection roData, data, code;
        int nextString;
        friend bool findStr(ModuleFile::InternalString *i);
        void DoMD5() throw (EFileError);
        bool GetUsedFuncNumber(unsigned int*,std::string&);

    public:
        struct Header {
            std::string *comment;
            int version;
            int prop;
            unsigned int startF;
            unsigned int shutdownF;
        };

        ModuleFile();
        ~ModuleFile();
        void open(const char *fileName) throw (EFileError);
        void close() throw (EFileError);
        void closeFD() throw (EFileError);
        void defHeader(ModuleFile::Header &header);
        // relocations must be sorted by offset
        void writeCodeRelocations(std::vector<ModuleFile::Relocation*> *relocations);
        void writeDataRelocations(std::vector<ModuleFile::Relocation*> *relocations);
        void writeRODataRelocations(std::vector<ModuleFile::Relocation*> *relocations);
        void writeInterfaces(std::vector<Interface*> *interfaces);
        void writeSection(char *buffer, unsigned int size, ModuleFile::SectionType type);
        // relocations must be sorted by offset
        void writeUsedFuncRelocations(std::vector<ModuleFile::UsedFuncRelocation*> *relocations);
        void writeUsedFunctions(std::vector<UsedInterface*> *usedFunctions);
};

#endif /*MODULE_FILE_H*/
