/*
* Class to parse the spec file
* Copyright (C) 2000-2003, Luiz Henrique Shigunov (luizshigunov@users.sourceforge.net)
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330,
* Boston, MA 02111-1307, USA.
*/

#include <iostream>
#include "syntaxe.h"

extern unsigned int lineNumber;

using namespace std;

Syntaxe::Syntaxe() {
}

void Syntaxe::open(char* filename) throw (EFileError, ESyntaxeError) {
	lexFile.open(filename, ios::in | ios::binary);
	lex = new yyFlexLexer(&lexFile);
	if (yyparser())
		throw ESyntaxeError(string("Syntaxe error."));
}

int Syntaxe::getLine() {
	return lineNumber;
}

string *Syntaxe::getComment() {
	return comment;
}

string *Syntaxe::getStartF() {
    return start;
}

string *Syntaxe::getShutdownF() {
    return shutdown;
}

unsigned int Syntaxe::getVersion() {
	return version;
}

unsigned int Syntaxe::getProp() {
	return prop;
}

vector<Interface*> *Syntaxe::getInterfaces() {
	return &interfaces;
}

vector<UsedInterface*> *Syntaxe::getUsedFunctions() {
    if (!usedInterfaces.empty())
        return &usedInterfaces;
    return NULL;
}

void yyerror(char*);

int Syntaxe::yylex() {
	return lex->yylex();
}

// include parser
// make yyparse a Syntaxe member
#define yyparse Syntaxe::yyparser
#include "grammar.tab.c"

#undef yyparse
