#ifndef SYNTAXE_H
#define SYNTAXE_H

#include <fstream>
#include <string>
#include <vector>
#include <FlexLexer.h>
#include "common.h"
#include "exceptions_file.h"

class ESyntaxeError : public std::logic_error {
    public:
        ESyntaxeError(const std::string& what_arg) : logic_error(what_arg) { };
};

class EInvalidFunction : public std::logic_error {
    public:
        EInvalidFunction(const std::string& what_arg) : logic_error(what_arg) { };
};

class Syntaxe {
    protected:
        std::ifstream lexFile;
        yyFlexLexer *lex;
        std::string *comment, *start, *shutdown;
        unsigned int version, prop;
        Interface *newInt;
        Implementation *newImp;
        UsedInterface *newUsedInt;
        UsedImplementation *newUsedImp;
        std::vector<Interface*> interfaces;
        std::vector<UsedInterface*> usedInterfaces;
        int yyparser();
        int yylex();

    public:
        Syntaxe();
        void open(char *fileName) throw (EFileError, ESyntaxeError);
        int getLine();
        std::string *getComment();
        std::string *getStartF();
        std::string *getShutdownF();
        unsigned int getVersion();
        unsigned int getProp();
        std::vector<Interface*> *getInterfaces();
        std::vector<UsedInterface*> *getUsedFunctions();
};

#endif /*SYNTAXE_H*/
