%union {
   string* str;
   unsigned int uint;
}

%token <str> TOK_HEADER
%token <str> TOK_STRING
%token <uint> TOK_HEXNUMBER
%token <str> TOK_INTERFACE
%token <str> TOK_IMPLEMENTATION
%token <str> TOK_USEDFUNCTIONS
%token <str> TOK_OPEN
%token <str> TOK_CLOSE

%type <str> def header interface more_interfaces usedfunctions more_functions more_implementations more_usedfunctions

%start def

%%

def: header interface usedfunctions
     ;

header: TOK_HEADER TOK_OPEN TOK_STRING TOK_HEXNUMBER TOK_HEXNUMBER TOK_STRING TOK_STRING TOK_STRING TOK_CLOSE
        { this->comment = $3;
          this->version = $4;
          this->prop = $5;
          this->phase0Start = $6;
          this->phase1Start = $7;
          this->shutdown = $8;
        }
     ;

interface: TOK_INTERFACE TOK_OPEN TOK_STRING TOK_IMPLEMENTATION TOK_OPEN TOK_STRING TOK_STRING TOK_HEXNUMBER
           { if (!$3)
                throw ESyntaxeError(string("Interface must have a name."));
             if ($3->length() > 31)
                throw ESyntaxeError(string("Interface name cannot be bigger than 31 chars."));
             this->newInt = new Interface;
             this->newInt->name = $3;
             this->interfaces.push_back(this->newInt);
             if (!$6)
                throw ESyntaxeError(string("Implementation must have a name."));
             if ($6->length() > 31)
                throw ESyntaxeError(string("Implementation name cannot be bigger than 31 chars."));
             this->newImp = new Implementation;
             this->newImp->name = $6;
             this->newInt->implementations.push_back(this->newImp);
             if (!$7)
                throw ESyntaxeError(string("Function must have a name."));
             Function *newF = new Function;
             newF->name = $7;
             newF->prop = $8;
             this->newImp->functions.push_back(newF);
           }
           more_functions TOK_CLOSE more_implementations TOK_CLOSE more_interfaces
     ;

more_interfaces: /* empty */  { }
                 | interface
     ;

usedfunctions: /* empty */  { }
               | TOK_USEDFUNCTIONS TOK_OPEN TOK_STRING TOK_STRING TOK_STRING TOK_HEXNUMBER
               { if (!$3)
                    throw ESyntaxeError(string("Used function must have a name."));
                 UsedFunction *newUF = new UsedFunction;
                 newUF->name = $3;
                 if (!$4)
                    throw ESyntaxeError(string("Used function interface must have a name."));
                 if ($4->length() > 31)
                    throw ESyntaxeError(string("Used function interface name cannot be bigger than 31 chars."));
                 if ($5 && $5->length() > 31)
                    throw ESyntaxeError(string("Used function implementation name cannot be bigger than 31 chars."));
                 newUF->interface = $4;
                 newUF->implementation = $5;
                 newUF->number = $6;
                 this->usedFunctions.push_back(newUF);
               }
               more_usedfunctions TOK_CLOSE
     ;

more_functions: /* empty */ { }
                | TOK_STRING TOK_HEXNUMBER
                { if (!$1)
                     throw ESyntaxeError(string("Function must have a name."));
                  Function *newF = new Function;
                  newF->name = $1;
                  newF->prop = $2;
                  this->newImp->functions.push_back(newF);
                }
                more_functions
     ;

more_implementations: /* empty */ { }
                      | TOK_IMPLEMENTATION TOK_OPEN TOK_STRING TOK_STRING TOK_HEXNUMBER
                      { if (!$3)
                           throw ESyntaxeError(string("Implementation must have a name."));
                        if ($3->length() > 31)
                            throw ESyntaxeError(string("Implementation name cannot be bigger than 31 chars."));
                        this->newImp = new Implementation;
                        this->newImp->name = $3;
                        this->newInt->implementations.push_back(this->newImp);
                        if (!$4)
                           throw ESyntaxeError(string("Function must have a name."));
                        Function *newF = new Function;
                        newF->name = $4;
                        newF->prop = $5;
                        this->newImp->functions.push_back(newF);
                      }
                      more_functions TOK_CLOSE more_implementations
     ;

more_usedfunctions: /* empty */ { }
                    | TOK_STRING TOK_STRING TOK_STRING TOK_HEXNUMBER
                    { if (!$1)
                         throw ESyntaxeError(string("Used function must have a name."));
                      UsedFunction *newUF = new UsedFunction;
                      newUF->name = $1;
                      if (!$2)
                         throw ESyntaxeError(string("Used function interface must have a name."));
                      if ($2->length() > 31)
                         throw ESyntaxeError(string("Used function interface name cannot be bigger than 31 chars."));
                      if ($3 && $3->length() > 31)
                         throw ESyntaxeError(string("Used function implementation name cannot be bigger than 31 chars."));
                      newUF->interface = $2;
                      newUF->implementation = $3;
                      newUF->number = $4;
                      this->usedFunctions.push_back(newUF);
                    }
                    more_usedfunctions
     ;

%%

void yyerror (char *s) {
    cout << s << endl;
}

