#ifndef MODULE_FILE_H
#define MODULE_FILE_H

#include <fstream>
#include <string>
#include <vector>
#include <algorithm>
#include <functional>
#include "common.h"
#include "exceptions_file.h"

class ModuleFile {
    protected:
        struct InternalHeader {
            char md5[16];
            char sign[4];
            int code;
            int codeSize;
            int data;
            int dataSize;
            int uDataSize;
            int usedFuncs;
            int usedFuncsSize;
            int usedFuncReloc;
            int usedFuncRelocSize;
            int interfaces;
            int interfacesSize;
            int dataReloc;
            int dataRelocSize;
            int codeReloc;
            int codeRelocSize;
            int strings;
            short int stringsSize;
            short int version;
            short int prop;
            short int comment;
            int phase0StartF;
            int phase1StartF;
            int shutdownF;
        };

        struct InternalString {
            string str;
            short int iStr;
        };

        struct InternalSection {
            char *buffer;
            unsigned int size;
        };

	public:
		enum SectionType {Data, BSS, Code};

		enum RelocationType {Relative, Abs};

		struct Relocation {
			ModuleFile::SectionType what;
			unsigned int offset;
			int add;
		};

		struct UsedFuncRelocation {
			unsigned int offset;
			ModuleFile::RelocationType type;
			string name;
		};

	protected:
		fstream fileOut;
		ModuleFile::InternalHeader header;
		vector<ModuleFile::InternalString*> strings;
		vector<Interface*> *interfaces;
		vector<ModuleFile::Relocation*> *dataRelocs;
		vector<ModuleFile::Relocation*> *codeRelocs;
		vector<ModuleFile::UsedFuncRelocation*> *usedFuncRelocations;
		vector<UsedFunction*> *usedFunctions;
		ModuleFile::InternalSection data, code;
		int nextString;
		friend bool findStr(ModuleFile::InternalString *i);
        void DoMD5() throw (EFileError);

    public:
        struct Header {
            string *comment;
            int version;
            int prop;
            unsigned int phase0StartF;
            unsigned int phase1StartF;
            unsigned int shutdownF;
        };

        ModuleFile();
        ~ModuleFile();
        void open(const char *fileName) throw (EFileError);
        void close() throw (EFileError);
        void closeFD() throw (EFileError);
        void defHeader(ModuleFile::Header &header);
        // relocations must be sorted by offset
        void writeCodeRelocations(vector<ModuleFile::Relocation*> *relocations);
        // relocations must be sorted by offset
        void writeDataRelocations(vector<ModuleFile::Relocation*> *relocations);
        void writeInterfaces(vector<Interface*> *interfaces);
        void writeSection(char *buffer, unsigned int size, ModuleFile::SectionType type);
        // relocations must be sorted by offset
        void writeUsedFuncRelocations(vector<ModuleFile::UsedFuncRelocation*> *relocations);
        void writeUsedFunctions(vector<UsedFunction*> *usedFunctions);
};

#endif /*MODULE_FILE_H*/
