#ifndef SYNTAXE_H
#define SYNTAXE_H

#include <fstream>
#include <string>
#include <vector>
#include <FlexLexer.h>
#include "common.h"
#include "exceptions_file.h"

class ESyntaxeError : public logic_error {
    public:
        ESyntaxeError(const string& what_arg) : logic_error(what_arg) { };
};

class EInvalidFunction : public logic_error {
    public:
        EInvalidFunction(const string& what_arg) : logic_error(what_arg) { };
};

class Syntaxe {
    protected:
        ifstream lexFile;
        yyFlexLexer *lex;
        string *comment, *phase0Start, *phase1Start, *shutdown;
        unsigned int version, prop;
        Interface *newInt;
        Implementation *newImp;
        vector<Interface*> interfaces;
        vector<UsedFunction*> usedFunctions;
        int yyparser();
        int yylex();

    public:
        Syntaxe();
        void open(char *fileName) throw (EFileError, ESyntaxeError);
        int getLine();
        string *getComment();
        string *getPhase0StartF();
        string *getPhase1StartF();
        string *getShutdownF();
        unsigned int getVersion();
        unsigned int getProp();
        vector<Interface*> *getInterfaces();
        vector<UsedFunction*> *getUsedFunctions();
};

#endif /*SYNTAXE_H*/
