/* -*-Asm-*- */
/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 1999,2000,2001   Free Software Foundation, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
* Modified for use as ModulOS's boot sector
* Copyright (C) 2002, Luiz Henrique Shigunov (luizshigunov@users.sourceforge.net)
*/

/*
 *  defines for the code go here
 */

/* First 2 bytes: first sector to read, next 2 bytes: number of sectors to read */
#define SECTOR_ENTRY 4

/* stack top */
#define STACK_TOP 0x2000

/* where to start reading sectors */
#define BUFFER_SEG 0x7e0

/* Absolute addresses
   This makes the assembler generate the address without support
   from the linker. (ELF can't relocate 16-bit addresses!) */
#define ABS(x) (x-_start+0x7c00)

/* Print message string */
#define MSG(x)  movw $ABS(x), %si; call message

    .file   "bootsec.S"

    .text

/* Tell GAS to generate 16-bit instructions so that this code works in real mode. */
    .code16

.globl _start; _start:
    /*
     * _start is loaded at 0x7c00 and is jumped to with CS:IP 0:0x7c00
     */

    /* general setup */
    cli     /* we're not safe here! */

    /*
     * ljmp to the next instruction because some bogus BIOSes
     * jump to 07C0:0000 instead of 0000:7C00.
     */
    ljmp    $0, $ABS(real_start)

real_start:

    /* set up %ds and %ss as offset from 0 */
    xorw    %ax, %ax
    movw    %ax, %ds
    movw    %ax, %ss

    /* set up the REAL stack */
    movw    $STACK_TOP, %sp

    sti     /* we're safe again */

    /* save drive reference first thing! */
    pushw   %dx

    /*
     *  Determine the hard disk geometry from the BIOS!
     *  We do this first, so that LS-120 IDE floppies work correctly.
     */
    movb    $8, %ah
    int     $0x13
    /*
     *  The call failed, so maybe use the floppy probe instead.
     */
    jc      floppy_probe

final_init:
    movw    $ABS(sectors), %si

    /* save number of heads */
    xorl    %eax, %eax
    movb    %dh, %al
    incw    %ax
    movl    %eax, 4(%si)

    xorw    %dx, %dx
    movb    %cl, %dl
    shlw    $2, %dx
    movb    %ch, %al
    movb    %dh, %ah

    /* save number of cylinders */
    incw    %ax
    movw    %ax, 8(%si)

    xorw    %ax, %ax
    movb    %dl, %al
    shrb    $2, %al

    /* save number of sectors */
    movl    %eax, (%si)

    /* save segment with disk buffer */
    movw    $BUFFER_SEG, %bx
    pushw   %bx

    /* this sets up for the first run through "bootloop" */
    movw    $ABS(firstlist - SECTOR_ENTRY), %di

    /* this is the loop for reading the secondary boot-loader in */
bootloop:
    /* check the number of sectors to read */
    cmpb    $0, 2(%di)

    /* if zero, go to the start function */
    je      bootit

setup_sectors:	
    /* load logical sector start (bottom half) */
    movzxw  (%di), %eax

    /* zero %edx */
    xorl    %edx, %edx

    /* divide by number of sectors */
    divl    (%si)

    /* save sector start */
    movb    %dl, 10(%si)

    xorl    %edx, %edx  /* zero %edx */
    divl    4(%si)      /* divide by number of heads */

    /* save head start */
    movb    %dl, 11(%si)

    /* save cylinder start */
    movw    %ax, 12(%si)

    /* do we need too many cylinders? */
    cmpw    8(%si), %ax
    jge     geometry_error

    /* determine the maximum sector length of this read */
    movw    (%si), %ax      /* get number of sectors per track/head */

    /* subtract sector start */
    subb    10(%si), %al

    /* how many do we really want to read? */
    cmpw    %ax, 2(%di)     /* compare against total number of sectors */

    /* which is greater? */
    jg      2f

    /* if less than, set to total */
    movw    2(%di), %ax

2:
    /* subtract from total */
    subw    %ax, 2(%di)

    /* add into logical sector start */
    addw    %ax, (%di)

/*
 *  This is the loop for taking care of BIOS geometry translation (ugh!)
 */

    /* get high bits of cylinder */
    movb    13(%si), %dl

    shlb    $6, %dl         /* shift left by 6 bits */
    movb    10(%si), %cl    /* get sector */

    incb    %cl             /* normalize sector (sectors go from 1-N, not 0-(N-1) ) */
    orb     %dl, %cl        /* composite together */
    movb    12(%si), %ch    /* sector+hcyl in cl, cylinder in ch */

    /* restore segment */
    popw    %bx
    movw    %bx, %es

    /* restore %dx */
    popw    %dx
    pushw   %dx

    /* save segment */
    pushw   %bx

    /* head number */
    movb    11(%si), %dh

    pushw   %ax     /* save %ax from destruction! */

/*
 * BIOS call "INT 0x13 Function 0x2" to read sectors from disk into memory
 *  Call with   %ah = 0x2
 *              %al = number of sectors
 *              %ch = cylinder
 *              %cl = sector (bits 6-7 are high bits of "cylinder")
 *              %dh = head
 *              %dl = drive (0x80 for hard disk, 0x0 for floppy disk)
 *              %es:%bx = segment:offset of buffer
 *  Return:
 *          %al = 0x0 on success; err code on failure
 */

    xorw    %bx, %bx    /* %bx = 0, put it at 0 in the segment */
    movb    $0x2, %ah   /* function 2 */
    int     $0x13

    jc  read_error

    /* restore %ax */
    popw    %ax

    /* get current buffer segment */
    popw    %bx

    /* determine the next possible destination address (presuming 512 byte sectors!) */
    shlw    $5, %ax         /* shift %ax five bits to the left */
    addw    %ax, %bx        /* add the corrected value to the destination address for next time */
    pushw   %bx

    /* check if finished with this dataset */
    cmpw    $0, 2(%di)
    jne     setup_sectors

    /* update position to load from */
    subw    $SECTOR_ENTRY, %di

    /* jump to bootloop */
    jmp     bootloop

/* END OF MAIN LOOP */

bootit:
    /* remove saved %es */
    popw    %bx
    popw    %dx	/* this makes sure %dl is our "boot" drive */
    ljmp    $0, $0x7e00

/* END OF MAIN LOOP */

/*
 * BIOS Geometry translation error (past the end of the disk geometry!).
 */
geometry_error:
    MSG(geometry_error_string)
    jmp     general_error

/*
 * Read error on the disk.
 */
read_error:
    MSG(read_error_string)

general_error:
    MSG(general_error_string)

/* go here when you need to stop the machine hard after an error condition */
stop:   jmp     stop

sectors:
    .long   0
heads:
    .long   0
cylinders:
    .word   0
sector_start:
    .byte   0
head_start:
    .byte   0
cylinder_start:
    .word   0

geometry_error_string:  .string "Geom"
read_error_string:      .string "Read"
general_error_string:   .string " Error"

/*
 * message: write the string pointed to by %si
 *
 *   WARNING: trashes %si, %ax, and %bx
 */

    /*
     * Use BIOS "int 10H Function 0Eh" to write character in teletype mode
     *  %ah = 0xe   %al = character
     *  %bh = page  %bl = foreground color (graphics modes)
     */
1:
    movw    $0x0001, %bx
    movb    $0xe, %ah
    int     $0x10           /* display a byte */
message:
    lodsb
    cmpb    $0, %al
    jne     1b          /* if not end of string, jmp to display */
    ret

probe_values:
    .byte   36, 18, 15, 9, 0

floppy_probe:
/*
 *  Perform floppy probe.
 */

    movw    $ABS(probe_values-1), %si

probe_loop:
    /* reset floppy controller INT 13h AH=0 */
    xorw    %ax, %ax
    int     $0x13

    incw    %si
    movb    (%si), %cl

    /* if number of sectors is 0, display error and die */
    cmpb    $0, %cl
    jne     1f

/*
 * Floppy disk probe failure.
 */
    MSG(fd_probe_error_string)
    jmp     general_error

fd_probe_error_string:  .string "Floppy"

1:
    /* perform read */
    xorw    %bx, %bx
    movw    $0x201, %ax
    movb    $0, %ch
    movb    $0, %dh
    int     $0x13

    /* if error, jump to "probe_loop" */
    jc      probe_loop

    /* %cl is already the correct value! */
    movb    $1, %dh
    movb    $79, %ch

    jmp     final_init

     . = _start + 0x1fe

firstlist:  /* this label has to be after the list data!!! */

/* the last 2 bytes in the sector 0 contain the signature */
    .word   0xaa55
