/* disk_io.c - implement abstract BIOS disk input and output */
/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 1999, 2000, 2001  Free Software Foundation, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
* Modified for use as ModulOS's bootloader
* Copyright (C) 2002, Luiz Henrique Shigunov (luizshigunov@users.sourceforge.net)
*/

#include "shared.h"

/*
 *  Global variables describing details of the filesystem
 */

/* disk buffer parameters */
int buf_drive = -1;
int buf_track = -1;
static struct geometry buf_geom = {0, };

int rawread(int drive, int sector, int nsectors, char *buf) {
    /*
     *  Check track buffer.  If it isn't valid or it is from the
     *  wrong disk, then reset the disk geometry.
     */
    if (buf_drive != drive) {
        if (get_diskinfo(drive, &buf_geom))
            return 0;
        buf_drive = drive;
        buf_track = -1;
    }

    do {
        int soff, num_sect, bufaddr, track, size;

        /*  Get first sector of track  */
        soff = sector % buf_geom.sectors;
        track = sector - soff;
        num_sect = buf_geom.sectors - soff;
        bufaddr = BUFFERADDR + (soff * SECTOR_SIZE);

        if (track != buf_track) {
            int read_start = track, read_len = buf_geom.sectors;

            /*
             *  If there's more than one read in this entire loop, then
             *  only make the earlier reads for the portion needed.  This
             *  saves filling the buffer with data that won't be used!
             */
            if (nsectors > num_sect) {
                read_start = sector;
                read_len = num_sect;
                bufaddr = BUFFERADDR;
            }

            if (biosdisk(BIOSDISK_READ, drive, &buf_geom, read_start, read_len, BUFFERSEG))
                return 0;
            else
                buf_track = track;
        }

        if (nsectors > num_sect)
            size = num_sect;
        else
            size = nsectors;

        memmove(buf, (char*)bufaddr, size * SECTOR_SIZE);

        buf += size * SECTOR_SIZE;
        nsectors -= size;
        sector += num_sect;
    } while (nsectors);

    return 1;
}

int devread(int sector, int nsectors, char *buf) {
    return rawread(current_drive, sector, nsectors, buf);
}
