/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 1999, 2001  Free Software Foundation, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
* Modified for use as ModulOS's bootloader
* Copyright (C) 2002, Luiz Henrique Shigunov (luizshigunov@users.sourceforge.net)
*/

#include "shared.h"

/* Display an ASCII character. */
void putchar(int c) {
    if (c == '\n')
        console_putchar('\r');
    console_putchar(c);
}

void print(char *str) {
    while (*str)
        putchar(*(str++));
}

int substring (char *s1, char *s2) {
    while (*s1 == *s2) {
        /* The strings match exactly. */
        if (! *(s1++))
            return 0;
        s2++;
    }

    /* S1 is a substring of S2. */
    if (*s1 == 0)
        return -1;

    /* S1 isn't a substring. */
    return 1;
}

void *
grub_memmove (void *to, const void *from, int len)
{
       /* This assembly code is stolen from
          linux-2.2.2/include/asm-i386/string.h. This is not very fast
          but compact.  */
       int d0, d1, d2;

       if (to < from)
         {
           asm volatile ("cld\n\t"
                         "rep\n\t"
                         "movsb"
                         : "=&c" (d0), "=&S" (d1), "=&D" (d2)
                         : "0" (len),"1" (from),"2" (to)
                         : "memory");
         }
       else
         {
           asm volatile ("std\n\t"
                         "rep\n\t"
                         "movsb\n\t"
                         "cld"
                         : "=&c" (d0), "=&S" (d1), "=&D" (d2)
                         : "0" (len),
                         "1" (len - 1 + (const char *) from),
                         "2" (len - 1 + (char *) to)
                         : "memory");
         }

   return to;
}
