/* bios.c - implement C part of low-level BIOS disk input and output */
/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 1999,2000  Free Software Foundation, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
* Modified for use as ModulOS's starter
* Copyright (C) 2002, Luiz Henrique Shigunov (luizshigunov@users.sourceforge.net)
*/

#include "bios.h"

/* These are defined in asm.S, and never be used elsewhere, so declare the
   prototypes here.  */
extern int asm_biosdisk_standard (int ah, int drive,
			      int coff, int hoff, int soff,
			      int nsec, int segment);
extern int asm_get_diskinfo_standard (int drive,
				  unsigned long *cylinders,
				  unsigned long *heads,
				  unsigned long *sectors);

/* Read/write NSEC sectors starting from SECTOR in DRIVE disk with GEOMETRY
   from/into SEGMENT segment. If READ is BIOSDISK_READ, then read it,
   else if READ is BIOSDISK_WRITE, then write it. If an geometry error
   occurs, return BIOSDISK_ERROR_GEOMETRY, and if other error occurs, then
   return the error number. Otherwise, return 0.  */
int bios_disk(int read, int drive, struct geometry *geometry, int sector, int nsec, int segment) {
    int cylinder_offset, head_offset, sector_offset;
    int head;

    /* SECTOR_OFFSET is counted from one, while HEAD_OFFSET and
       CYLINDER_OFFSET are counted from zero.  */
    sector_offset = sector % geometry->sectors + 1;
    head = sector / geometry->sectors;
    head_offset = head % geometry->heads;
    cylinder_offset = head / geometry->heads;

    if (cylinder_offset >= geometry->cylinders)
        return 1;

    return asm_biosdisk_standard(read + 0x02, drive, cylinder_offset, head_offset, sector_offset,  nsec, segment);
}

/* Return the geometry of DRIVE in GEOMETRY. If an error occurs, return
   non-zero, otherwise zero.  */
int bios_get_diskinfo(int drive, struct geometry *geometry) {
    /* Clear the flags.  */
    geometry->flags = 0;

    /* First, try INT 13 AH=8h call.  */
    if (asm_get_diskinfo_standard(drive, &geometry->cylinders, &geometry->heads, &geometry->sectors))
        return 1;

    geometry->total_sectors = (geometry->cylinders * geometry->heads * geometry->sectors);

    return 0;
}
