#ifndef COMMON_H
#define COMMON_H

#define ALIGN_DATA_UP(data, align) ((((unsigned int)(data)) + (align)-1) & ~((align)-1))

/* stack */
#define STACK_ADDR 0x7c00

/* asm code that must be below 1MB */
#define LOW1MB_ASM 0x7c00

/* disk buffer */
#define DISK_BUFFERLEN   0x8000
#define DISK_BUFFERADDR  0x8000
#define DISK_BUFFERSEG   0x800
#define SECTOR_SIZE      0x200
#define SECTOR_BITS      9

#ifndef ASM
#include "avltree_in.h"

extern char *tmpAddr;

#define ALLOC_TMP_MEM(size) ((tmpAddr+ALIGN_DATA_UP(size, 4)) < (char*)0x90000 ? tmpAddr : NULL); tmpAddr += ALIGN_DATA_UP(size, 4)

typedef struct _Module {
    struct _Module *next;
    unsigned int fileSize;
    int prop;
    char *pos;
    char *codePos, *dataPos;
    char *path;
    char *pathPos;
    char *fileHeader;
    unsigned int refCounter;
    char *memStart;
    unsigned int memSize;
} Module;

typedef struct _Function {
    char *code;
    unsigned char prop;
    /*
    0x01 - system function
    0x02 - not implemented
    */
    unsigned char words;
    short reserved;
} Function;

typedef struct _Implementation {
    struct _Implementation *next;
    const char *name;
    char *namePos;
    char *pos;
    Function *functions;
    char *functionPos;
    Module *module;
    struct _Interface *interface;
    unsigned short interN;
    unsigned short impN;
} Implementation;

typedef struct _Interface {
    AVLTree_In_Node node;
    struct _Interface *next;
    const char *name;
    char *namePos;
    unsigned int nFunctions;
    Implementation *imps;
    char *pos;
} Interface;

typedef struct {
    /* first module of the list */
    char *modules;
    /* avl tree root */
    char *interfaces;
    char *pgDir;
    /* start of used memory */
    char *start;
    char *cfgStart;
    unsigned int cfgSize;
} StarterInfo;

#endif /* !ASM */
#endif /* COMMON_H */
