/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 1999, 2001  Free Software Foundation, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
* Modified for use as ModulOS's starter
* Copyright (C) 2002, Luiz Henrique Shigunov (luizshigunov@users.sourceforge.net)
*/

#include <stddef.h>

#include "common.h"
#include "disk_io.h"
#include "utils.h"

/* sizes are always in bytes, BLOCK values are always in DEV_BSIZE (sectors) */
#define DEV_BSIZE 512

/* include/linux/fs.h */
#define BLOCK_SIZE 1024		/* initial block size for superblock read */
/* made up, defaults to 1 but can be passed via mount_opts */
#define WHICH_SUPER 1
/* kind of from fs/ext2/super.c */
#define SBLOCK (WHICH_SUPER * BLOCK_SIZE / DEV_BSIZE)	/* = 2 */

/* include/asm-i386/types.h */
typedef __signed__ char __s8;
typedef unsigned char __u8;
typedef __signed__ short __s16;
typedef unsigned short __u16;
typedef __signed__ int __s32;
typedef unsigned int __u32;

/*
 * Constants relative to the data blocks, from ext2_fs.h
 */
#define EXT2_NDIR_BLOCKS                12
#define EXT2_IND_BLOCK                  EXT2_NDIR_BLOCKS
#define EXT2_DIND_BLOCK                 (EXT2_IND_BLOCK + 1)
#define EXT2_TIND_BLOCK                 (EXT2_DIND_BLOCK + 1)
#define EXT2_N_BLOCKS                   (EXT2_TIND_BLOCK + 1)

/* include/linux/ext2_fs.h */
struct ext2_super_block
  {
    __u32 s_inodes_count;	/* Inodes count */
    __u32 s_blocks_count;	/* Blocks count */
    __u32 s_r_blocks_count;	/* Reserved blocks count */
    __u32 s_free_blocks_count;	/* Free blocks count */
    __u32 s_free_inodes_count;	/* Free inodes count */
    __u32 s_first_data_block;	/* First Data Block */
    __u32 s_log_block_size;	/* Block size */
    __s32 s_log_frag_size;	/* Fragment size */
    __u32 s_blocks_per_group;	/* # Blocks per group */
    __u32 s_frags_per_group;	/* # Fragments per group */
    __u32 s_inodes_per_group;	/* # Inodes per group */
    __u32 s_mtime;		/* Mount time */
    __u32 s_wtime;		/* Write time */
    __u16 s_mnt_count;		/* Mount count */
    __s16 s_max_mnt_count;	/* Maximal mount count */
    __u16 s_magic;		/* Magic signature */
    __u16 s_state;		/* File system state */
    __u16 s_errors;		/* Behaviour when detecting errors */
    __u16 s_pad;
    __u32 s_lastcheck;		/* time of last check */
    __u32 s_checkinterval;	/* max. time between checks */
    __u32 s_creator_os;		/* OS */
    __u32 s_rev_level;		/* Revision level */
    __u16 s_def_resuid;		/* Default uid for reserved blocks */
    __u16 s_def_resgid;		/* Default gid for reserved blocks */
    __u32 s_reserved[235];	/* Padding to the end of the block */
  };

struct ext2_group_desc
  {
    __u32 bg_block_bitmap;	/* Blocks bitmap block */
    __u32 bg_inode_bitmap;	/* Inodes bitmap block */
    __u32 bg_inode_table;	/* Inodes table block */
    __u16 bg_free_blocks_count;	/* Free blocks count */
    __u16 bg_free_inodes_count;	/* Free inodes count */
    __u16 bg_used_dirs_count;	/* Directories count */
    __u16 bg_pad;
    __u32 bg_reserved[3];
  };

struct ext2_inode
  {
    __u16 i_mode;		/* File mode */
    __u16 i_uid;		/* Owner Uid */
    __u32 i_size;		/* 4: Size in bytes */
    __u32 i_atime;		/* Access time */
    __u32 i_ctime;		/* 12: Creation time */
    __u32 i_mtime;		/* Modification time */
    __u32 i_dtime;		/* 20: Deletion Time */
    __u16 i_gid;		/* Group Id */
    __u16 i_links_count;	/* 24: Links count */
    __u32 i_blocks;		/* Blocks count */
    __u32 i_flags;		/* 32: File flags */
    union
      {
	struct
	  {
	    __u32 l_i_reserved1;
	  }
	linux1;
	struct
	  {
	    __u32 h_i_translator;
	  }
	hurd1;
	struct
	  {
	    __u32 m_i_reserved1;
	  }
	masix1;
      }
    osd1;			/* OS dependent 1 */
    __u32 i_block[EXT2_N_BLOCKS];	/* 40: Pointers to blocks */
    __u32 i_version;		/* File version (for NFS) */
    __u32 i_file_acl;		/* File ACL */
    __u32 i_dir_acl;		/* Directory ACL */
    __u32 i_faddr;		/* Fragment address */
    union
      {
	struct
	  {
	    __u8 l_i_frag;	/* Fragment number */
	    __u8 l_i_fsize;	/* Fragment size */
	    __u16 i_pad1;
	    __u32 l_i_reserved2[2];
	  }
	linux2;
	struct
	  {
	    __u8 h_i_frag;	/* Fragment number */
	    __u8 h_i_fsize;	/* Fragment size */
	    __u16 h_i_mode_high;
	    __u16 h_i_uid_high;
	    __u16 h_i_gid_high;
	    __u32 h_i_author;
	  }
	hurd2;
	struct
	  {
	    __u8 m_i_frag;	/* Fragment number */
	    __u8 m_i_fsize;	/* Fragment size */
	    __u16 m_pad1;
	    __u32 m_i_reserved2[2];
	  }
	masix2;
      }
    osd2;			/* OS dependent 2 */
  };

/* linux/limits.h */
#define NAME_MAX         255	/* # chars in a file name */

/* linux/posix_type.h */
typedef long linux_off_t;

/* linux/ext2fs.h */
#define EXT2_NAME_LEN 255
struct ext2_dir_entry
  {
    __u32 inode;		/* Inode number */
    __u16 rec_len;		/* Directory entry length */
    __u8 name_len;		/* Name length */
    __u8 file_type;
    char name[EXT2_NAME_LEN];	/* File name */
  };

/* linux/ext2fs.h */
/*
 * EXT2_DIR_PAD defines the directory entries boundaries
 *
 * NOTE: It must be a multiple of 4
 */
#define EXT2_DIR_PAD                    4
#define EXT2_DIR_ROUND                  (EXT2_DIR_PAD - 1)
#define EXT2_DIR_REC_LEN(name_len)      (((name_len) + 8 + EXT2_DIR_ROUND) & \
                                         ~EXT2_DIR_ROUND)


/* ext2/super.c */
#define log2(n) ffz(~(n))

#define EXT2_SUPER_MAGIC      0xEF53	/* include/linux/ext2_fs.h */
#define EXT2_ROOT_INO              2	/* include/linux/ext2_fs.h */
#define PATH_MAX                1024	/* include/linux/limits.h */
#define MAX_LINK_COUNT             5	/* number of symbolic links to follow */

#define SUPERBLOCK st_super_block
#define GROUP_DESC cacheGroupBlk
#define DATABLOCK1 st_data_block1
#define DATABLOCK2 st_data_block2

/* linux/ext2_fs.h */
#define EXT2_ADDR_PER_BLOCK(s)          (EXT2_BLOCK_SIZE(s) / sizeof (__u32))
#define EXT2_ADDR_PER_BLOCK_BITS(s)		(log2(EXT2_ADDR_PER_BLOCK(s)))

/* linux/ext2_fs.h */
#define EXT2_BLOCK_SIZE_BITS(s)        ((s)->s_log_block_size + 10)
/* kind of from ext2/super.c */
#define EXT2_BLOCK_SIZE(s)	(1 << EXT2_BLOCK_SIZE_BITS(s))
/* linux/ext2fs.h */
#define EXT2_DESC_PER_BLOCK(s) \
     (EXT2_BLOCK_SIZE(s) / sizeof (struct ext2_group_desc))
/* linux/stat.h */
#define S_IFMT  00170000
#define S_IFLNK  0120000
#define S_IFREG  0100000
#define S_IFDIR  0040000
#define S_ISLNK(m)	(((m) & S_IFMT) == S_IFLNK)
#define S_ISREG(m)      (((m) & S_IFMT) == S_IFREG)
#define S_ISDIR(m)      (((m) & S_IFMT) == S_IFDIR)

static int mapblock1 = -1;
static int mapblock2 = -1;
static struct ext2_super_block *st_super_block = NULL;
static struct ext2_inode *st_root_inode = NULL;
static struct ext2_group_desc *cacheGroupBlk = NULL;
static int cacheGroupBlkNum = -1;
static char *cacheInodesBlk = NULL;
static int cacheInodesBlkNum = -1;
static char *st_data_block1 = NULL;
static char *st_data_block2 = NULL;
static char *cacheDirBlk = NULL;
static int cacheDirBlkNum = -1;
static char *cacheDirName = NULL;
static int cacheDirNameLen = 0;
static struct ext2_inode *cacheDirInode = NULL;

int ext2fs_dir(char*,struct ext2_inode*);
int ext2fs_mount(void);
int ext2fs_read(char*,unsigned int*,struct ext2_inode*);
int read_inode(int,struct ext2_inode*);

/* include/asm-i386/bitops.h */
/*
 * ffz = Find First Zero in word. Undefined if no zero exists,
 * so code should check against ~0UL first..
 */
static __inline__ unsigned long
ffz (unsigned long word)
{
  __asm__ ("bsfl %1,%0"
:	   "=r" (word)
:	   "r" (~word));
  return word;
}

/* read whole file to buf */
/* returns: 0 - success; 1 - error */
int fsys_read_file(char *filename, char *buf, unsigned int *size) {
    struct ext2_inode inode;

    /* find filename inode */
    if (ext2fs_dir(filename, &inode))
        return 1;

    /* ok, now read it! */
    if (ext2fs_read(buf, size, &inode))
        return 1;

    return 0;
}

/* check filesystem types and read superblock into memory buffer */
/* returns: 0 - success; 1 - error */
int fsys_mount(void) {
    st_super_block = (struct ext2_super_block*)ALLOC_TMP_MEM(sizeof(struct ext2_super_block));
    st_root_inode = (struct ext2_inode*)ALLOC_TMP_MEM(sizeof(struct ext2_inode));
    cacheDirName = ALLOC_TMP_MEM(1024);
    cacheDirInode = (struct ext2_inode*)ALLOC_TMP_MEM(sizeof(struct ext2_inode));
    if (!st_super_block || !st_root_inode || !cacheDirName || !cacheDirInode) {
        util_print("Not enough memory.\n");
        return 1;
    }

    if (disk_devread(SBLOCK, 1, (char*)SUPERBLOCK)) {
        util_print("IO error.\n");
        return 1;
    }

    if (SUPERBLOCK->s_magic != EXT2_SUPER_MAGIC) {
        util_print("Filesystem error.\n");
        return 1;
    }

    cacheInodesBlk = ALLOC_TMP_MEM(EXT2_BLOCK_SIZE(SUPERBLOCK));
    cacheGroupBlk = (struct ext2_group_desc*)ALLOC_TMP_MEM(EXT2_BLOCK_SIZE(SUPERBLOCK));
    st_data_block1 = ALLOC_TMP_MEM(2*EXT2_BLOCK_SIZE(SUPERBLOCK));
    st_data_block2 = st_data_block1 + EXT2_BLOCK_SIZE(SUPERBLOCK);
    cacheDirBlk = ALLOC_TMP_MEM(EXT2_BLOCK_SIZE(SUPERBLOCK));
    if (!cacheInodesBlk || !cacheGroupBlk || !st_data_block1 || !cacheDirBlk) {
        util_print("Not enough memory.\n");
        return 1;
    }

    /* read root inode */
    if (read_inode(EXT2_ROOT_INO, st_root_inode)) {
        return 1;
    }

    return 0;
}

/* Takes a file system block number and reads it into BUFFER. */
/* returns: 0 - success; 1 - error */
static int ext2_rdfsb(int fsblock, int buffer) {
    if (disk_devread(fsblock *(EXT2_BLOCK_SIZE(SUPERBLOCK) / DEV_BSIZE), EXT2_BLOCK_SIZE(SUPERBLOCK) / DEV_BSIZE, (char*)buffer)) {
        util_print("IO error.\n");
        return 1;
    }
    return 0;
}

/* from
  ext2/inode.c:ext2_bmap()
*/
/* Maps LOGICAL_BLOCK (the file offset divided by the blocksize) into
   a physical block (the location in the file system) via an inode. */
/* returns: > 0 - success; -1 - error */
static int ext2fs_block_map(int logical_block, struct ext2_inode *inode) {
    /* if it is directly pointed to by the inode, return that physical addr */
    if (logical_block < EXT2_NDIR_BLOCKS)
        return inode->i_block[logical_block];
    /* else */
    logical_block -= EXT2_NDIR_BLOCKS;
    /* try the indirect block */
    if (logical_block < EXT2_ADDR_PER_BLOCK (SUPERBLOCK)) {
        if (mapblock1 != 1 && ext2_rdfsb (inode->i_block[EXT2_IND_BLOCK], (int)DATABLOCK1))
            return -1;
        mapblock1 = 1;
        return ((__u32 *) DATABLOCK1)[logical_block];
    }
    /* else */
    logical_block -= EXT2_ADDR_PER_BLOCK (SUPERBLOCK);
    /* now try the double indirect block */
    if (logical_block < (1 << (EXT2_ADDR_PER_BLOCK_BITS (SUPERBLOCK) * 2))) {
        int bnum;
        if (mapblock1 != 2 && ext2_rdfsb (inode->i_block[EXT2_DIND_BLOCK], (int)DATABLOCK1))
            return -1;
        mapblock1 = 2;
        if ((bnum = (((__u32 *) DATABLOCK1) [logical_block >> EXT2_ADDR_PER_BLOCK_BITS (SUPERBLOCK)]))
            != mapblock2 && ext2_rdfsb (bnum, (int)DATABLOCK2))
            return -1;
        mapblock2 = bnum;
        return ((__u32 *) DATABLOCK2)[logical_block & (EXT2_ADDR_PER_BLOCK (SUPERBLOCK) - 1)];
    }
    /* else */
    mapblock2 = -1;
    logical_block -= (1 << (EXT2_ADDR_PER_BLOCK_BITS (SUPERBLOCK) * 2));
    if (mapblock1 != 3 && ext2_rdfsb (inode->i_block[EXT2_TIND_BLOCK], (int)DATABLOCK1))
        return -1;
    mapblock1 = 3;
    if (ext2_rdfsb (((__u32 *) DATABLOCK1)[logical_block >> (EXT2_ADDR_PER_BLOCK_BITS (SUPERBLOCK) * 2)], (int)DATABLOCK2))
        return -1;
    if (ext2_rdfsb (((__u32 *) DATABLOCK2)[(logical_block >> EXT2_ADDR_PER_BLOCK_BITS (SUPERBLOCK)) & (EXT2_ADDR_PER_BLOCK (SUPERBLOCK) - 1)], (int)DATABLOCK2))
        return -1;
    return ((__u32 *) DATABLOCK2)[logical_block & (EXT2_ADDR_PER_BLOCK (SUPERBLOCK) - 1)];
}

/* preconditions: all preconds of ext2fs_block_map */
/* reads all file to buf */
/* returns: 0 - success; 1 - error */
int ext2fs_read(char *buf, unsigned int *size, struct ext2_inode *inode) {
    int logical_block, map, blk_len;

    *size = inode->i_size;
    blk_len = (ALIGN_DATA_UP(inode->i_size, EXT2_BLOCK_SIZE(SUPERBLOCK))) >> EXT2_BLOCK_SIZE_BITS(SUPERBLOCK);
    logical_block = 0;
    do {
        map = ext2fs_block_map(logical_block, inode);
        if (map < 0)
            break;
        if (disk_devread(map * (EXT2_BLOCK_SIZE(SUPERBLOCK) / DEV_BSIZE), EXT2_BLOCK_SIZE(SUPERBLOCK) / DEV_BSIZE, buf)) {
            util_print("ext2fs_read: IO error.\n");
            return 1;
        }
        buf += EXT2_BLOCK_SIZE(SUPERBLOCK);
        logical_block++;
    } while (--blk_len);
    return 0;
}

/* This functions reads inode into buf
 * preconditions: ext2fs_mount already executed, therefore supblk in buffer
 *   known as SUPERBLOCK
 * postconditions: on a nonzero return, buffer known as INODE contains the inode
 * side effects: messes up GROUP_DESC buffer area
 */
/* returns: 0 - success; 1 - error */
int read_inode(int inode, struct ext2_inode *buf) {
    int group_id;                   /* which group the inode is in */
    int group_desc;                 /* fs pointer to that group */
    int desc;                       /* index within that group */
    int ino_blk;                    /* fs pointer of the inode's information */
    struct ext2_group_desc *gdp;
    struct ext2_inode *raw_inode;   /* inode info corresponding to current_ino */

    /* look up an inode */
    group_id = (inode - 1) / (SUPERBLOCK->s_inodes_per_group);
    group_desc = group_id >> log2 (EXT2_DESC_PER_BLOCK(SUPERBLOCK));
    desc = group_id & (EXT2_DESC_PER_BLOCK(SUPERBLOCK) - 1);
    if (group_desc != cacheGroupBlkNum) {
        if (ext2_rdfsb ((WHICH_SUPER + group_desc + SUPERBLOCK->s_first_data_block), (int)GROUP_DESC))
            return 1;
        cacheGroupBlkNum = group_desc;
    }
    gdp = GROUP_DESC;
    ino_blk = gdp[desc].bg_inode_table + (((inode - 1) % (SUPERBLOCK->s_inodes_per_group))
              >> log2(EXT2_BLOCK_SIZE(SUPERBLOCK) / sizeof(struct ext2_inode)));
    if (ino_blk != cacheInodesBlkNum) {
        if (ext2_rdfsb (ino_blk, (int)cacheInodesBlk))
            return 1;
        cacheInodesBlkNum = ino_blk;
    }

    /* reset indirect blocks! */
    mapblock2 = mapblock1 = -1;

    raw_inode = (struct ext2_inode*)cacheInodesBlk + ((inode - 1) & (EXT2_BLOCK_SIZE(SUPERBLOCK) / sizeof(struct ext2_inode) - 1));

    /* copy inode to fixed location */
    util_memmove(buf, (void*)raw_inode, sizeof(struct ext2_inode));

    return 0;
}

/* This functions look ups filename and put its inode in inode
 * preconditions: ext2fs_mount already executed, therefore supblk in buffer
 *   known as SUPERBLOCK
 * postconditions: on a nonzero return, inode contains the
 *   inode of the file we were trying to look up
 * side effects: messes up GROUP_DESC buffer area
 */
/* returns: 0 - success; 1 - error */
int ext2fs_dir(char *filename, struct ext2_inode *inode) {
    int str_chk;    /* used to hold the results of a string compare */
    int off;        /* offset within block of directory entry (off mod blocksize) */
    int loc;        /* location within a directory */
    int blk;        /* which data blk within dir entry (off div blocksize) */
    long map;       /* fs pointer of a particular block from dir entry */
    struct ext2_dir_entry *dp;  /* pointer to directory entry */
    char *rest, ch, *lastS;
    int dirLen;
    struct ext2_inode *currInode;

    /* first check cache */
    lastS = util_strrchr(filename, '/');
    dirLen = lastS - filename + 1;
    if (dirLen == cacheDirNameLen && !util_strncmp(filename, cacheDirName, dirLen)) {
        /* get cache values */
        filename = lastS + 1;
    } else {
        /* put into the cache */
        cacheDirNameLen = dirLen;
        util_memmove(cacheDirName, filename, dirLen);
        /* start from root */
        util_memmove(cacheDirInode, st_root_inode, sizeof(struct ext2_inode));
        cacheDirBlkNum = -1;
    }
    currInode = cacheDirInode;

    while (1) {
        /* skip over slashes */
        while (*filename == '/') {
            filename++;
        }

        /* if this isn't a directory of sufficient size to hold our file, abort */
        if (!(currInode->i_size) || !S_ISDIR (currInode->i_mode)) {
            util_print(filename);
            util_print(" not found.\n");
            return 1;
        }

        /* skip to next slash or end of filename (space) */
        for (rest = filename; (ch = *rest) && ch != '/'; rest++);

        /* look through this directory and find the next filename component */
        /* invariant: rest points to slash after the next filename component */
        *rest = 0;
        loc = 0;
        do {
            /* if our location/byte offset into the directory exceeds the size, give up */
            if (loc >= currInode->i_size) {
                util_print(filename);
                util_print(" not found.\n");
                return 1;
            }

            /* else, find the (logical) block component of our location */
            blk = loc >> EXT2_BLOCK_SIZE_BITS (SUPERBLOCK);

            /* we know which logical block of the directory entry we are looking
            for, now we have to translate that to the physical (fs) block on
            the disk */
            if (cacheDirBlkNum != blk) {
                map = ext2fs_block_map (blk, currInode);
                if ((map < 0) || ext2_rdfsb (map, (int)cacheDirBlk))
                    return 1;
                cacheDirBlkNum = blk;
            }
            off = loc & (EXT2_BLOCK_SIZE (SUPERBLOCK) - 1);
            dp = (struct ext2_dir_entry *) (cacheDirBlk + off);
            /* advance loc prematurely to next on-disk directory entry  */
            loc += dp->rec_len;

            /* NOTE: ext2fs filenames are NOT null-terminated */
            str_chk = -1;
            if (dp->inode) {
                str_chk = util_strncmp(filename, dp->name, dp->name_len);
            }
        } while (!dp->inode || str_chk);

        filename = rest;
        *filename = ch;

        /* we've found the file */
        if (!*filename) {
            currInode = inode;
        }

        /* ok, now read inode */
        if (read_inode(dp->inode, currInode))
            return 1;

        if (S_ISLNK(currInode->i_mode)) {
            util_print(filename);
            util_print(" is a symbolic link.\n");
            return 1;
        }

        if (!*filename) {
            /* If not regular file, error! */
            if (!S_ISREG(currInode->i_mode)) {
                util_print("Invalid file type!\n");
                return 1;
            }
            return 0;
        }
        /* new entry. Reset it */
        cacheDirBlkNum = -1;
    }
}
