/*
* setup_mheader - sets up the multiboot header of the starter
* Copyright (C) 2002, Luiz Henrique Shigunov (luizshigunov@users.sourceforge.net)
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330,
* Boston, MA 02111-1307, USA.
*/

#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "multiboot.h"

/* Must pass video information to OS.  */
#define MULTIBOOT_VIDEO_MODE            0x00000004

/* This flag indicates the use of the address fields in the header.  */
#define MULTIBOOT_AOUT_KLUDGE           0x00010000

/*
 * The entire multiboot_header must be contained
 * within the first MULTIBOOT_SEARCH bytes of the kernel image.
 */
#define MULTIBOOT_SEARCH                8192
#define MULTIBOOT_FOUND(addr, len) \
  (! ((addr) & 0x3) \
   && (len) >= 12 \
   && *((int *) (addr)) == MULTIBOOT_HEADER_MAGIC \
   && ! (*((unsigned *) (addr)) + *((unsigned *) (addr + 4)) \
         + *((unsigned *) (addr + 8))) \
   && (! (MULTIBOOT_AOUT_KLUDGE & *((int *) (addr + 4))) || (len) >= 32) \
   && (! (MULTIBOOT_VIDEO_MODE & *((int *) (addr + 4))) || (len) >= 48))

int main(int argc, char *argv[]) {
    char buffer[MULTIBOOT_SEARCH];
    struct multiboot_header *header;
    struct stat st_buf;
    int fd, len, i, size;

    fd = open("starter", O_RDWR);
    if (fd < 0) {
        perror("Cannot open starter");
        return 1;
    }

    if (!(len = read (fd, buffer, MULTIBOOT_SEARCH)) || len < 32) {
        perror("Cannot read enough bytes from starter");
        return 1;
    }

    header = NULL;
    for (i = 0; i < len; i++) {
        if (MULTIBOOT_FOUND ((int) (buffer + i), len - i)) {
            header = (struct multiboot_header*)(buffer + i);
            break;
        }
    }
    if (!header) {
        printf("Cannot find multiboot header!\n");
        return 1;
    }

    /* get starter size */
    if (fstat(fd, &st_buf) == -1) {
        perror("Cannot stat starter");
        return 1;
    }
    if (lseek(fd, (int)((char*)&header->load_end_addr-buffer), SEEK_SET) == -1) {
        perror("Cannot seek to header");
        return 1;
    }
    size = st_buf.st_size + header->load_addr;
    if (write(fd, (char*)&size, 4) != 4) {
        perror("Cannot write header");
        return 1;
    }
    close(fd);
    return 0;
}
