/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 1999, 2001  Free Software Foundation, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
* Modified for use as ModulOS's starter
* Copyright (C) 2002, Luiz Henrique Shigunov (luizshigunov@users.sourceforge.net)
*/

#include <stddef.h>

#include "asm.h"

/* Display an ASCII character. */
void util_putchar(int c) {
    if (c == '\n')
        asm_console_putchar('\r');
    asm_console_putchar(c);
}

void util_print(char *str) {
    while (*str)
        util_putchar(*(str++));
}

void *util_memmove (void *to, const void *from, int len)
{
   /* This assembly code is stolen from
      linux-2.2.2/include/asm-i386/string.h. This is not very fast
      but compact.  */
   int d0, d1, d2;

   if (to < from)
     {
       asm volatile ("cld\n\t"
                     "rep\n\t"
                     "movsb"
                     : "=&c" (d0), "=&S" (d1), "=&D" (d2)
                     : "0" (len),"1" (from),"2" (to)
                     : "memory");
     }
   else
     {
       asm volatile ("std\n\t"
                     "rep\n\t"
                     "movsb\n\t"
                     "cld"
                     : "=&c" (d0), "=&S" (d1), "=&D" (d2)
                     : "0" (len),
                     "1" (len - 1 + (const char *) from),
                     "2" (len - 1 + (char *) to)
                     : "memory");
     }

   return to;
}

void util_memset(void *mem, unsigned int c, unsigned int count) {
    int d0, d1;
    __asm__ __volatile__(
        "cld\n\t"
        "rep\n\t"
        "stosb"
        : "=&c" (d0), "=&D" (d1)
        :"a" (c),"1" (mem),"0" (count)
        :"memory");
}

char *util_strrchr(const char *s, int c) {
    int d0, d1;
    register char * __res;
    __asm__ __volatile__(
        "cld\n\t"
        "movb %%al,%%ah\n"
        "1:\tlodsb\n\t"
        "cmpb %%ah,%%al\n\t"
        "jne 2f\n\t"
        "leal -1(%%esi),%0\n"
        "2:\ttestb %%al,%%al\n\t"
        "jne 1b"
        :"=g" (__res), "=&S" (d0), "=&a" (d1) :"0" (0),"1" (s),"2" (c));
    return __res;
}

int util_strcmp(const char *s1, const char *s2) {
    int d0, d1;
    register int __res;
    __asm__ __volatile__(
        "cld\n"
        "1:\tlodsb\n\t"
        "scasb\n\t"
        "jne 2f\n\t"
        "testb %%al,%%al\n\t"
        "jne 1b\n\t"
        "xorl %%eax,%%eax\n\t"
        "jmp 3f\n"
        "2:\tsbbl %%eax,%%eax\n\t"
        "orb $1,%%al\n"
        "3:"
        :"=a" (__res), "=&S" (d0), "=&D" (d1)
        :"1" (s1),"2" (s2));
    return __res;
}

int util_strncmp(const char *s1, const char *s2, unsigned int count) {
    register int __res;
    int d0, d1, d2;
    __asm__ __volatile__(
        "cld\n"
        "1:\tdecl %3\n\t"
        "js 2f\n\t"
        "lodsb\n\t"
        "scasb\n\t"
        "jne 3f\n\t"
        "testb %%al,%%al\n\t"
        "jne 1b\n"
        "2:\txorl %%eax,%%eax\n\t"
        "jmp 4f\n"
        "3:\tsbbl %%eax,%%eax\n\t"
        "orb $1,%%al\n"
        "4:"
    :"=a" (__res), "=&S" (d0), "=&D" (d1), "=&c" (d2)
    :"1" (s1),"2" (s2),"3" (count));
        return __res;
}

unsigned int util_strlen(const char *s) {
    int d0;
    register int __res;
    __asm__ __volatile__(
        "cld\n\t"
        "repne\n\t"
        "scasb\n\t"
        "notl %0\n\t"
        "decl %0"
        :"=c" (__res), "=&D" (d0) :"1" (s),"a" (0), "0" (0xffffffff));
    return __res;
}

#define ULONG_MAX       4294967295UL
#define LONG_MAX        2147483647L
#define LONG_MIN        (-LONG_MAX - 1L)

long int util_strtol(const char *str, char **endstr, int base) {
    int negative, overflow;
    register unsigned long int cutoff, i;
    register unsigned int cutlim;
    register const char *s;
    register unsigned char c;
    const char *save;

    if (base < 0 || base == 1 || base > 36) {
        return 0;
    }
    save = s = str;

    /* Skip white space */
    while ((*s == ' ') || (*s == '\f') || (*s == '\n') || (*s == '\r') || (*s == '\t') || (*s == '\v'))
            ++s;
    if (*s == '\0')
            goto noconv;

    /* Check for a sign */
    if (*s == '-') {
            negative = 1;
            ++s;
    } else if (*s == '+') {
            negative = 0;
            ++s;
    } else
            negative = 0;

    /* Recognize number prefix and if BASE is zero, figure it out ourselves */
    if (*s == '0') {
            if ((base == 0 || base == 16) && (s[1] & 0xdf) == 'X') {
                    s += 2;
                    base = 16;
            } else if (base == 0)
                    base = 8;
    } else if (base == 0)
            base = 10;

    /* Save the pointer so we can check later if anything happened */
    save = s;
    cutoff = ULONG_MAX / (unsigned long int) base;
    cutlim = ULONG_MAX % (unsigned long int) base;
    overflow = 0;
    i = 0;
    c = *s;
    for (;c != '\0'; c = *++s) {
            if (c >= '0' && c <= '9')
                    c -= '0';
            else if ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z'))
                    c = (c & 0xdf) - 'A' + 10;
            else
                    break;
            if ((int) c >= base)
                    break;
            /* Check for overflow */
            if (i > cutoff || (i == cutoff && c > cutlim))
                    overflow = 1;
            else {
                    i *= (unsigned long int) base;
                    i += c;
            }
    }

    /* Check if anything actually happened */
    if (s == save)
            goto noconv;

    /* Store in endstr the address of one character
    past the last character we converted */
    if (endstr != NULL)
            *endstr = (char *)s;

    /* Check for a value that is within the range of
    `unsigned LONG int', but outside the range of `LONG int' */
    if (overflow == 0 && i > (negative ? -((unsigned long int) (LONG_MIN + 1)) + 1
          : (unsigned long int) LONG_MAX))
            overflow = 1;
    if (overflow) {
            return negative ? LONG_MIN : LONG_MAX;
    }
    /* Return the result of the appropriate sign */
    return negative ? -i : i;

noconv:
    /* We must handle a special case here: the base is 0 or 16 and the
    first two characters are '0' and 'x', but the rest are no
    hexadecimal digits.  This is no error case.  We return 0 and
    endstr points to the `x` */
    if (endstr != NULL) {
            if (save - str >= 2 && (save[-1] & 0xdf) == 'X' && save[-2] == '0')
                    *endstr = (char *) &save[-1];
            else
                    /* There was no number to convert */
                    *endstr = (char *) str;
    }
    return 0L;
}
