/*
*****************************************
* This file implements the IntController_i386 interface
*
* Version: 0.2.1
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 20/04/2003
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2003, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include "common_def.h"

static int spin;
unsigned char masterICW2 = 0xf0; /* IRQ0 interruption number */
unsigned char slaveICW2  = 0xf8; /* IRQ8 interruption number */
unsigned char masterOCW1 = 0xfb; /* master PIC mask */
unsigned char slaveOCW1  = 0xff; /* slave PIC mask */

void IntController_i386_EndIRQ(unsigned int irq) {
    __asm__ __volatile__ ("outb %%al,$0x20" : :"al" (0x20));
    if (irq > 7)
        __asm__ __volatile__ ("outb %%al,$0xa0" : : "al" (0x20));
    return;
}

int IntController_i386_MaskIRQ(unsigned int irq) {
    int port;
    unsigned char *mask;

    if (irq > 15)
        return E_BAD_VALUE;
    port = 0x21;
    mask = &masterOCW1;
    if (irq > 7) {
        port = 0xa1;
        irq = irq - 8;
        mask = &slaveOCW1;
    }
    TaskManager_SpinLock(&spin);
    *mask = *mask | (1 << irq);
    __asm__ __volatile__ ("outb %%al, %%dx" : : "a" (*mask), "d" (port));
    TaskManager_SpinUnlock(&spin);
    return E_OK;
}

int IntController_i386_RedirectIRQ(unsigned int interrupt) {
    int flags;

    if (interrupt > 240)
        return E_BAD_VALUE;
    flags = IntManager_i386_NoInt();
    masterICW2 = interrupt;
    slaveICW2 = interrupt + 8;
    /* restart PIC with new values */
    __asm__ __volatile__ (
        "movb $0x11, %%al\n"
        "outb %%al, $0x20\n"
        "outb %%al, $0xa0\n"
        "movb (masterICW2), %%al\n"
        "outb %%al, $0x21\n"
        "movb (slaveICW2), %%al\n"
        "outb %%al, $0xa1\n"
        "movb $0x4, %%al\n"
        "outb %%al, $0x21\n"
        "movb $0x2, %%al\n"
        "outb %%al, $0xa1\n"
        "movb $0x1, %%al\n"
        "outb %%al, $0x21\n"
        "outb %%al, $0xa1\n"
        "movb (masterOCW1), %%al\n"
        "outb %%al, $0x21\n"
        "movb (slaveOCW1), %%al\n"
        "outb %%al, $0xa1" : : : "%eax");
    IntManager_i386_RestoreInt(flags);
    return E_OK;
}

int IntController_i386_UnmaskIRQ(unsigned int irq) {
    int port;
    unsigned char *mask;

    if (irq > 15)
        return E_BAD_VALUE;
    port = 0x21;
    mask = &masterOCW1;
    if (irq > 7) {
        port = 0xa1;
        irq = irq - 8;
        mask = &slaveOCW1;
    }
    TaskManager_SpinLock(&spin);
    *mask = *mask & !(1 << irq);
    __asm__ __volatile__ ("outb %%al, %%dx" : : "a" (*mask), "d" (port));
    TaskManager_SpinUnlock(&spin);
    return E_OK;
}
