/*
*****************************************
* This file implements the start and shutdown functions
*
* Version: 0.2.4
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 20/04/2003
*
* Copyright (C) 2000-2003, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include "common_def.h"

SysModManager_Module *ID;
extern DMAController_i386_DMA_Channel dmaChannels[8];
extern unsigned char masterICW2;
extern unsigned char slaveICW2;
extern unsigned char masterOCW1;
extern unsigned char slaveOCW1;

int Shutdown(void) {
    int i, ret;

    for (i=0; i<= DMAController_i386_MAX_DMA_CHANNEL; i++) {
        if (dmaChannels[i].logAddr) {
            ret = MemManager_i386_FreePages(dmaChannels[i].logAddr, dmaChannels[i].size, 0);
            ASSERT_ERROR(ret);
        }
    }
    ret = IOPortManager_Free(ID, 0);
    ASSERT_ERROR(ret);
    ret = IOPortManager_Free(ID, 0x80);
    ASSERT_ERROR(ret);
    ret = IOPortManager_Free(ID, 0xc0);
    ASSERT_ERROR(ret);
    ret = IOPortManager_Free(ID, 0x40);
    ASSERT_ERROR(ret);
    ret = IOPortManager_Free(ID, 0x20);
    ASSERT_ERROR(ret);
    ret = IOPortManager_Free(ID, 0xa0);
    ASSERT_ERROR(ret);
    return 0;
}

int Phase0Start(SysModManager_Module *module) {
    int ret;

    ID = module;
    /* start all used modules in phase 0 */
    if ((ret = SysModManager_StartModule((unsigned int)IntManager_i386_NoInt)) ||
        (ret = SysModManager_StartModule((unsigned int)IOPortManager_Alloc))) {
        DEBUG_MSG_PRE(ret);
        goto errorReturn;
    }
    /* PIC ports */
    if ((ret = IOPortManager_Alloc(module, 0x20, 0x21, 0))) {
        DEBUG_MSG_PRE(ret);
        goto errorReturn;
    }
    if ((ret = IOPortManager_Alloc(module, 0xa0, 0xa1, 0))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreePort20;
    }
    return 0;

errorFreePort20:
    ret = IOPortManager_Free(module, 0x20);
    ASSERT_ERROR(ret);
errorReturn:
    return 1;
}

int Phase1Start(SysModManager_Module *module) {
    int flags, ret;

    /* start all used modules in phase 1 */
    if ((ret = SysModManager_StartModule((unsigned int)IntManager_i386_NoInt)) ||
        (ret = SysModManager_StartModule((unsigned int)IOPortManager_Alloc)) ||
        (ret = SysModManager_StartModule((unsigned int)MemManager_i386_FreePages)) ||
        (ret = SysModManager_StartModule((unsigned int)TaskManager_SpinLock))) {
        DEBUG_MSG_PRE(ret);
        goto errorReturn;
    }
    /* Timer ports */
    if ((ret = IOPortManager_Alloc(module, 0x40, 0x43, 0))) {
        DEBUG_MSG_PRE(ret);
        goto errorReturn;
    }
    /* DMA ports */
    if ((ret = IOPortManager_Alloc(module, 0, 0x1f, 0))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreePort40;
    }
    if ((ret = IOPortManager_Alloc(module, 0x80, 0x8f, 0))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreePort00;
    }
    if ((ret = IOPortManager_Alloc(module, 0xc0, 0xdf, 0))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreePort80;
    }
    /* DMA channel 4 is used by DMA */
    dmaChannels[4].module = module;

    flags = IntManager_i386_NoInt();
    /* Setup PIC */
    __asm__ __volatile__ (
        "movb $0x11, %%al\n"
        "outb %%al, $0x20\n"
        "outb %%al, $0xa0\n"
        "movb (masterICW2), %%al\n"
        "outb %%al, $0x21\n"
        "movb (slaveICW2), %%al\n"
        "outb %%al, $0xa1\n"
        "movb $0x4, %%al\n"
        "outb %%al, $0x21\n"
        "movb $0x2, %%al\n"
        "outb %%al, $0xa1\n"
        "movb $0x1, %%al\n"
        "outb %%al, $0x21\n"
        "outb %%al, $0xa1\n"
        "movb (masterOCW1), %%al\n"
        "outb %%al, $0x21\n"
        "movb (slaveOCW1), %%al\n"
        "outb %%al, $0xa1" : : : "%eax");
    /* Setup DMA */
    __asm__ __volatile__ (
        "movb $0xe, %%al\n"
        "outb %%al, $0xf\n"
        "outb %%al, $0xde\n"
        "movb $0xc0, %%al\n"
        "outb %%al, $0xd6\n"
        ::: "eax");
    IntManager_i386_RestoreInt(flags);
    return 0;

errorFreePort80:
    ret = IOPortManager_Free(module, 0x80);
    ASSERT_ERROR(ret);
errorFreePort00:
    ret = IOPortManager_Free(module, 0x00);
    ASSERT_ERROR(ret);
errorFreePort40:
    ret = IOPortManager_Free(module, 0x40);
    ASSERT_ERROR(ret);
errorReturn:
    return 1;
}
