/*
*****************************************
* This file implements the TimerController_i386 interface
*
* Version: 0.2.1
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 20/04/2002
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2003, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include "common_def.h"

static TimerController_i386_Timer timers[TimerController_i386_NUM_TIMERS];

int TimerController_i386_Alloc(SysModManager_Module *module, unsigned int timer) {
    if (timer >= TimerController_i386_NUM_TIMERS)
        return E_BAD_VALUE;
    if (timers[timer].module)
        return E_IN_USE;
    timers[timer].module = module;
    return E_OK;
}

int TimerController_i386_Free(SysModManager_Module *module, unsigned int timer) {
    if (timer >= TimerController_i386_NUM_TIMERS)
        return E_BAD_VALUE;
    if (timers[timer].module != module)
        return E_NOT_IN_USE;
    timers[timer].module = NULL;
    return E_OK;
}

int TimerController_i386_Setup(SysModManager_Module *module, unsigned int timer, unsigned int value, unsigned int config) {
    int flags;

    if (timer >= TimerController_i386_NUM_TIMERS)
        return E_BAD_VALUE;
    if (timers[timer].module != module)
        return E_NOT_IN_USE;
    if (config >= 0xc)
        return E_BAD_VALUE;
    config |= timer;
    config |= 0x30;
    flags = IntManager_i386_NoInt();
    __asm__ __volatile__ (
        "outb %%al, $0x43\n"
        "xorl %%edx, %%edx\n"
        "movb $0x40, %%dl\n"
        "shrb $0x6, %%al\n"
        "addb %%al, %%dl\n"
        "movb %%bl, %%al\n"
        "outb %%al, %%dx\n"
        "movb %%bh, %%al\n"
        "outb %%al, %%dx"
        : : "a" (config), "b" (value) : "%edx");
    IntManager_i386_RestoreInt(flags);
    return E_OK;
}
