/*
*****************************************
* Debug macros
*
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 17/06/2005
*
* See the documentation for details.
*
* Copyright (C) 2000-2005, Luiz Henrique Shigunov
*****************************************
*/

#if !defined(__SYS_DEBUG_H)
#define __SYS_DEBUG_H
#include <errormanager.h>
#include <sysmodmanager.h>

#ifdef DEBUG
#define DEBUG_MSG(STRING) SysModManager_DebugMsg STRING
#define DEBUG_MSG_PRE(errCode) SysModManager_DebugMsg(__FILE__ ": debug point at %u: 0x%x caller: 0x%x\n", __LINE__, errCode, __builtin_return_address(0))
#else
#define DEBUG_MSG(STRING) do {;} while(0)
#define DEBUG_MSG_PRE(errCode) do {;} while(0)
#endif

#ifdef DEBUG_ASSERT_MSG
#define ASSERT_MSG(STRING) SysModManager_DebugMsg STRING
#define ASSERT_MSG_PRE(errCode) SysModManager_DebugMsg(__FILE__ ": assert fail at %u: 0x%x caller: 0x%x\n", __LINE__, errCode, __builtin_return_address(0))
#define ASSERT_ERROR(errCode) if (errCode) ASSERT_MSG_PRE(errCode)
#define ASSERT_ERROR_RET(errCode, ret) if (errCode) do { ASSERT_MSG_PRE(errCode); return ret;} while (0)
#define ASSERT_ERROR_PTR(errCode) if (IS_PTR_ERROR(errCode)) ASSERT_MSG_PRE(PTR_TO_ERROR(errCode))
#define ASSERT_ERROR_PTR_RET(errCode, ret) if (IS_PTR_ERROR(errCode)) do { ASSERT_MSG_PRE(PTR_TO_ERROR(errCode)); return ret;} while (0)
#define ASSERT(exp) if (!(exp)) do { ASSERT_MSG_PRE(0); while(1) __asm__ __volatile__ ("hlt"); } while (0)
#define ASSERT_RET(exp, ret) if (!(exp)) do { ASSERT_MSG_PRE(0); return ret;} while (0)
#define ASSERT_NO_RET(exp) if (!(exp)) do { ASSERT_MSG_PRE(0); return;} while (0)
#else
#define ASSERT_MSG(STRING) do {;} while(0)
#define ASSERT_MSG_PRE(errCode) do {;} while(0)
#define ASSERT_ERROR(errCode) do {;} while(0)
#define ASSERT_ERROR_RET(errCode, ret) do {;} while(0)
#define ASSERT_ERROR_PTR(errCode) do {;} while(0)
#define ASSERT_ERROR_PTR_RET(errCode, ret) do {;} while(0)
#define ASSERT(exp) do {;} while(0)
#define ASSERT_RET(exp, ret) do {;} while(0)
#define ASSERT_NO_RET(exp) do {;} while(0)
#endif
#endif /* __SYS_DEBUG_H */
