/*
*****************************************
* Interface FSManager
*
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 28/02/2003
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2003, Luiz Henrique Shigunov
*****************************************
*/

#if !defined(__FSMANAGER_H)
#define __FSMANAGER_H
#include "taskmanager.h"
#define FSManager__CLOSE 0x00
#define FSManager__MOUNT 0x02
#define FSManager__OPEN 0x03
#define FSManager__READ 0x05
#define FSManager__REMOVE 0x07
#define FSManager__REWINDDIR 0x09
#define FSManager__SEEK 0x0a
#define FSManager__STATUS 0x0b
#define FSManager__STATUSFD 0x01
#define FSManager__UNMOUNT 0x0c
#define FSManager__WRITE 0x0d
#define FSManager__UCLOSE 0x0e
#define FSManager__UCHANGEDIR 0x0f
#define FSManager__UCHANGEROOT 0x10
#define FSManager__UMOUNT 0x11
#define FSManager__UOPEN 0x12
#define FSManager__UREAD 0x14
#define FSManager__UREMOVE 0x16
#define FSManager__UREWINDDIR 0x17
#define FSManager__USEEK 0x18
#define FSManager__USTATUS 0x1a
#define FSManager__USTATUSFD 0x04
#define FSManager__UUNMOUNT 0x1b
#define FSManager__UWRITE 0x1c
#define FSManager_TYPE_MASK 0170000
#define FSManager_TYPE_SOCKET 0140000
#define FSManager_TYPE_LINK 0120000
#define FSManager_TYPE_FILE 0100000
#define FSManager_TYPE_BLOCK_DEV 0060000
#define FSManager_TYPE_DIR 0040000
#define FSManager_TYPE_CHAR_DEV 0020000
#define FSManager_TYPE_FIFO 0010000
#define FSManager_UID_BIT 0004000
#define FSManager_GID_BIT 0002000
#define FSManager_STICKY_BIT 0001000
#define FSManager_USER_RWX 0000700
#define FSManager_USER_READ 0000400
#define FSManager_USER_WRITE 0000200
#define FSManager_USER_EXEC 0000100
#define FSManager_GROUP_RWX 0000070
#define FSManager_GROUP_READ 0000040
#define FSManager_GROUP_WRITE 0000020
#define FSManager_GROUP_EXEC 0000010
#define FSManager_OTHER_RWX 0000007
#define FSManager_OTHER_READ 0000004
#define FSManager_OTHER_WRITE 0000002
#define FSManager_OTHER_EXEC 0000001
#define FSManager_MOUNT_RO 0x01
#define FSManager_MOUNT_PERMANENT 0x0100
#define FSManager_READ 0x01
#define FSManager_WRITE 0x02
#define FSManager_CREATE 0x04
#define FSManager_EXCLUSIVE 0x08
#define FSManager_TRUNCATE 0x10
#define FSManager_APPEND 0x20
#define FSManager_NOFOLLOW 0x40
#define FSManager_DIRECTORY 0x80
#define FSManager_O_USER_RWX 070000000
#define FSManager_O_USER_READ 040000000
#define FSManager_O_USER_WRITE 020000000
#define FSManager_O_USER_EXEC 010000000
#define FSManager_O_GROUP_RWX 007000000
#define FSManager_O_GROUP_READ 004000000
#define FSManager_O_GROUP_WRITE 002000000
#define FSManager_O_GROUP_EXEC 001000000
#define FSManager_O_OTHERS_RWX 000700000
#define FSManager_O_OTHERS_READ 000400000
#define FSManager_O_OTHERS_WRITE 000200000
#define FSManager_O_OTHERS_EXEC 000100000
#define FSManager_SEEK_START 0x01
#define FSManager_SEEK_CURRENT 0x02
#define FSManager_SEEK_END 0x04
#define FSManager_STATUS_LINK 0x01
#define FSManager_USER_BUF 0x80
#define FSManager_USER_BUF_IO 0x01
#define FSManager_FORCE_UNMOUNT 0x01
#define FSManager_UNMOUNT_PERMANENT 0x0100
#define FSMANAGER_INAME_LEN 24
#ifndef __NO_FSMANAGER_TYPES
struct _FSManager_Handle {struct _FileSystem_Handle *fd;};
#endif
typedef struct _FSManager_Handle FSManager_Handle;
typedef struct _FSManager_FS FSManager_FS;
typedef struct {unsigned short size;char name[1];} FSManager_DirEntry;
typedef struct {unsigned int mode;unsigned int nlink;unsigned int uid;unsigned int gid;int size;unsigned int blksize;unsigned int blocks;unsigned int atime;unsigned int mtime;unsigned int ctime;} FSManager_EntryStatus;
typedef struct {const char *str;unsigned int len;unsigned int hash;} FSManager_ExStr;
typedef struct FSManager__Dentry {unsigned int count;unsigned int mode;unsigned int flags;struct FSManager__Dentry *parent;FSManager_FS *fs;struct FSManager__Dentry *hashNext;struct FSManager__Dentry *hashPrior;struct FSManager__Dentry *lruNext;struct FSManager__Dentry *lruPrior;struct FSManager__Dentry *mounted;unsigned int version;TaskManager_Semaphore lock;FSManager_ExStr name;void *fsData;char iname[FSMANAGER_INAME_LEN];} FSManager_Dentry;
typedef struct {FSManager_Handle *fd;int flags;void *buf;int size;} FSManager_FDIO;
#include "filesystem.h"
/* system functions */
int FSManager_Close(FSManager_Handle*);
int FSManager_Mount(const char*,const char*,unsigned int,const char*,int,const char*);
int FSManager_Open(const char*,int,FSManager_Handle**);
int FSManager_Read(FSManager_FDIO*);
int FSManager_Remove(const char*);
int FSManager_RewindDir(FSManager_Handle*);
int FSManager_Seek(FSManager_Handle*,int,int,int*);
int FSManager_Status(const char*,FSManager_EntryStatus*,int);
int FSManager_StatusFD(FSManager_Handle*,FSManager_EntryStatus*,int);
int FSManager_Unmount(const char*,int);
int FSManager_Write(FSManager_FDIO*);
/* user functions */
int FSManager_UClose(int);
int FSManager_UChangeDir(const char*);
int FSManager_UChangeRoot(const char*);
int FSManager_UMount(const char*,const char*,unsigned int,const char*,int,const char*);
int FSManager_UOpen(const char*,int);
int FSManager_URead(int,void*,int);
int FSManager_URemove(const char*);
int FSManager_URewindDir(int);
int FSManager_USeek(int,int,int);
int FSManager_UStatus(const char*,FSManager_EntryStatus*,int);
int FSManager_UStatusFD(int,FSManager_EntryStatus*,int);
int FSManager_UUnmount(const char*,int);
int FSManager_UWrite(int,void*,int);
#endif /* __FSMANAGER_H */
