/*
*****************************************
* Interface SysModManager
*
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 17/01/2004
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2004, Luiz Henrique Shigunov
*****************************************
*/

#if !defined(__SYSMODMANAGER_H)
#define __SYSMODMANAGER_H
#define SysModManager__BROADCAST 0x00
#define SysModManager__DEBUGMSG 0x02
#define SysModManager__DELMODINFO 0x01
#define SysModManager__GETFUNCTION 0x05
#define SysModManager__GETFUNCTIONINFOINDEX 0x07
#define SysModManager__GETIMPID 0x0c
#define SysModManager__GETIMPINFO 0x0d
#define SysModManager__GETIMPINFOINDEX 0x0e
#define SysModManager__GETINTERID 0x13
#define SysModManager__GETINTERINFO 0x0f
#define SysModManager__GETINTERINFOINDEX 0x10
#define SysModManager__GETMODINFO 0x06
#define SysModManager__GETMODULEINFO 0x11
#define SysModManager__GETMODULEINFOINDEX 0x12
#define SysModManager__PUTIMPID 0x03
#define SysModManager__REGISTER 0x08
#define SysModManager__SETMODINFO 0x09
#define SysModManager__STARTMODULE 0x0a
#define SysModManager__UNREGISTER 0x0b
typedef struct _SysModManager_Module SysModManager_Module;
typedef struct _SysModManager_Implementation SysModManager_Imp;
typedef struct _SysModManager_Interface SysModManager_Interface;
typedef struct {unsigned int code;int prop;unsigned int words;} SysModManager_FunctionInfo;
typedef struct {SysModManager_Imp *imp;SysModManager_Module *module;SysModManager_Interface *interface;char *name;} SysModManager_ImpInfo;
typedef struct {SysModManager_Interface *interface;unsigned int nFunctions;char *name;} SysModManager_InterInfo;
typedef struct {SysModManager_Module *module;int prop;unsigned int memStart;unsigned int memSize;unsigned int refCounter;char *path;unsigned int phase0Function;unsigned int phase1Function;unsigned int shutdownFunction;char md5[16];} SysModManager_ModuleInfo;
#define SysModManager_USER_FUNCTION 0x01
#define SysModManager_SYS_FUNCTION 0x01
#define SysModManager_NOT_IMPLEMENTED 0x02
#define SysModManager_PERMANENT 0x01
#define SysModManager_AUTOMATIC 0x02
#define SysModManager_LOAD_AT_BOOT 0x04
void SysModManager_Broadcast(const char*,unsigned int,void*,void*);
void SysModManager_DebugMsg(const char*,...);
void *SysModManager_DelModInfo(SysModManager_Module*,SysModManager_Module*);
void *SysModManager_GetFunction(SysModManager_Imp*,unsigned int,int,int*);
int SysModManager_GetFunctionInfoIndex(SysModManager_Imp*,unsigned int,SysModManager_FunctionInfo*,void*,unsigned int);
SysModManager_Imp *SysModManager_GetImpID(const char*,const char*);
int SysModManager_GetImpInfo(SysModManager_Imp*,SysModManager_ImpInfo*,void*,unsigned int);
int SysModManager_GetImpInfoIndex(SysModManager_Interface*,unsigned int,SysModManager_ImpInfo*,void*,unsigned int);
SysModManager_Interface *SysModManager_GetInterID(const char*);
int SysModManager_GetInterInfo(SysModManager_Interface*,SysModManager_InterInfo*,void*,unsigned int);
int SysModManager_GetInterInfoIndex(unsigned int,SysModManager_InterInfo*,void*,unsigned int);
void *SysModManager_GetModInfo(SysModManager_Module*,SysModManager_Module*);
int SysModManager_GetModuleInfo(SysModManager_Module*,SysModManager_ModuleInfo*,void*,unsigned int);
int SysModManager_GetModuleInfoIndex(unsigned int,SysModManager_ModuleInfo*,void*,unsigned int);
int SysModManager_PutImpID(SysModManager_Imp*);
int SysModManager_Register(const char*,int);
int SysModManager_SetModInfo(SysModManager_Module*,SysModManager_Module*,void*,int,void**);
int SysModManager_StartModule(unsigned int);
int SysModManager_Unregister(const char*,int);
#endif /* __SYSMODMANAGER_H */
