/*
*****************************************
* Interface TaskManager
*
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 15/06/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2005, Luiz Henrique Shigunov
*****************************************
*/

#if !defined(__TASKMANAGER_H)
#define __TASKMANAGER_H
#define TaskManager__CANCELWAIT 0x00
#define TaskManager__CREATETASK 0x01
#define TaskManager__CREATETHREAD 0x02
#define TaskManager__DELTASKINFO 0x03
#define TaskManager__DELTHREADINFO 0x04
#define TaskManager__DESTROYABLE 0x2d
#define TaskManager__DESTROYABLETHREAD 0x30
#define TaskManager__DESTROYTASK 0x05
#define TaskManager__DESTROYTHREAD 0x06
#define TaskManager__GETLASTERROR 0x18
#define TaskManager__GETMSCOUNTER 0x07
#define TaskManager__GETPARENTTASKHANDLE 0x26
#define TaskManager__GETTASKHANDLEID 0x27
#define TaskManager__GETTASKHANDLE 0x08
#define TaskManager__GETTASKID 0x28
#define TaskManager__GETTASKINFO 0x09
#define TaskManager__GETTASKMMAP 0x1d
#define TaskManager__GETTHREADHANDLEID 0x29
#define TaskManager__GETTHREADHANDLE 0x0a
#define TaskManager__GETTHREADID 0x2a
#define TaskManager__GETTHREADINFO 0x0b
#define TaskManager__GETTHREADSTATUS 0x0c
#define TaskManager__LOCKREAD 0x0d
#define TaskManager__LOCKWRITE 0x0e
#define TaskManager__P 0x0f
#define TaskManager__RESUME 0x10
#define TaskManager__SCHEDULE 0x11
#define TaskManager__SETLASTERROR 0x2e
#define TaskManager__SETTASKINFO 0x12
#define TaskManager__SETTHREADINFO 0x13
#define TaskManager__SPINLOCK 0x14
#define TaskManager__SPINUNLOCK 0x15
#define TaskManager__SUSPEND 0x16
#define TaskManager__TOUSERLEVEL 0x17
#define TaskManager__UNDESTROYABLETHREAD 0x2f
#define TaskManager__UNLOCKRW 0x19
#define TaskManager__V 0x1a
#define TaskManager__WAIT 0x1b
#define TaskManager__WAITQUEUEADD 0x1e
#define TaskManager__WAITQUEUEWAKEUP 0x1f
#define TaskManager__YIELD 0x1c
#define TaskManager__UCREATETHREAD 0x20
#define TaskManager__UDESTROYTHREAD 0x21
#define TaskManager__UGETLASTERROR 0x31
#define TaskManager__UGETPARENTTASKID 0x2b
#define TaskManager__UGETTASKID 0x2c
#define TaskManager__UGETTHREADID 0x22
#define TaskManager__UP 0x23
#define TaskManager__USETLASTERROR 0x32
#define TaskManager__UTRYP 0x24
#define TaskManager__UV 0x25
/* arch specific .h */
#ifdef __i386__
#include <taskmanager_i386.h>
#endif
#ifndef __NO_TASKMANAGER_TYPES
struct _TaskManager_RWSync {void *data[4];};
struct _TaskManager_Semaphore {int count;void *data;};
struct _TaskManager_WaitQueue {void *data;};
#endif
typedef struct _TaskManager_Task TaskManager_Task;
typedef struct _TaskManager_Thread TaskManager_Thread;
typedef struct _TaskManager_RWSync TaskManager_RWSync;
typedef struct _TaskManager_Semaphore TaskManager_Semaphore;
typedef struct _TaskManager_WaitQueue TaskManager_WaitQueue;
/* you must include arch specific .h for this! */
typedef struct _TaskManager_Regs TaskManager_Regs;
#define TaskManager_WAIT_SUS 0x01
#define TaskManager_WAIT_KEEP_CPU 0x02
#define TaskManager_CREATE_SUS 0x01
#define TaskManager_CREATE_SYSTEM 0x02
#define TaskManager_LOCK_DONT_BLOCK 0x01
#define TaskManager_RESUME_KEEP_CPU 0x01
#define TaskManager_SET_INFO_KEEP_DATA 0x01
#define TaskManager_THREAD_INT 0x04
#include <sysmodmanager.h>
/* you must include arch specific .h for this! */
static __inline__ void TaskManager_AtomicAdd(unsigned int *var, unsigned int value) {
TaskManager_INLINE_ATOMICADD
}
static __inline__ void TaskManager_AtomicDec(unsigned int *var) {
TaskManager_INLINE_ATOMICDEC
}
static __inline__ void TaskManager_AtomicInc(unsigned int *var) {
TaskManager_INLINE_ATOMICINC
}
static __inline__ void TaskManager_AtomicSub(unsigned int *var, unsigned int value) {
TaskManager_INLINE_ATOMICSUB
}
int TaskManager_CancelWait(TaskManager_Thread*,int);
int TaskManager_CreateTask(void*,void*,int,TaskManager_Task**,TaskManager_Thread**);
int TaskManager_CreateThread(TaskManager_Task*,void*,void*,int,TaskManager_Thread**);
void *TaskManager_DelTaskInfo(TaskManager_Task*,SysModManager_Module*);
void *TaskManager_DelThreadInfo(TaskManager_Thread*,SysModManager_Module*);
void TaskManager_Destroyable(TaskManager_Task*);
void TaskManager_DestroyableThread(TaskManager_Thread*);
int TaskManager_DestroyTask(TaskManager_Task*);
int TaskManager_DestroyThread(TaskManager_Thread*);
long TaskManager_GetLastError(const char**,int*);
unsigned int TaskManager_GetMsCounter(void);
TaskManager_Task *TaskManager_GetParentTaskHandle(TaskManager_Task*);
TaskManager_Task *TaskManager_GetTaskHandleID(int);
TaskManager_Task *TaskManager_GetTaskHandle(TaskManager_Thread*);
int TaskManager_GetTaskID(TaskManager_Task*);
void *TaskManager_GetTaskInfo(TaskManager_Task*,SysModManager_Module*);
int TaskManager_GetTaskMMap(TaskManager_Task*);
TaskManager_Thread *TaskManager_GetThreadHandleID(TaskManager_Task*,int);
TaskManager_Thread *TaskManager_GetThreadHandle(void);
int TaskManager_GetThreadID(TaskManager_Thread*);
void *TaskManager_GetThreadInfo(TaskManager_Thread*,SysModManager_Module*);
int TaskManager_GetThreadStatus(TaskManager_Thread*);
int TaskManager_LockRead(TaskManager_RWSync*,int);
int TaskManager_LockWrite(TaskManager_RWSync*,int);
int TaskManager_P(TaskManager_Semaphore*,int,unsigned int);
int TaskManager_Resume(TaskManager_Thread*,int);
void TaskManager_Schedule(void);
long TaskManager_SetLastError(long,const char*,int);
void *TaskManager_SetTaskInfo(TaskManager_Task*,SysModManager_Module*,void*,int);
void *TaskManager_SetThreadInfo(TaskManager_Thread*,SysModManager_Module*,void*,int);
void TaskManager_SpinLock(int*);
void TaskManager_SpinUnlock(int*);
int TaskManager_Suspend(TaskManager_Thread*);
void TaskManager_ToUserLevel(TaskManager_Regs*);
void TaskManager_UndestroyableThread(void);
int TaskManager_UnlockRW(TaskManager_RWSync*,int);
int TaskManager_V(TaskManager_Semaphore*,int);
int TaskManager_Wait(TaskManager_Thread*,unsigned int);
int TaskManager_WaitQueueAdd(TaskManager_WaitQueue*,TaskManager_Semaphore*,int);
int TaskManager_WaitQueueWakeup(TaskManager_WaitQueue*);
void TaskManager_Yield(void);
int TaskManager_UCreateThread(void*,void*,int);
int TaskManager_UDestroyThread(int,void*,unsigned int);
long TaskManager_UGetLastError(char*,int*);
int TaskManager_UGetParentTaskID(void);
int TaskManager_UGetTaskID(void);
int TaskManager_UGetThreadID(void);
int TaskManager_UP(TaskManager_Semaphore*);
long TaskManager_USetLastError(long,const char*,int);
int TaskManager_UTryP(TaskManager_Semaphore*);
int TaskManager_UV(TaskManager_Semaphore*);
#endif  /* __TASKMANAGER_H */
