/*
*****************************************
* This file implements the KbdLayout interface
*
* Version: 0.2.4
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 01/07/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2005, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include <stddef.h>
#include "kbdlayout.h"
#include "keyboard.h"

/* deadkeys table */
KbdLayout_VKDeadKey deadkeys[] = {
    {{{Keyboard_VK_PUNCT5, 0x00, Keyboard_VK_A, 0x00}}, 0xe1,0x00},
    {{{Keyboard_VK_PUNCT5, 0x00, Keyboard_VK_E, 0x00}}, 0xe9,0x00},
    {{{Keyboard_VK_PUNCT5, 0x00, Keyboard_VK_I, 0x00}}, 0xed,0x00},
    {{{Keyboard_VK_PUNCT5, 0x00, Keyboard_VK_O, 0x00}}, 0xf3,0x00},
    {{{Keyboard_VK_PUNCT5, 0x00, Keyboard_VK_U, 0x00}}, 0xfa,0x00},
    {{{Keyboard_VK_PUNCT5, 0x00, Keyboard_VK_Y, 0x00}}, 0xfd,0x00},
    {{{Keyboard_VK_PUNCT5, 0x00, Keyboard_VK_A, 0x01}}, 0xc1,0x00},
    {{{Keyboard_VK_PUNCT5, 0x00, Keyboard_VK_E, 0x01}}, 0xc9,0x00},
    {{{Keyboard_VK_PUNCT5, 0x00, Keyboard_VK_I, 0x01}}, 0xcd,0x00},
    {{{Keyboard_VK_PUNCT5, 0x00, Keyboard_VK_O, 0x01}}, 0xd3,0x00},
    {{{Keyboard_VK_PUNCT5, 0x00, Keyboard_VK_U, 0x01}}, 0xda,0x00},
    {{{Keyboard_VK_PUNCT5, 0x00, Keyboard_VK_Y, 0x01}}, 0xdd,0x00},
    {{{Keyboard_VK_PUNCT5, 0x00, Keyboard_VK_C, 0x00}}, 0xe7,0x00},
    {{{Keyboard_VK_PUNCT5, 0x00, Keyboard_VK_C, 0x01}}, 0xc7,0x00},
    {{{Keyboard_VK_PUNCT5, 0x00, Keyboard_VK_SPACE, 0x00}}, '\'',0x00},
    {{{Keyboard_VK_PUNCT5, 0x00, Keyboard_VK_PUNCT5, 0x00}}, '\'',0x00},
    {{{Keyboard_VK_PUNCT5, 0x01, Keyboard_VK_A, 0x00}}, 0xe4,0x00},
    {{{Keyboard_VK_PUNCT5, 0x01, Keyboard_VK_E, 0x00}}, 0xeb,0x00},
    {{{Keyboard_VK_PUNCT5, 0x01, Keyboard_VK_I, 0x00}}, 0xef,0x00},
    {{{Keyboard_VK_PUNCT5, 0x01, Keyboard_VK_O, 0x00}}, 0xf6,0x00},
    {{{Keyboard_VK_PUNCT5, 0x01, Keyboard_VK_U, 0x00}}, 0xfc,0x00},
    {{{Keyboard_VK_PUNCT5, 0x01, Keyboard_VK_Y, 0x00}}, 0xff,0x00},
    {{{Keyboard_VK_PUNCT5, 0x01, Keyboard_VK_A, 0x01}}, 0xc4,0x00},
    {{{Keyboard_VK_PUNCT5, 0x01, Keyboard_VK_E, 0x01}}, 0xcb,0x00},
    {{{Keyboard_VK_PUNCT5, 0x01, Keyboard_VK_I, 0x01}}, 0xcf,0x00},
    {{{Keyboard_VK_PUNCT5, 0x01, Keyboard_VK_O, 0x01}}, 0xd6,0x00},
    {{{Keyboard_VK_PUNCT5, 0x01, Keyboard_VK_U, 0x01}}, 0xdc,0x00},
    {{{Keyboard_VK_PUNCT5, 0x01, Keyboard_VK_SPACE, 0x00}}, '\"',0x00},
    {{{Keyboard_VK_PUNCT5, 0x01, Keyboard_VK_PUNCT5, 0x01}}, '\"',0x00},
    {{{Keyboard_VK_6, 0x01, Keyboard_VK_A, 0x00}}, 0xe2,0x00},
    {{{Keyboard_VK_6, 0x01, Keyboard_VK_E, 0x00}}, 0xea,0x00},
    {{{Keyboard_VK_6, 0x01, Keyboard_VK_I, 0x00}}, 0xee,0x00},
    {{{Keyboard_VK_6, 0x01, Keyboard_VK_O, 0x00}}, 0xf4,0x00},
    {{{Keyboard_VK_6, 0x01, Keyboard_VK_U, 0x00}}, 0xfb,0x00},
    {{{Keyboard_VK_6, 0x01, Keyboard_VK_A, 0x01}}, 0xc2,0x00},
    {{{Keyboard_VK_6, 0x01, Keyboard_VK_E, 0x01}}, 0xca,0x00},
    {{{Keyboard_VK_6, 0x01, Keyboard_VK_I, 0x01}}, 0xce,0x00},
    {{{Keyboard_VK_6, 0x01, Keyboard_VK_O, 0x01}}, 0xd4,0x00},
    {{{Keyboard_VK_6, 0x01, Keyboard_VK_U, 0x01}}, 0xdb,0x00},
    {{{Keyboard_VK_6, 0x01, Keyboard_VK_SPACE, 0x00}}, '^',0x00},
    {{{Keyboard_VK_6, 0x01, Keyboard_VK_6, 0x01}}, '^',0x00},
    {{{Keyboard_VK_PUNCT6, 0x00, Keyboard_VK_A, 0x00}}, 0xe0,0x00},
    {{{Keyboard_VK_PUNCT6, 0x00, Keyboard_VK_E, 0x00}}, 0xe8,0x00},
    {{{Keyboard_VK_PUNCT6, 0x00, Keyboard_VK_I, 0x00}}, 0xec,0x00},
    {{{Keyboard_VK_PUNCT6, 0x00, Keyboard_VK_O, 0x00}}, 0xf2,0x00},
    {{{Keyboard_VK_PUNCT6, 0x00, Keyboard_VK_U, 0x00}}, 0xf9,0x00},
    {{{Keyboard_VK_PUNCT6, 0x00, Keyboard_VK_A, 0x01}}, 0xc0,0x00},
    {{{Keyboard_VK_PUNCT6, 0x00, Keyboard_VK_E, 0x01}}, 0xc8,0x00},
    {{{Keyboard_VK_PUNCT6, 0x00, Keyboard_VK_I, 0x01}}, 0xcc,0x00},
    {{{Keyboard_VK_PUNCT6, 0x00, Keyboard_VK_O, 0x01}}, 0xd2,0x00},
    {{{Keyboard_VK_PUNCT6, 0x00, Keyboard_VK_U, 0x01}}, 0xd9,0x00},
    {{{Keyboard_VK_PUNCT6, 0x00, Keyboard_VK_SPACE, 0x00}}, '`',0x00},
    {{{Keyboard_VK_PUNCT6, 0x00, Keyboard_VK_PUNCT6, 0x00}}, '`',0x00},
    {{{Keyboard_VK_PUNCT6, 0x01, Keyboard_VK_A, 0x00}}, 0xe3,0x00},
    {{{Keyboard_VK_PUNCT6, 0x01, Keyboard_VK_O, 0x00}}, 0xf5,0x00},
    {{{Keyboard_VK_PUNCT6, 0x01, Keyboard_VK_N, 0x00}}, 0xf1,0x00},
    {{{Keyboard_VK_PUNCT6, 0x01, Keyboard_VK_A, 0x01}}, 0xc3,0x00},
    {{{Keyboard_VK_PUNCT6, 0x01, Keyboard_VK_O, 0x01}}, 0xd5,0x00},
    {{{Keyboard_VK_PUNCT6, 0x01, Keyboard_VK_N, 0x01}}, 0xd1,0x00},
    {{{Keyboard_VK_PUNCT6, 0x01, Keyboard_VK_SPACE, 0x00}}, '~',0x00},
    {{{Keyboard_VK_PUNCT6, 0x01, Keyboard_VK_PUNCT6, 0x01}}, '~',0x00},
    {{{0x00, 0x00, 0x00, 0x00}}, 0x00,0x00},
};

/* virtual key to character with shift */
unsigned short shiftTable[256] = {
    0x00, ')', '!', '@', '#', '$', '%', 0xf000, '&', '*',          /*  0 -  9 */
    '(', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I',           /* 10 - 19 */
    'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S',           /* 20 - 29 */
    'T', 'U', 'V', 'W', 'X', 'Y', 'Z', ' ', 0x00, 0x00,         /* 30 - 39 */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 40 - 49 */
    0x00, 0x00, 0x00, 0x00, '_', '+', '{', '}', ':', 0xf000,      /* 50 - 59 */
    0xf000, '|', '<', '>', '?', '|', 0x00, 0x00, 0x00, 0x00,       /* 60 - 69 */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x1b, '\b', '\t', '\r', '\r', /* 70 - 79 */
};

/* main keyboard layout struct */
KbdLayout_MapTables mapTable = {
    /* scancode to virtual key */
    {0x00, Keyboard_VK_ESC, Keyboard_VK_1, Keyboard_VK_2, Keyboard_VK_3, Keyboard_VK_4, Keyboard_VK_5, Keyboard_VK_6, Keyboard_VK_7, Keyboard_VK_8,                                                                                /*  0 -  9 */
     Keyboard_VK_9, Keyboard_VK_0, Keyboard_VK_PUNCT0, Keyboard_VK_PUNCT1, Keyboard_VK_BACK, Keyboard_VK_TAB, Keyboard_VK_Q, Keyboard_VK_W, Keyboard_VK_E, Keyboard_VK_R,                                                          /* 10 - 19 */
     Keyboard_VK_T, Keyboard_VK_Y, Keyboard_VK_U, Keyboard_VK_I, Keyboard_VK_O, Keyboard_VK_P, Keyboard_VK_PUNCT2, Keyboard_VK_PUNCT3, Keyboard_VK_RETURN, Keyboard_VK_LCTRL,                                                      /* 20 - 29 */
     Keyboard_VK_A, Keyboard_VK_S, Keyboard_VK_D, Keyboard_VK_F, Keyboard_VK_G, Keyboard_VK_H, Keyboard_VK_J, Keyboard_VK_K, Keyboard_VK_L, Keyboard_VK_PUNCT4,                                                                    /* 30 - 39 */
     Keyboard_VK_PUNCT5, Keyboard_VK_PUNCT6, Keyboard_VK_LSHIFT, Keyboard_VK_PUNCT7, Keyboard_VK_Z, Keyboard_VK_X, Keyboard_VK_C, Keyboard_VK_V, Keyboard_VK_B, Keyboard_VK_N,                                                     /* 40 - 49 */
     Keyboard_VK_M, Keyboard_VK_PUNCT8, Keyboard_VK_PUNCT9, Keyboard_VK_PUNCT10, Keyboard_VK_RSHIFT, Keyboard_VK_KEYPAD_MUL, Keyboard_VK_LALT, Keyboard_VK_SPACE, Keyboard_VK_CAPSLOCK, Keyboard_VK_F1,                            /* 50 - 59 */
     Keyboard_VK_F2, Keyboard_VK_F3, Keyboard_VK_F4, Keyboard_VK_F5, Keyboard_VK_F6, Keyboard_VK_F7, Keyboard_VK_F8, Keyboard_VK_F9, Keyboard_VK_F10, Keyboard_VK_NUMLOCK,                                                         /* 60 - 69 */
     Keyboard_VK_SCROLLLOCK, Keyboard_VK_KEYPAD7, Keyboard_VK_KEYPAD8, Keyboard_VK_KEYPAD9, Keyboard_VK_KEYPAD_MINUS, Keyboard_VK_KEYPAD4, Keyboard_VK_KEYPAD5, Keyboard_VK_KEYPAD6, Keyboard_VK_KEYPAD_PLUS, Keyboard_VK_KEYPAD1, /* 70 - 79 */
     Keyboard_VK_KEYPAD2, Keyboard_VK_KEYPAD3, Keyboard_VK_KEYPAD0, Keyboard_VK_KEYPAD_DOT, Keyboard_VK_SYSRQ, 0x00, Keyboard_VK_PUNCT11, Keyboard_VK_F11, Keyboard_VK_F12, 0x00,                                                  /* 80 - 89 */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
     /* scancode with prefix 0xe0 to virtual key */
    {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,   /*  0 -  9 */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,   /* 10 - 19 */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, Keyboard_VK_KEYBAD_RETURN, Keyboard_VK_RCTRL,                                                      /* 20 - 29 */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,                                                                                        /* 30 - 39 */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,                                                                                        /* 40 - 49 */
     0x00, 0x00, 0x00, Keyboard_VK_KEYPAD_DIV, 0x00, Keyboard_VK_SYSRQ, Keyboard_VK_RALT, 0x00, 0x00, 0x00,                                             /* 50 - 59 */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,                                                                                        /* 60 - 69 */
     Keyboard_VK_PAUSE, Keyboard_VK_HOME, Keyboard_VK_UP, Keyboard_VK_PAGEUP, 0x00, Keyboard_VK_LEFT, 0x00, Keyboard_VK_RIGHT, 0x00, Keyboard_VK_END,   /* 70 - 79 */
     Keyboard_VK_DOWN, Keyboard_VK_PAGEDOWN, Keyboard_VK_INSERT, Keyboard_VK_DELETE, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,                                /* 80 - 89 */
     0x00, Keyboard_VK_LWIN, Keyboard_VK_RWIN, Keyboard_VK_MENU, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,                                                    /* 90 - 99 */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,   /* 100 - 109 */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,   /* 110 - 119 */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},              /* 120 - 127 */
     /* virtual key to characters - no shift */
     {{0x00, 0x00}, {'0', 0x00}, {'1', 0x00}, {'2', 0x00}, {'3', 0x00}, {'4', 0x00}, {'5', 0x00}, {'6', 0x00}, {'7', 0x00}, {'8', 0x00}, /*  0 -  9 */
      {'9', 0x00}, {'a', 0x01}, {'b', 0x01}, {'c', 0x01}, {'d', 0x01}, {'e', 0x01}, {'f', 0x01}, {'g', 0x01}, {'h', 0x01}, {'i', 0x01},  /* 10 - 19 */
      {'j', 0x01}, {'k', 0x01}, {'l', 0x01}, {'m', 0x01}, {'n', 0x01}, {'o', 0x01}, {'p', 0x01}, {'q', 0x01}, {'r', 0x01}, {'s', 0x01},  /* 20 - 29 */
      {'t', 0x01}, {'u', 0x01}, {'v', 0x01}, {'w', 0x01}, {'x', 0x01}, {'y', 0x01}, {'z', 0x01}, {' ', 0x00}, {'0', 0x00}, {'1', 0x00},  /* 30 - 39 */
      {'2', 0x00}, {'3', 0x00}, {'4', 0x00}, {'5', 0x00}, {'6', 0x00}, {'7', 0x00}, {'8', 0x00}, {'9', 0x00}, {'/', 0x00}, {'*', 0x00},  /* 40 - 49 */
      {'-', 0x00}, {'+', 0x00}, {'.', 0x00}, {',', 0x00}, {'-', 0x00}, {'=', 0x00}, {'[', 0x00}, {']', 0x00}, {';', 0x00}, {0xf000, 0x00},  /* 50 - 59 */
      {0xf000, 0x00}, {'\\', 0x00}, {',', 0x00}, {'.', 0x00}, {'/', 0x00}, {'\\', 0x00}, {0x00, 0x00}, {0x00, 0x00}, {0x00, 0x00}, {0x00, 0x00},  /* 60 - 69 */
      {0x00, 0x00}, {0x00, 0x00}, {0x00, 0x00}, {0x00, 0x00}, {0x00, 0x00}, {0x1b, 0x00}, {'\b', 0x00}, {'\t', 0x00}, {'\r', 0x00}, {'\r', 0x00},  /* 70 - 79 */
     },
     /* virtual key to character with shift */
     {shiftTable, NULL, NULL, NULL, NULL, NULL, NULL, NULL},
     /* deadkeys table */
     deadkeys
};

KbdLayout_MapTables *GetMapTables(void) {
    return &mapTable;
}
