/*
*****************************************
* This file implements the MemMan interface
*
* Version: 0.21
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 16/09/2002
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2002, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* Functions taked from Linux
*	NO Copyright (C) 1991, 1992 Linus Torvalds,
*	consider these trivial functions to be PD.
*****************************************
*/

#include <stddef.h>

int MemMan_Compare(const void *s1, const void *s2, unsigned int count) {
	register int __res;
	__asm__ __volatile__(
		"cld\n\t"
		"repe\n\t"
		"cmpsb\n\t"
		"je 1f\n\t"
		"sbbl %0,%0\n\t"
		"orb $1,%b0\n"
		"1:"
		:"=a" (__res)
		:"0" (0), "S" (s1), "D" (s2),"c" (count));
	return __res;
}

void MemMan_Copy(void *dest, const void *src, unsigned int count) {
	int d0, d1, d2;
	__asm__ __volatile__(
		"cld\n\t"
		"rep ; movsl\n\t"
		"testb $2,%b4\n\t"
		"je 1f\n\t"
		"movsw\n"
		"1:\ttestb $1,%b4\n\t"
		"je 2f\n\t"
		"movsb\n"
		"2:"
		: "=&c" (d0), "=&D" (d1), "=&S" (d2)
		:"0" (count/4), "q" (count),"1" ((long) dest),"2" ((long) src)
		: "memory");
}

void *MemMan_FindByte(const void *mem, int c, unsigned int count) {
	int d0;
	register void * __res;
	if (!count)
		return NULL;
	__asm__ __volatile__(
		"cld\n\t"
		"repne\n\t"
		"scasb\n\t"
		"je 1f\n\t"
		"movl $1,%0\n"
		"1:\tdecl %0"
		:"=D" (__res), "=&c" (d0) : "a" (c),"0" (mem),"1" (count));
	return __res;
}

void MemMan_Move(void *dest, const void *src, unsigned int count) {
	int d0, d1, d2;
	if (dest<src)
	__asm__ __volatile__(
		"cld\n\t"
		"rep\n\t"
		"movsb"
		: "=&c" (d0), "=&S" (d1), "=&D" (d2)
		:"0" (count),"1" (src),"2" (dest)
		: "memory");
	else
	__asm__ __volatile__(
		"std\n\t"
		"rep\n\t"
		"movsb\n\t"
		"cld"
		: "=&c" (d0), "=&S" (d1), "=&D" (d2)
		:"0" (count),
		 "1" (count-1+(const char *)src),
		 "2" (count-1+(char *)dest)
		:"memory");
}

void MemMan_Set(void *mem, unsigned int c, unsigned int count) {
    int d0, d1;
    __asm__ __volatile__(
        "cld\n\t"
        "rep\n\t"
        "stosb"
        : "=&c" (d0), "=&D" (d1)
        :"a" (c),"1" (mem),"0" (count)
        :"memory");
}
