/*
*****************************************
* This file implements the StringMan interface
*
* Version: 0.2.2
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 10/05/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2005, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* Some functions taked from Linux
*	NO Copyright (C) 1991, 1992 Linus Torvalds,
*	consider these trivial functions to be PD.
*
* Some functions taked from Linux
*  linux/lib/vsprintf.c
*
*  Copyright (C) 1991, 1992  Linus Torvalds
* vsprintf.c -- Lars Wirzenius & Linus Torvalds.
*
* Wirzenius wrote this portably, Torvalds fucked it up :-)
*
* Fri Jul 13 2001 Crutcher Dunnavant <crutcher+kernel@datastacks.com>
* - changed to provide snprintf and vsnprintf functions
*
*****************************************
*/

#include <stdarg.h>
#include <stddef.h>

#define _U      0x01    /* upper */
#define _L      0x02    /* lower */
#define _D      0x04    /* digit */
#define _C      0x08    /* cntrl */
#define _P      0x10    /* punct */
#define _S      0x20    /* white space (space/lf/tab) */
#define _X      0x40    /* hex digit */
#define _SP     0x80    /* hard space (0x20) */

unsigned char _ctype[] = {
_C,_C,_C,_C,_C,_C,_C,_C,                        /* 0-7 */
_C,_C|_S,_C|_S,_C|_S,_C|_S,_C|_S,_C,_C,         /* 8-15 */
_C,_C,_C,_C,_C,_C,_C,_C,                        /* 16-23 */
_C,_C,_C,_C,_C,_C,_C,_C,                        /* 24-31 */
_S|_SP,_P,_P,_P,_P,_P,_P,_P,                    /* 32-39 */
_P,_P,_P,_P,_P,_P,_P,_P,                        /* 40-47 */
_D,_D,_D,_D,_D,_D,_D,_D,                        /* 48-55 */
_D,_D,_P,_P,_P,_P,_P,_P,                        /* 56-63 */
_P,_U|_X,_U|_X,_U|_X,_U|_X,_U|_X,_U|_X,_U,      /* 64-71 */
_U,_U,_U,_U,_U,_U,_U,_U,                        /* 72-79 */
_U,_U,_U,_U,_U,_U,_U,_U,                        /* 80-87 */
_U,_U,_U,_P,_P,_P,_P,_P,                        /* 88-95 */
_P,_L|_X,_L|_X,_L|_X,_L|_X,_L|_X,_L|_X,_L,      /* 96-103 */
_L,_L,_L,_L,_L,_L,_L,_L,                        /* 104-111 */
_L,_L,_L,_L,_L,_L,_L,_L,                        /* 112-119 */
_L,_L,_L,_P,_P,_P,_P,_C,                        /* 120-127 */
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,                /* 128-143 */
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,                /* 144-159 */
_S|_SP,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,   /* 160-175 */
_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,       /* 176-191 */
_U,_U,_U,_U,_U,_U,_U,_U,_U,_U,_U,_U,_U,_U,_U,_U,       /* 192-207 */
_U,_U,_U,_U,_U,_U,_U,_P,_U,_U,_U,_U,_U,_U,_U,_L,       /* 208-223 */
_L,_L,_L,_L,_L,_L,_L,_L,_L,_L,_L,_L,_L,_L,_L,_L,       /* 224-239 */
_L,_L,_L,_L,_L,_L,_L,_P,_L,_L,_L,_L,_L,_L,_L,_L};      /* 240-255 */

#define __ismask(x) (_ctype[(int)(unsigned char)(x)])

#define isalnum(c)      ((__ismask(c)&(_U|_L|_D)) != 0)
#define isalpha(c)      ((__ismask(c)&(_U|_L)) != 0)
#define iscntrl(c)      ((__ismask(c)&(_C)) != 0)
#define isdigit(c)      ((__ismask(c)&(_D)) != 0)
#define isgraph(c)      ((__ismask(c)&(_P|_U|_L|_D)) != 0)
#define islower(c)      ((__ismask(c)&(_L)) != 0)
#define isprint(c)      ((__ismask(c)&(_P|_U|_L|_D|_SP)) != 0)
#define ispunct(c)      ((__ismask(c)&(_P)) != 0)
#define isspace(c)      ((__ismask(c)&(_S)) != 0)
#define isupper(c)      ((__ismask(c)&(_U)) != 0)
#define isxdigit(c)     ((__ismask(c)&(_D|_X)) != 0)

#define isascii(c) (((unsigned char)(c))<=0x7f)
#define toascii(c) (((unsigned char)(c))&0x7f)

static inline unsigned char __tolower(unsigned char c)
{
        if (isupper(c))
                c -= 'A'-'a';
        return c;
}

static inline unsigned char __toupper(unsigned char c)
{
        if (islower(c))
                c -= 'a'-'A';
        return c;
}

#define tolower(c) __tolower(c)
#define toupper(c) __toupper(c)

#define do_div(n,base) ({ \
        unsigned long __upper, __low, __high, __mod; \
        asm("":"=a" (__low), "=d" (__high):"A" (n)); \
        __upper = __high; \
        if (__high) { \
                __upper = __high % (base); \
                __high = __high / (base); \
        } \
        asm("divl %2":"=a" (__low), "=d" (__mod):"rm" (base), "0" (__low), "1" (__upper)); \
        asm("":"=A" (n):"a" (__low),"d" (__high)); \
        __mod; \
})

long StringMan_NPrintFArgs(char*,unsigned long,const char*,va_list);

long StringMan_Compare(const char *s1, const char *s2) {
	long d0, d1;
	register long __res;
	__asm__ __volatile__(
		"cld\n"
		"1:\tlodsb\n\t"
		"scasb\n\t"
		"jne 2f\n\t"
		"testb %%al,%%al\n\t"
		"jne 1b\n\t"
		"xorl %%eax,%%eax\n\t"
		"jmp 3f\n"
		"2:\tsbbl %%eax,%%eax\n\t"
		"orb $1,%%al\n"
		"3:"
		:"=a" (__res), "=&S" (d0), "=&D" (d1)
		:"1" (s1),"2" (s2));
	return __res;
}

void StringMan_Concatenate(char *dest, const char *src) {
	long d0, d1, d2, d3;
	__asm__ __volatile__(
		"cld\n\t"
		"repne\n\t"
		"scasb\n\t"
		"decl %1\n"
		"1:\tlodsb\n\t"
		"stosb\n\t"
		"testb %%al,%%al\n\t"
		"jne 1b"
		: "=&S" (d0), "=&D" (d1), "=&a" (d2), "=&c" (d3)
		: "0" (src), "1" (dest), "2" (0), "3" (0xffffffff):"memory");
}

void StringMan_Copy(char *dest, const char *src) {
	long d0, d1, d2;
	__asm__ __volatile__(
		"cld\n"
		"1:\tlodsb\n\t"
		"stosb\n\t"
		"testb %%al,%%al\n\t"
		"jne 1b"
		: "=&S" (d0), "=&D" (d1), "=&a" (d2)
		:"0" (src),"1" (dest) : "memory");
}

char *StringMan_FindChar(const char *s, long c) {
	long d0;
	register char * __res;
	__asm__ __volatile__(
		"cld\n\t"
		"movb %%al,%%ah\n"
		"1:\tlodsb\n\t"
		"cmpb %%ah,%%al\n\t"
		"je 2f\n\t"
		"testb %%al,%%al\n\t"
		"jne 1b\n\t"
		"movl $1,%1\n"
		"2:\tmovl %1,%0\n\t"
		"decl %0"
		:"=a" (__res), "=&S" (d0) : "1" (s),"0" (c));
	return __res;
}

char *StringMan_FindLastChar(const char *s, long c) {
	long d0, d1;
	register char * __res;
	__asm__ __volatile__(
		"cld\n\t"
		"movb %%al,%%ah\n"
		"1:\tlodsb\n\t"
		"cmpb %%ah,%%al\n\t"
		"jne 2f\n\t"
		"leal -1(%%esi),%0\n"
		"2:\ttestb %%al,%%al\n\t"
		"jne 1b"
		:"=g" (__res), "=&S" (d0), "=&a" (d1) :"0" (0),"1" (s),"2" (c));
	return __res;
}

unsigned long StringMan_Length(const char *s) {
	long d0;
	register long __res;
	__asm__ __volatile__(
		"cld\n\t"
		"repne\n\t"
		"scasb\n\t"
		"notl %0\n\t"
		"decl %0"
		:"=c" (__res), "=&D" (d0) :"1" (s),"a" (0), "0" (0xffffffff));
	return __res;
}

long StringMan_NCompare(const char *s1, const char *s2, unsigned long count) {
	register long __res;
	long d0, d1, d2;
	__asm__ __volatile__(
		"cld\n"
		"1:\tdecl %3\n\t"
		"js 2f\n\t"
		"lodsb\n\t"
		"scasb\n\t"
		"jne 3f\n\t"
		"testb %%al,%%al\n\t"
		"jne 1b\n"
		"2:\txorl %%eax,%%eax\n\t"
		"jmp 4f\n"
		"3:\tsbbl %%eax,%%eax\n\t"
		"orb $1,%%al\n"
		"4:"
    :"=a" (__res), "=&S" (d0), "=&D" (d1), "=&c" (d2)
    :"1" (s1),"2" (s2),"3" (count));
	return __res;
}

void StringMan_NConcatenate(char *dest, const char *src, unsigned long count) {
	long d0, d1, d2, d3;
	__asm__ __volatile__(
		"cld\n\t"
		"repne\n\t"
		"scasb\n\t"
		"decl %1\n\t"
		"movl %8,%3\n"
		"1:\tdecl %3\n\t"
		"js 2f\n\t"
		"lodsb\n\t"
		"stosb\n\t"
		"testb %%al,%%al\n\t"
		"jne 1b\n"
		"2:\txorl %2,%2\n\t"
		"stosb"
		: "=&S" (d0), "=&D" (d1), "=&a" (d2), "=&c" (d3)
		: "0" (src),"1" (dest),"2" (0),"3" (0xffffffff), "g" (count)
		: "memory");
}

long StringMan_NCopy(char *dest, const char *src, unsigned long count) {
    long d0, d1, d2, d3;
    __asm__ __volatile__(
        "cld\n"
        "testl %2,%2\n\t"
        "je 4f\n\t"
        "decl %2\n\t"
        "1:\tdecl %2\n\t"
        "js 2f\n\t"
        "lodsb\n\t"
        "stosb\n\t"
        "testb %%al,%%al\n\t"
        "jne 1b\n\t"
        "decl %0\n\t"
        "2:\tlodsb\n\t"
        "testb %%al,%%al\n\t"
        "jne 3f\n\t"
        "xorl %2,%2\n\t"
        "3:\txorl %%eax,%%eax\n\t"
        "stosb\n"
        "4:"
        : "=&S" (d0), "=&D" (d1), "=&c" (d2), "=&a" (d3)
        :"0" (src),"1" (dest),"2" (count) : "memory");
    return d2;
}

unsigned long StringMan_NLength(const char *s, unsigned long count) {
    long d0;
    register long __res;
    __asm__ __volatile__(
        "movl %2,%0\n\t"
        "jmp 2f\n"
        "1:\tcmpb $0,(%0)\n\t"
        "je 3f\n\t"
        "incl %0\n"
        "2:\tdecl %1\n\t"
        "cmpl $-1,%1\n\t"
        "jne 1b\n"
        "3:\tsubl %2,%0"
        :"=a" (__res), "=&d" (d0)
        :"c" (s),"1" (count));
    return __res;
}

/**
 * simple_strtoul - convert a string to an unsigned long
 * @cp: The start of the string
 * @endp: A pointer to the end of the parsed string will be placed here
 * @base: The number base to use
 */
static unsigned long simple_strtoul(const char *cp,char **endp,unsigned long base)
{
    unsigned long result = 0,value;

    if (!base) {
        base = 10;
        if (*cp == '0') {
            base = 8;
            cp++;
            if ((*cp == 'x') && isxdigit(cp[1])) {
                cp++;
                base = 16;
            }
        }
    }
    while (isxdigit(*cp) && (value = isdigit(*cp) ? *cp-'0' : toupper(*cp)-'A'+10) < base) {
        result = result*base + value;
        cp++;
    }
    if (endp)
        *endp = (char *)cp;
    return result;
}

/**
 * simple_strtol - convert a string to a signed long
 * @cp: The start of the string
 * @endp: A pointer to the end of the parsed string will be placed here
 * @base: The number base to use
 */
static long simple_strtol(const char *cp,char **endp,unsigned long base)
{
    if(*cp=='-')
        return -simple_strtoul(cp+1,endp,base);
    return simple_strtoul(cp,endp,base);
}

/**
 * simple_strtoull - convert a string to an unsigned long long
 * @cp: The start of the string
 * @endp: A pointer to the end of the parsed string will be placed here
 * @base: The number base to use
 */
static unsigned long long simple_strtoull(const char *cp,char **endp,unsigned long base)
{
    unsigned long long result = 0,value;

    if (!base) {
        base = 10;
        if (*cp == '0') {
            base = 8;
            cp++;
            if ((*cp == 'x') && isxdigit(cp[1])) {
                cp++;
                base = 16;
            }
        }
    }
    while (isxdigit(*cp) && (value = isdigit(*cp) ? *cp-'0' : (islower(*cp)
           ? toupper(*cp) : *cp)-'A'+10) < base) {
        result = result*base + value;
        cp++;
    }
    if (endp)
        *endp = (char *)cp;
    return result;
}

/**
 * simple_strtoll - convert a string to a signed long long
 * @cp: The start of the string
 * @endp: A pointer to the end of the parsed string will be placed here
 * @base: The number base to use
 */
static long long simple_strtoll(const char *cp,char **endp,unsigned long base)
{
    if(*cp=='-')
        return -simple_strtoull(cp+1,endp,base);
    return simple_strtoull(cp,endp,base);
}

static long skip_atoi(const char **s)
{
    long i=0;

    while (isdigit(**s))
        i = i*10 + *((*s)++) - '0';
    return i;
}

#define ZEROPAD 1               /* pad with zero */
#define SIGN    2               /* unsigned/signed long */
#define PLUS    4               /* show plus */
#define SPACE   8               /* space if plus */
#define LEFT    16              /* left justified */
#define SPECIAL 32              /* 0x */
#define LARGE   64              /* use 'ABCDEF' instead of 'abcdef' */

static char * number(char * buf, char * end, long long num, long base, long size, long precision, long type)
{
    char c,sign,tmp[66];
    const char *digits;
    const char small_digits[] = "0123456789abcdefghijklmnopqrstuvwxyz";
    const char large_digits[] = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
    int i;

    digits = (type & LARGE) ? large_digits : small_digits;
    if (type & LEFT)
        type &= ~ZEROPAD;
    if (base < 2 || base > 36)
        return 0;
    c = (type & ZEROPAD) ? '0' : ' ';
    sign = 0;
    if (type & SIGN) {
        if (num < 0) {
            sign = '-';
            num = -num;
            size--;
        } else if (type & PLUS) {
            sign = '+';
            size--;
        } else if (type & SPACE) {
            sign = ' ';
            size--;
        }
    }
    if (type & SPECIAL) {
        if (base == 16)
            size -= 2;
        else if (base == 8)
            size--;
    }
    i = 0;
    if (num == 0)
        tmp[i++]='0';
    else while (num != 0)
        tmp[i++] = digits[do_div(num,base)];
    if (i > precision)
        precision = i;
    size -= precision;
    if (!(type&(ZEROPAD+LEFT))) {
        while(size-->0) {
            if (buf <= end)
                *buf = ' ';
            ++buf;
        }
    }
    if (sign) {
        if (buf <= end)
            *buf = sign;
        ++buf;
    }
    if (type & SPECIAL) {
        if (base==8) {
            if (buf <= end)
                *buf = '0';
            ++buf;
        } else if (base==16) {
            if (buf <= end)
                *buf = '0';
            ++buf;
            if (buf <= end)
                *buf = digits[33];
            ++buf;
        }
    }
    if (!(type & LEFT)) {
        while (size-- > 0) {
            if (buf <= end)
                *buf = c;
            ++buf;
        }
    }
    while (i < precision--) {
        if (buf <= end)
            *buf = '0';
        ++buf;
    }
    while (i-- > 0) {
        if (buf <= end)
            *buf = tmp[i];
        ++buf;
    }
    while (size-- > 0) {
        if (buf <= end)
            *buf = ' ';
        ++buf;
    }
    return buf;
}

long StringMan_NPrintF(char *buf, unsigned long size, const char *fmt, ...) {
    va_list args;
    long i;

    va_start(args, fmt);
    i = StringMan_NPrintFArgs(buf, size, fmt, args);
    va_end(args);
    return i;
}

long StringMan_NPrintFArgs(char *buf, unsigned long size, const char *fmt, va_list args) {
    int len;
    unsigned long long num;
    int i, base;
    char *str, *end, c;
    const char *s;

    int flags;              /* flags to number() */

    int field_width;        /* width of output field */
    int precision;          /* min. # of digits for integers; max
                            number of chars for from string */
    int qualifier;          /* 'h', 'l', or 'L' for integer fields */
                            /* 'z' support added 23/7/1999 S.H.    */
                            /* 'z' changed to 'Z' --davidm 1/25/99 */

    str = buf;
    end = buf + size - 1;

    if (end < buf - 1) {
        end = ((void *) -1);
        size = end - buf + 1;
    }

    for (; *fmt ; ++fmt) {
        if (*fmt != '%') {
            if (str <= end)
                *str = *fmt;
            ++str;
            continue;
        }

        /* process flags */
        flags = 0;
        repeat:
            ++fmt;          /* this also skips first '%' */
            switch (*fmt) {
                case '-': flags |= LEFT; goto repeat;
                case '+': flags |= PLUS; goto repeat;
                case ' ': flags |= SPACE; goto repeat;
                case '#': flags |= SPECIAL; goto repeat;
                case '0': flags |= ZEROPAD; goto repeat;
            }

        /* get field width */
        field_width = -1;
        if (isdigit(*fmt))
            field_width = skip_atoi(&fmt);
        else if (*fmt == '*') {
            ++fmt;
            /* it's the next argument */
            field_width = va_arg(args, int);
            if (field_width < 0) {
                field_width = -field_width;
                flags |= LEFT;
            }
        }

        /* get the precision */
        precision = -1;
        if (*fmt == '.') {
            ++fmt;
            if (isdigit(*fmt))
                precision = skip_atoi(&fmt);
            else if (*fmt == '*') {
                ++fmt;
                /* it's the next argument */
                precision = va_arg(args, int);
            }
            if (precision < 0)
                precision = 0;
        }

        /* get the conversion qualifier */
        qualifier = -1;
        if (*fmt == 'h' || *fmt == 'l' || *fmt == 'L' || *fmt =='Z') {
            qualifier = *fmt;
            ++fmt;
            if (qualifier == 'l' && *fmt == 'l') {
                qualifier = 'L';
                ++fmt;
            }
        }

        /* default base */
        base = 10;

        switch (*fmt) {
            case 'c':
                if (!(flags & LEFT)) {
                    while (--field_width > 0) {
                        if (str <= end)
                            *str = ' ';
                        ++str;
                    }
                }
                c = (unsigned char) va_arg(args, int);
                if (str <= end)
                    *str = c;
                ++str;
                while (--field_width > 0) {
                    if (str <= end)
                        *str = ' ';
                    ++str;
                }
                continue;

            case 's':
                s = va_arg(args, char *);
                if (!s)
                    s = "<NULL>";

                len = StringMan_NLength(s, precision);

                if (!(flags & LEFT)) {
                    while (len < field_width--) {
                        if (str <= end)
                            *str = ' ';
                        ++str;
                    }
                }
                for (i = 0; i < len; ++i) {
                    if (str <= end)
                        *str = *s;
                    ++str; ++s;
                }
                while (len < field_width--) {
                    if (str <= end)
                        *str = ' ';
                    ++str;
                }
                continue;

            case 'p':
                if (field_width == -1) {
                    field_width = 2*sizeof(void *);
                    flags |= ZEROPAD;
                }
                str = number(str, end,  (unsigned long) va_arg(args, void *), 16, field_width, precision, flags);
                continue;

            case 'n':
                /* FIXME:
                * What does C99 say about the overflow case here? */
                if (qualifier == 'l') {
                        long * ip = va_arg(args, long *);
                        *ip = (str - buf);
                } else if (qualifier == 'Z') {
                        size_t * ip = va_arg(args, size_t *);
                        *ip = (str - buf);
                } else {
                        int * ip = va_arg(args, int *);
                        *ip = (str - buf);
                }
                continue;

            case '%':
                if (str <= end)
                        *str = '%';
                ++str;
                continue;

                /* integer number formats - set up the flags and "break" */
            case 'o':
                base = 8;
                break;

            case 'X':
                flags |= LARGE;
            case 'x':
                base = 16;
                break;

            case 'd':
            case 'i':
                flags |= SIGN;
            case 'u':
                break;

            default:
                if (str <= end)
                    *str = '%';
                ++str;
                if (*fmt) {
                    if (str <= end)
                        *str = *fmt;
                    ++str;
                } else {
                    --fmt;
                }
                continue;
        }
        if (qualifier == 'L')
            num = va_arg(args, long long);
        else if (qualifier == 'l') {
            num = va_arg(args, unsigned long);
            if (flags & SIGN)
                num = (signed long) num;
        } else if (qualifier == 'Z') {
            num = va_arg(args, size_t);
        } else if (qualifier == 'h') {
            num = (unsigned short) va_arg(args, int);
            if (flags & SIGN)
                num = (signed short) num;
        } else {
            num = va_arg(args, unsigned int);
            if (flags & SIGN)
                num = (signed int) num;
        }
        str = number(str, end, num, base, field_width, precision, flags);
    }
    if (str <= end)
        *str = '\0';
    else if (size > 0)
        /* don't write out a null byte if the buf size is zero */
        *end = '\0';
    /* the trailing null byte doesn't count towards the total
    * ++str;
    */
    return str-buf;
}

long StringMan_PrintF(char *buf, const char *fmt, ...) {
    va_list args;
    long i;

    va_start(args, fmt);
    i = StringMan_NPrintFArgs(buf, 0xffffffff, fmt, args);
    va_end(args);
    return i;
}

#define INT_MAX 2147483647

long StringMan_ScanFArgs(const char *buf, const char *fmt, va_list args) {
    const char *str = buf;
    char *next;
    int num = 0;
    int qualifier;
    int base;
    int field_width = -1;
    int is_sign = 0;

    while(*fmt && *str) {
        /* skip any white space in format */
        /* white space in format matchs any amount of
        * white space, including none, in the input.
        */
        if (isspace(*fmt)) {
            while (isspace(*fmt))
                ++fmt;
            while (isspace(*str))
                ++str;
        }

        /* anything that is not a conversion must match exactly */
        if (*fmt != '%' && *fmt) {
            if (*fmt++ != *str++)
                break;
            continue;
        }

        if (!*fmt)
            break;
        ++fmt;

        /* skip this conversion.
        * advance both strings to next white space
        */
        if (*fmt == '*') {
            while (!isspace(*fmt) && *fmt)
                fmt++;
            while (!isspace(*str) && *str)
                str++;
            continue;
        }

        /* get field width */
        if (isdigit(*fmt))
            field_width = skip_atoi(&fmt);

        /* get conversion qualifier */
        qualifier = -1;
        if (*fmt == 'h' || *fmt == 'l' || *fmt == 'L' || *fmt == 'Z') {
            qualifier = *fmt;
            fmt++;
        }
        base = 10;
        is_sign = 0;

        if (!*fmt || !*str)
            break;

        switch(*fmt++) {
        case 'c':
        {
            char *s = (char *) va_arg(args,char*);
            if (field_width == -1)
                field_width = 1;
            do {
                *s++ = *str++;
            } while(field_width-- > 0 && *str);
            num++;
        }
        continue;
        case 's':
        {
            char *s = (char *) va_arg(args, char *);
            if(field_width == -1)
                field_width = INT_MAX;
            /* first, skip leading white space in buffer */
            while (isspace(*str))
                str++;

            /* now copy until next white space */
            while (*str && !isspace(*str) && field_width--) {
                *s++ = *str++;
            }
            *s = '\0';
            num++;
        }
        continue;
        case 'n':
            /* return number of characters read so far */
        {
            int *i = (int *)va_arg(args,int*);
            *i = str - buf;
        }
        continue;
        case 'o':
            base = 8;
            break;
        case 'x':
        case 'X':
            base = 16;
            break;
        case 'd':
        case 'i':
            is_sign = 1;
        case 'u':
            break;
        case '%':
            /* looking for '%' in str */
            if (*str++ != '%')
                return num;
            continue;
        default:
            /* invalid format; stop here */
            return num;
        }

        /* have some sort of integer conversion.
        * first, skip white space in buffer.
        */
        while (isspace(*str))
            str++;

        if (!*str || !isdigit(*str))
            break;

        switch(qualifier) {
            case 'h':
                if (is_sign) {
                    short *s = (short *) va_arg(args,short *);
                    *s = (short) simple_strtol(str,&next,base);
                } else {
                    unsigned short *s = (unsigned short *) va_arg(args, unsigned short *);
                    *s = (unsigned short) simple_strtoul(str, &next, base);
                }
                break;
            case 'l':
                if (is_sign) {
                    long *l = (long *) va_arg(args,long *);
                    *l = simple_strtol(str,&next,base);
                } else {
                    unsigned long *l = (unsigned long*) va_arg(args,unsigned long*);
                    *l = simple_strtoul(str,&next,base);
                }
                break;
            case 'L':
                if (is_sign) {
                    long long *l = (long long*) va_arg(args,long long *);
                    *l = simple_strtoll(str,&next,base);
                } else {
                    unsigned long long *l = (unsigned long long*) va_arg(args,unsigned long long*);
                    *l = simple_strtoull(str,&next,base);
                }
                break;
            case 'Z':
                {
                    size_t *s = (size_t*) va_arg(args,size_t*);
                    *s = (size_t) simple_strtoul(str,&next,base);
                }
                break;
            default:
                if (is_sign) {
                    int *i = (int *) va_arg(args, int*);
                    *i = (int) simple_strtol(str,&next,base);
                } else {
                    unsigned int *i = (unsigned int*) va_arg(args, unsigned int*);
                    *i = (unsigned int) simple_strtoul(str,&next,base);
                }
                break;
        }
        num++;

        if (!next)
            break;
        str = next;
    }
    return num;
}

long StringMan_ScanF(const char *buf, const char *fmt, ...) {
    va_list args;
    long i;

    va_start(args, fmt);
    i = StringMan_ScanFArgs(buf, fmt, args);
    va_end(args);
    return i;
}

/*
*Code taked from glibc 2.1 - http://www.gnu.org
*Copyright (C) 1991,92,94,95,96,97,98,99 Free Software Foundation, Inc.
*/

#define ULONG_MAX	4294967295UL
#define LONG_MAX	2147483647L
#define LONG_MIN	(-LONG_MAX - 1L)

long StringMan_StrToInt(const char *str, char **endstr, long base, long *error) {
	int negative, overflow;
	register unsigned long int cutoff, i;
	register unsigned int cutlim;
	register const char *s;
	register unsigned char c;
	const char *save;

	if (base < 0 || base == 1 || base > 36) {
  	if (error != NULL)
			*error = 1;
		return 0;
	}
	save = s = str;

	/* Skip white space */
	while ((*s == ' ') || (*s == '\f') || (*s == '\n') || (*s == '\r') || (*s == '\t') || (*s == '\v'))
		++s;
	if (*s == '\0')
		goto noconv;

	/* Check for a sign */
	if (*s == '-') {
		negative = 1;
		++s;
	} else if (*s == '+') {
		negative = 0;
		++s;
	} else
		negative = 0;

	/* Recognize number prefix and if BASE is zero, figure it out ourselves */
	if (*s == '0') {
		if ((base == 0 || base == 16) && (s[1] & 0xdf) == 'X') {
			s += 2;
			base = 16;
		} else if (base == 0)
			base = 8;
	} else if (base == 0)
		base = 10;

	/* Save the pointer so we can check later if anything happened */
	save = s;
	cutoff = ULONG_MAX / (unsigned long int) base;
	cutlim = ULONG_MAX % (unsigned long int) base;
	overflow = 0;
	i = 0;
	c = *s;
	for (;c != '\0'; c = *++s) {
		if (c >= '0' && c <= '9')
			c -= '0';
		else if ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z'))
			c = (c & 0xdf) - 'A' + 10;
		else
			break;
		if ((int) c >= base)
			break;
		/* Check for overflow */
		if (i > cutoff || (i == cutoff && c > cutlim))
			overflow = 1;
		else {
			i *= (unsigned long int) base;
			i += c;
		}
	}

	/* Check if anything actually happened */
	if (s == save)
		goto noconv;

	/* Store in endstr the address of one character
	past the last character we converted */
	if (endstr != NULL)
		*endstr = (char *)s;

	/* Check for a value that is within the range of
	`unsigned LONG int', but outside the range of `LONG int' */
	if (overflow == 0 && i > (negative ? -((unsigned long int) (LONG_MIN + 1)) + 1
	      : (unsigned long int) LONG_MAX))
		overflow = 1;
	if (overflow) {
		if (error != NULL)
			*error = 2;
		return negative ? LONG_MIN : LONG_MAX;
	}
	/* Return the result of the appropriate sign */
	return negative ? -i : i;

noconv:
	/* We must handle a special case here: the base is 0 or 16 and the
	first two characters are '0' and 'x', but the rest are no
	hexadecimal digits.  This is no error case.  We return 0 and
	endstr points to the `x` */
	if (endstr != NULL) {
		if (save - str >= 2 && (save[-1] & 0xdf) == 'X' && save[-2] == '0')
			*endstr = (char *) &save[-1];
		else
			/* There was no number to convert */
			*endstr = (char *) str;
	}
	return 0L;
}
