/*
*****************************************
* This file implements the StringMan interface
*
* Version: 0.1
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 11/02/2002
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2002, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* Some functions derived from Linux
*   NO Copyright (C) 1991, 1992 Linus Torvalds,
*   consider these trivial functions to be PD.
*****************************************
*/

#include <stdarg.h>
#include <stddef.h>

int StringMan_BufPrintFArgs(char*,char*,va_list);

int StringMan_Compare(const char *s1, const char *s2) {
    int d0, d1;
    register int __res;
    __asm__ __volatile__(
        "cld\n"
        "1:\tlodsb\n\t"
        "scasb\n\t"
        "jne 2f\n\t"
        "testb %%al,%%al\n\t"
        "jne 1b\n\t"
        "xorl %%eax,%%eax\n\t"
        "jmp 3f\n"
        "2:\tsbbl %%eax,%%eax\n\t"
        "orb $1,%%al\n"
        "3:"
        :"=a" (__res), "=&S" (d0), "=&D" (d1)
        :"1" (s1),"2" (s2));
    return __res;
}

void StringMan_Concatenate(char *dest, const char *src) {
    int d0, d1, d2, d3;
    __asm__ __volatile__(
        "cld\n\t"
        "repne\n\t"
        "scasb\n\t"
        "decl %1\n"
        "1:\tlodsb\n\t"
        "stosb\n\t"
        "testb %%al,%%al\n\t"
        "jne 1b"
        : "=&S" (d0), "=&D" (d1), "=&a" (d2), "=&c" (d3)
        : "0" (src), "1" (dest), "2" (0), "3" (0xffffffff):"memory");
}

void StringMan_Copy(char *dest, const char *src) {
    int d0, d1, d2;
    __asm__ __volatile__(
        "cld\n"
        "1:\tlodsb\n\t"
        "stosb\n\t"
        "testb %%al,%%al\n\t"
        "jne 1b"
        : "=&S" (d0), "=&D" (d1), "=&a" (d2)
        :"0" (src),"1" (dest) : "memory");
}

char *StringMan_FindChar(const char *s, int c) {
    int d0;
    register char * __res;
    __asm__ __volatile__(
        "cld\n\t"
        "movb %%al,%%ah\n"
        "1:\tlodsb\n\t"
        "cmpb %%ah,%%al\n\t"
        "je 2f\n\t"
        "testb %%al,%%al\n\t"
        "jne 1b\n\t"
        "movl $1,%1\n"
        "2:\tmovl %1,%0\n\t"
        "decl %0"
        :"=a" (__res), "=&S" (d0) : "1" (s),"0" (c));
    return __res;
}

char *StringMan_FindLastChar(const char *s, int c) {
    int d0, d1;
    register char * __res;
    __asm__ __volatile__(
        "cld\n\t"
        "movb %%al,%%ah\n"
        "1:\tlodsb\n\t"
        "cmpb %%ah,%%al\n\t"
        "jne 2f\n\t"
        "leal -1(%%esi),%0\n"
        "2:\ttestb %%al,%%al\n\t"
        "jne 1b"
        :"=g" (__res), "=&S" (d0), "=&a" (d1) :"0" (0),"1" (s),"2" (c));
    return __res;
}

unsigned int StringMan_Length(const char *s) {
    int d0;
    register int __res;
    __asm__ __volatile__(
        "cld\n\t"
        "repne\n\t"
        "scasb\n\t"
        "notl %0\n\t"
        "decl %0"
        :"=c" (__res), "=&D" (d0) :"1" (s),"a" (0), "0" (0xffffffff));
    return __res;
}

int StringMan_NCompare(const char *s1, const char *s2, unsigned int count) {
    register int __res;
    int d0, d1, d2;
    __asm__ __volatile__(
        "cld\n"
        "1:\tdecl %3\n\t"
        "js 2f\n\t"
        "lodsb\n\t"
        "scasb\n\t"
        "jne 3f\n\t"
        "testb %%al,%%al\n\t"
        "jne 1b\n"
        "2:\txorl %%eax,%%eax\n\t"
        "jmp 4f\n"
        "3:\tsbbl %%eax,%%eax\n\t"
        "orb $1,%%al\n"
        "4:"
    :"=a" (__res), "=&S" (d0), "=&D" (d1), "=&c" (d2)
    :"1" (s1),"2" (s2),"3" (count));
    return __res;
}

void StringMan_NConcatenate(char *dest, const char *src, unsigned int count) {
    int d0, d1, d2, d3;
    __asm__ __volatile__(
        "cld\n\t"
        "repne\n\t"
        "scasb\n\t"
        "decl %1\n\t"
        "movl %8,%3\n"
        "1:\tdecl %3\n\t"
        "js 2f\n\t"
        "lodsb\n\t"
        "stosb\n\t"
        "testb %%al,%%al\n\t"
        "jne 1b\n"
        "2:\txorl %2,%2\n\t"
        "stosb"
        : "=&S" (d0), "=&D" (d1), "=&a" (d2), "=&c" (d3)
        : "0" (src),"1" (dest),"2" (0),"3" (0xffffffff), "g" (count)
        : "memory");
}

void StringMan_NCopy(char *dest, const char *src, unsigned int count) {
    int d0, d1, d2, d3;
    __asm__ __volatile__(
        "cld\n"
        "1:\tdecl %2\n\t"
        "js 2f\n\t"
        "lodsb\n\t"
        "stosb\n\t"
        "testb %%al,%%al\n\t"
        "jne 1b\n\t"
        "rep\n\t"
        "stosb\n"
        "2:"
        : "=&S" (d0), "=&D" (d1), "=&c" (d2), "=&a" (d3)
        :"0" (src),"1" (dest),"2" (count) : "memory");
}

int StringMan_BufPrintF(char *buf, char *fmt, ...) {
    va_list args;
    int i;

    va_start(args, fmt);
    i=StringMan_BufPrintFArgs(buf,fmt,args);
    va_end(args);
    return i;
}

int StringMan_BufPrintFArgs(char *buf, char *fmt, va_list args) {
    unsigned int tmp, i;
    char *str, *s;

    for (str=buf; *fmt; fmt++) {
        if (*fmt != '%') {
            *str++ = *fmt;
            continue;
        }
        fmt++;
        switch (*fmt) {
            case 'c':
                *str++ = (unsigned char) va_arg(args, int);
                break;
            case 's':
                s = va_arg(args, char *);
                if (!s)
                    s = "<NULL>";
                tmp = StringMan_Length(s);
                for (i=0; i < tmp; i++)
                    *str++ = *s++;
                break;
            case 'x':
                *str++ = '0';
                *str++ = 'x';
                tmp = va_arg(args, unsigned int);
                i = 32;
                do {
                    unsigned char c;
                    i -= 8;
                    c = tmp >> i;
                    c = c >> 4;
                    if (c > 9)
                        *str++ = c+0x37;
                    else
                        *str++ = c+0x30;
                    c = (tmp >> i) & 0xf;
                    if (c > 9)
                        *str++ = c+0x37;
                    else
                        *str++ = c+0x30;
                } while (i);
                break;
            default:
                if (*fmt != '%')
                    *str++ = '%';
                if (*fmt)
                    *str++ = *fmt;
                else
                    fmt--;
        }
    }
    *str = '\0';
    return str-buf;
}

/*
*Code taked from glibc 2.1 - http://www.gnu.org
*Copyright (C) 1991,92,94,95,96,97,98,99 Free Software Foundation, Inc.
*/

#define ULONG_MAX	4294967295UL
#define LONG_MAX	2147483647L
#define LONG_MIN	(-LONG_MAX - 1L)

long int StringMan_StrToInt(const char *str, char **endstr, int base, int *error) {
    int negative, overflow;
    register unsigned long int cutoff, i;
    register unsigned int cutlim;
    register const char *s;
    register unsigned char c;
    const char *save;

    if (base < 0 || base == 1 || base > 36) {
        if (error != NULL)
            *error = 1;
        return 0;
    }
    save = s = str;

    /* Skip white space */
    while ((*s == ' ') || (*s == '\f') || (*s == '\n') || (*s == '\r') || (*s == '\t') || (*s == '\v'))
        ++s;
    if (*s == '\0')
        goto noconv;

    /* Check for a sign */
    if (*s == '-') {
        negative = 1;
        ++s;
    } else if (*s == '+') {
        negative = 0;
        ++s;
    } else
        negative = 0;

    /* Recognize number prefix and if BASE is zero, figure it out ourselves */
    if (*s == '0') {
        if ((base == 0 || base == 16) && (s[1] & 0xdf) == 'X') {
            s += 2;
            base = 16;
        } else if (base == 0)
            base = 8;
    } else if (base == 0)
        base = 10;

    /* Save the pointer so we can check later if anything happened */
    save = s;
    cutoff = ULONG_MAX / (unsigned long int) base;
    cutlim = ULONG_MAX % (unsigned long int) base;
    overflow = 0;
    i = 0;
    c = *s;
    for (;c != '\0'; c = *++s) {
        if (c >= '0' && c <= '9')
            c -= '0';
        else if ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z'))
            c = (c & 0xdf) - 'A' + 10;
        else
            break;
        if ((int) c >= base)
            break;
        /* Check for overflow */
        if (i > cutoff || (i == cutoff && c > cutlim))
            overflow = 1;
        else {
            i *= (unsigned long int) base;
            i += c;
        }
    }

    /* Check if anything actually happened */
    if (s == save)
        goto noconv;

    /* Store in endstr the address of one character
    past the last character we converted */
    if (endstr != NULL)
        *endstr = (char *)s;

    /* Check for a value that is within the range of
    `unsigned LONG int', but outside the range of `LONG int' */
    if (overflow == 0 && i > (negative ? -((unsigned long int) (LONG_MIN + 1)) + 1
          : (unsigned long int) LONG_MAX))
        overflow = 1;
    if (overflow) {
        if (error != NULL)
            *error = 2;
        return negative ? LONG_MIN : LONG_MAX;
    }
    /* Return the result of the appropriate sign */
    return negative ? -i : i;

noconv:
    /* We must handle a special case here: the base is 0 or 16 and the
    first two characters are '0' and 'x', but the rest are no
    hexadecimal digits.  This is no error case.  We return 0 and
    endstr points to the `x` */
    if (endstr != NULL) {
        if (save - str >= 2 && (save[-1] & 0xdf) == 'X' && save[-2] == '0')
            *endstr = (char *) &save[-1];
        else
            /* There was no number to convert */
            *endstr = (char *) str;
    }
    return 0L;
}
