/*
*****************************************
* This file implements the UMemManager interface
*
* Version: 0.1.5
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 22/06/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2005, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*
* This code is derived from the GNU C Library version 2.2.4.
* See GNU C Library malloc for code comments.
*/

/* Malloc implementation for multiple threads without lock contention.
   Copyright (C) 1996,1997,1998,1999,2000,2001 Free Software Foundation, Inc.
   This file is part of the GNU C Library.
   Contributed by Wolfram Gloger <wmglo@dent.med.uni-muenchen.de>
   and Doug Lea <dl@cs.oswego.edu>, 1996.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.  */

/* $Id: memmanager.c,v 1.7 2005/06/30 21:49:26 luizshigunov Exp $

  This work is mainly derived from malloc-2.6.4 by Doug Lea
  <dl@cs.oswego.edu>, which is available from:

                 ftp://g.oswego.edu/pub/misc/malloc.c
*/

#include <stddef.h>
#include <lib/memman.h>
#include <lib/memmanager.h>
#include <lib/mutex.h>
#include <lib/stringman.h>
#include <lib/thread.h>
#include <errormanager.h>
#include <intermanager.h>
#include <memmanager.h>
#include <usermodmanager.h>
#include <userspacemanager.h>
#include <textvideo.h>

struct mallinfo {
    unsigned int totalMem;
    unsigned int allocAreas;
    unsigned int totalAreas;
    unsigned int totalAlloc;
    unsigned int totalFree;
    unsigned int trimArea;
};

/* option stuff */
#define OPTION_TRIM_THRESHOLD 0x01
#define OPTION_TOP_PAD 0x02
#define OPTION_MMAP_THRESHOLD 0x03
#define OPTION_MMAP_MAX 0x04

/* function renaming */
#define cALLOc              AllocClear
#define fREe                Free
#define mALLOc              Alloc
#define mEMALIGn            AllocAlign
#define rEALLOc             ReAlloc
#define vALLOc              AllocPageAlign
#define pvALLOc             AllocPageRound
#define mALLINFo            GetMemStats
#define mALLOPt             SetMemOption
#define mALLOC_USABLE_SIZe  GetUsableSize
#define mALLOC_TRIm         MemTrim

/* mutex stuff */
#define mutex_t User_Mutex_Mutex
#define MUTEX_INITIALIZER {{0,0,0,0}}
#define mutex_trylock(x) User_Mutex_TryLock(x)
#define mutex_unlock(x) User_Mutex_Unlock(x)
#define mutex_lock(x) User_Mutex_Lock(x)
#define mutex_init(x) User_Mutex_Init(x, 0)

/* TSD stuff */
typedef unsigned int tsd_key_t;
#define tsd_key_create(key, destr) User_Thread_AllocTSD(key, destr)
#define tsd_setspecific(key, data) User_Thread_SetTSD(key, data)
#define tsd_getspecific(key, vptr) (vptr = User_Thread_GetTSD(key))

/* memory stuff */
#define memcpy(dst, src, size) User_MemMan_Copy(dst, src, size)
#define memmove(dst, src, size) User_MemMan_Move(dst, src, size)
#define memset(buf, c, size) User_MemMan_Set(buf, c, size)

/* sbrk stuff */
char *MemManager_brkBase;
static __inline__ void *sbrk(int diff) {
    char *newBase, *tmp;

    newBase = UserSpaceManager_UBrk(MemManager_brkBase + diff);
    if (newBase == MemManager_brkBase)
        return (void*)-1;
    tmp = MemManager_brkBase;
    MemManager_brkBase = newBase;
    return tmp;
}
#define MORECORE sbrk
#define MORECORE_FAILURE -1
#define MORECORE_CLEARS 2

#define munmap MemManager_UFreeUserPages

#define size_t      unsigned int
#define malloc_getpagesize 4096

#define Void_t      void

#if defined(DEBUG) || defined(DEBUG_ASSERT)
void __assert_fail(const char *__assertion, const char *__file, unsigned int __line, const char *__function) {
    while(1) {
#ifdef DEBUG_ASSERT_MSG
        char buf[16];
        int (*TV_Write)(const char*,int);
        char outBuf[256];
        InterManager_OutputInfo *outInfo;

        /* get TextVideo functions */
        outInfo = (InterManager_OutputInfo*)outBuf;
        if (!InterManager_UGetOutputInfo(outInfo, sizeof(outBuf)) && !User_StringMan_Compare(outInfo->names, "TextVideo")) {
            TV_Write = UserModManager_UGetFunction(outInfo->names, outInfo->names+outInfo->interSize, TextVideo__UWRITE, UserModManager_SYS_FUNCTION);
            if (!IS_PTR_ERROR(TV_Write)) {
                TV_Write(__file, (int)User_StringMan_Length(__file));
                User_StringMan_BufPrintF(buf, ":%x: ", __line);
                TV_Write(buf, (int)User_StringMan_Length(buf));
                if (__function) {
                    TV_Write(__function, (int)User_StringMan_Length(__function));
                    TV_Write(": ", sizeof(": ")-1);
                }
                TV_Write("Assertion '", sizeof("Assertion '")-1);
                TV_Write(__assertion, (int)User_StringMan_Length(__assertion));
                TV_Write("' failed.\n", sizeof("' failed.\n")-1);
            }
        }
#endif
        UserModManager_UExit(1);
    }
}
#endif

#ifdef DEBUG
#define MALLOC_DEBUG 2
#include <assert.h>
#else
#ifdef DEBUG_ASSERT
#define MALLOC_DEBUG 1
#include <assert.h>
#else
#define assert(x) ((void)0)
#endif
#endif

#define INTERNAL_SIZE_T unsigned int

#define REALLOC_ZERO_BYTES_FREES

#define HAVE_MEMCPY 1

#define MALLOC_ZERO(charp, nbytes)                                            \
do {                                                                          \
  INTERNAL_SIZE_T mzsz = (nbytes);                                            \
  if(mzsz <= 9*sizeof(mzsz)) {                                                \
    INTERNAL_SIZE_T* mz = (INTERNAL_SIZE_T*) (charp);                         \
    if(mzsz >= 5*sizeof(mzsz)) {     *mz++ = 0;                               \
                                     *mz++ = 0;                               \
      if(mzsz >= 7*sizeof(mzsz)) {   *mz++ = 0;                               \
                                     *mz++ = 0;                               \
        if(mzsz >= 9*sizeof(mzsz)) { *mz++ = 0;                               \
                                     *mz++ = 0; }}}                           \
                                     *mz++ = 0;                               \
                                     *mz++ = 0;                               \
                                     *mz   = 0;                               \
  } else memset((charp), 0, mzsz);                                            \
} while(0)

#define MALLOC_COPY(dest,src,nbytes,overlap)                                  \
do {                                                                          \
  INTERNAL_SIZE_T mcsz = (nbytes);                                            \
  if(mcsz <= 9*sizeof(mcsz)) {                                                \
    INTERNAL_SIZE_T* mcsrc = (INTERNAL_SIZE_T*) (src);                        \
    INTERNAL_SIZE_T* mcdst = (INTERNAL_SIZE_T*) (dest);                       \
    if(mcsz >= 5*sizeof(mcsz)) {     *mcdst++ = *mcsrc++;                     \
                                     *mcdst++ = *mcsrc++;                     \
      if(mcsz >= 7*sizeof(mcsz)) {   *mcdst++ = *mcsrc++;                     \
                                     *mcdst++ = *mcsrc++;                     \
        if(mcsz >= 9*sizeof(mcsz)) { *mcdst++ = *mcsrc++;                     \
                                     *mcdst++ = *mcsrc++; }}}                 \
                                     *mcdst++ = *mcsrc++;                     \
                                     *mcdst++ = *mcsrc++;                     \
                                     *mcdst   = *mcsrc  ;                     \
  } else if(overlap)                                                          \
    memmove(dest, src, mcsz);                                                 \
  else                                                                        \
    memcpy(dest, src, mcsz);                                                  \
} while(0)

#define HAVE_MMAP 1

#define USE_ARENAS 1

#define MAP_FAILED NULL

#define DEFAULT_TRIM_THRESHOLD (128 * 1024)

#define DEFAULT_TOP_PAD        (0)

#define DEFAULT_MMAP_THRESHOLD (128 * 1024)

#define DEFAULT_MMAP_MAX       (1024)

#define HEAP_MIN_SIZE (32*1024)
#define HEAP_MAX_SIZE (1024*1024) /* must be a power of two */

#define THREAD_STATS 0

struct malloc_chunk
{
  INTERNAL_SIZE_T prev_size; /* Size of previous chunk (if free). */
  INTERNAL_SIZE_T size;      /* Size in bytes, including overhead. */
  struct malloc_chunk* fd;   /* double links -- used only if free. */
  struct malloc_chunk* bk;
};

typedef struct malloc_chunk* mchunkptr;

#define NAV             128   /* number of bins */

typedef struct malloc_chunk* mbinptr;

typedef struct _arena {
  mbinptr av[2*NAV + 2];
  struct _arena *next;
  size_t size;
#if THREAD_STATS
  long stat_lock_direct, stat_lock_loop, stat_lock_wait;
#endif
  mutex_t mutex;
} arena;

typedef struct _heap_info {
  arena *ar_ptr; /* Arena for this heap. */
  struct _heap_info *prev; /* Previous heap. */
  size_t size;   /* Current size in bytes. */
  size_t pad;    /* Make sure the following data is properly aligned. */
} heap_info;

static void      chunk_free(arena *ar_ptr, mchunkptr p);
static mchunkptr chunk_alloc(arena *ar_ptr, INTERNAL_SIZE_T size);
static mchunkptr chunk_realloc(arena *ar_ptr, mchunkptr oldp, INTERNAL_SIZE_T oldsize, INTERNAL_SIZE_T nb);
static mchunkptr chunk_align(arena *ar_ptr, INTERNAL_SIZE_T nb, size_t alignment);
static int       main_trim(size_t pad);
#if USE_ARENAS
static int       heap_trim(heap_info *heap, size_t pad);
#endif

#define SIZE_SZ                (sizeof(INTERNAL_SIZE_T))
#define MALLOC_ALIGNMENT      (SIZE_SZ + SIZE_SZ)
#define MALLOC_ALIGN_MASK      (MALLOC_ALIGNMENT - 1)
#define MINSIZE                (sizeof(struct malloc_chunk))

#define chunk2mem(p) ((Void_t*)((char*)(p) + 2*SIZE_SZ))
#define mem2chunk(mem) chunk_at_offset((mem), -2*SIZE_SZ)

#define request2size(req, nb) \
 ((nb = (req) + (SIZE_SZ + MALLOC_ALIGN_MASK)),\
  ((long)nb <= 0 || nb < (INTERNAL_SIZE_T) (req) \
   ? (1) \
   : ((nb < (MINSIZE + MALLOC_ALIGN_MASK) \
	   ? (nb = MINSIZE) : (nb &= ~MALLOC_ALIGN_MASK)), 0)))

#define aligned_OK(m)    (((unsigned long)((m)) & (MALLOC_ALIGN_MASK)) == 0)

#define PREV_INUSE 0x1UL

#define IS_MMAPPED 0x2UL

#define SIZE_BITS (PREV_INUSE|IS_MMAPPED)

#define next_chunk(p) chunk_at_offset((p), (p)->size & ~PREV_INUSE)

#define prev_chunk(p) chunk_at_offset((p), -(p)->prev_size)

#define chunk_at_offset(p, s)  ((mchunkptr)(((char*)(p)) + (s)))

#define inuse(p) (next_chunk(p)->size & PREV_INUSE)

#define prev_inuse(p)  ((p)->size & PREV_INUSE)

#define chunk_is_mmapped(p) ((p)->size & IS_MMAPPED)

#define set_inuse(p) (next_chunk(p)->size |= PREV_INUSE)

#define clear_inuse(p) (next_chunk(p)->size &= ~PREV_INUSE)

#define inuse_bit_at_offset(p, s) \
  (chunk_at_offset((p), (s))->size & PREV_INUSE)

#define set_inuse_bit_at_offset(p, s) \
  (chunk_at_offset((p), (s))->size |= PREV_INUSE)

#define clear_inuse_bit_at_offset(p, s) \
  (chunk_at_offset((p), (s))->size &= ~(PREV_INUSE))

#define chunksize(p)          ((p)->size & ~(SIZE_BITS))

#define set_head_size(p, s)   ((p)->size = (((p)->size & PREV_INUSE) | (s)))

#define set_head(p, s)        ((p)->size = (s))

#define set_foot(p, s)   (chunk_at_offset(p, s)->prev_size = (s))

#define bin_at(a, i)   _bin_at(a, i)
#define _bin_at(a, i)  ((mbinptr)((char*)&(((a)->av)[2*(i)+2]) - 2*SIZE_SZ))
#define init_bin(a, i) ((a)->av[2*(i)+2] = (a)->av[2*(i)+3] = bin_at((a), (i)))
#define next_bin(b)    ((mbinptr)((char*)(b) + 2 * sizeof(((arena*)0)->av[0])))
#define prev_bin(b)    ((mbinptr)((char*)(b) - 2 * sizeof(((arena*)0)->av[0])))

#define binblocks(a)      (bin_at(a,0)->size)/* bitvector of nonempty blocks */
#define top(a)            (bin_at(a,0)->fd)  /* The topmost chunk */
#define last_remainder(a) (bin_at(a,1))      /* remainder from last split */

#define initial_top(a)    ((mchunkptr)bin_at(a, 0))

#define first(b) ((b)->fd)
#define last(b)  ((b)->bk)

#define bin_index(sz)                                                         \
(((((unsigned long)(sz)) >> 9) ==    0) ?       (((unsigned long)(sz)) >>  3):\
 ((((unsigned long)(sz)) >> 9) <=    4) ?  56 + (((unsigned long)(sz)) >>  6):\
 ((((unsigned long)(sz)) >> 9) <=   20) ?  91 + (((unsigned long)(sz)) >>  9):\
 ((((unsigned long)(sz)) >> 9) <=   84) ? 110 + (((unsigned long)(sz)) >> 12):\
 ((((unsigned long)(sz)) >> 9) <=  340) ? 119 + (((unsigned long)(sz)) >> 15):\
 ((((unsigned long)(sz)) >> 9) <= 1364) ? 124 + (((unsigned long)(sz)) >> 18):\
                                          126)
#define MAX_SMALLBIN         63
#define MAX_SMALLBIN_SIZE   512
#define SMALLBIN_WIDTH        8

#define smallbin_index(sz)  (((unsigned long)(sz)) >> 3)

#define is_small_request(nb) ((nb) < MAX_SMALLBIN_SIZE - SMALLBIN_WIDTH)

#define BINBLOCKWIDTH     4   /* bins per block */

#define idx2binblock(ix)      ((unsigned)1 << ((ix) / BINBLOCKWIDTH))
#define mark_binblock(a, ii)  (binblocks(a) |= idx2binblock(ii))
#define clear_binblock(a, ii) (binblocks(a) &= ~(idx2binblock(ii)))

#define IAV(i) _bin_at(&main_arena, i), _bin_at(&main_arena, i)

static arena main_arena = {
    {
 0, 0,
 IAV(0),   IAV(1),   IAV(2),   IAV(3),   IAV(4),   IAV(5),   IAV(6),   IAV(7),
 IAV(8),   IAV(9),   IAV(10),  IAV(11),  IAV(12),  IAV(13),  IAV(14),  IAV(15),
 IAV(16),  IAV(17),  IAV(18),  IAV(19),  IAV(20),  IAV(21),  IAV(22),  IAV(23),
 IAV(24),  IAV(25),  IAV(26),  IAV(27),  IAV(28),  IAV(29),  IAV(30),  IAV(31),
 IAV(32),  IAV(33),  IAV(34),  IAV(35),  IAV(36),  IAV(37),  IAV(38),  IAV(39),
 IAV(40),  IAV(41),  IAV(42),  IAV(43),  IAV(44),  IAV(45),  IAV(46),  IAV(47),
 IAV(48),  IAV(49),  IAV(50),  IAV(51),  IAV(52),  IAV(53),  IAV(54),  IAV(55),
 IAV(56),  IAV(57),  IAV(58),  IAV(59),  IAV(60),  IAV(61),  IAV(62),  IAV(63),
 IAV(64),  IAV(65),  IAV(66),  IAV(67),  IAV(68),  IAV(69),  IAV(70),  IAV(71),
 IAV(72),  IAV(73),  IAV(74),  IAV(75),  IAV(76),  IAV(77),  IAV(78),  IAV(79),
 IAV(80),  IAV(81),  IAV(82),  IAV(83),  IAV(84),  IAV(85),  IAV(86),  IAV(87),
 IAV(88),  IAV(89),  IAV(90),  IAV(91),  IAV(92),  IAV(93),  IAV(94),  IAV(95),
 IAV(96),  IAV(97),  IAV(98),  IAV(99),  IAV(100), IAV(101), IAV(102), IAV(103),
 IAV(104), IAV(105), IAV(106), IAV(107), IAV(108), IAV(109), IAV(110), IAV(111),
 IAV(112), IAV(113), IAV(114), IAV(115), IAV(116), IAV(117), IAV(118), IAV(119),
 IAV(120), IAV(121), IAV(122), IAV(123), IAV(124), IAV(125), IAV(126), IAV(127)
    },
    &main_arena, /* next */
    0, /* size */
#if THREAD_STATS
    0, 0, 0, /* stat_lock_direct, stat_lock_loop, stat_lock_wait */
#endif
    MUTEX_INITIALIZER /* mutex */
};

#undef IAV

static tsd_key_t arena_key;
static mutex_t list_lock;

#if THREAD_STATS
static int stat_n_heaps;
#define THREAD_STAT(x) x
#else
#define THREAD_STAT(x) do ; while(0)
#endif

static unsigned long trim_threshold   = DEFAULT_TRIM_THRESHOLD;
static unsigned long top_pad          = DEFAULT_TOP_PAD;
static unsigned int  n_mmaps_max      = DEFAULT_MMAP_MAX;
static unsigned long mmap_threshold   = DEFAULT_MMAP_THRESHOLD;

static char* sbrk_base = (char*)(-1);

static unsigned long max_sbrked_mem;

#define sbrked_mem (main_arena.size)

static unsigned int n_mmaps;
static unsigned int max_n_mmaps;
static unsigned long mmapped_mem;
static unsigned long max_mmapped_mem;

static unsigned long arena_mem;

#define weak_variable

int __malloc_initialized = -1;

int Start(void) {
  if(__malloc_initialized >= 0) return E_OK;
  __malloc_initialized = 0;
  mutex_init(&main_arena.mutex);
  mutex_init(&list_lock);
  tsd_key_create(&arena_key, NULL);
  tsd_setspecific(arena_key, (Void_t *)&main_arena);
  __malloc_initialized = 1;
  MemManager_brkBase = UserSpaceManager_UBrk(NULL);
  return E_OK;
}

#if HAVE_MMAP

#if defined __GNUC__ && __GNUC__ >= 2
/* This function is only called from one place, inline it.  */
__inline__
#endif
static mchunkptr
mmap_chunk(size_t size)
{
  size_t page_mask = malloc_getpagesize - 1;
  mchunkptr p;

  /* For mmapped chunks, the overhead is one SIZE_SZ unit larger, because
   * there is no following chunk whose prev_size field could be used.
   */
  size = (size + SIZE_SZ + page_mask) & ~page_mask;

  p = (mchunkptr)MemManager_UAllocUserPages((void*)0, size, 0, MemManager_PROT_READ+MemManager_PROT_WRITE);
  if(IS_PTR_ERROR(p)) return 0;

  n_mmaps++;
  if (n_mmaps > max_n_mmaps) max_n_mmaps = n_mmaps;

  /* We demand that eight bytes into a page must be 8-byte aligned. */
  assert(aligned_OK(chunk2mem(p)));

  /* The offset to the start of the mmapped region is stored
   * in the prev_size field of the chunk; normally it is zero,
   * but that can be changed in memalign().
   */
  p->prev_size = 0;
  set_head(p, size|IS_MMAPPED);

  mmapped_mem += size;
  if ((unsigned long)mmapped_mem > (unsigned long)max_mmapped_mem)
    max_mmapped_mem = mmapped_mem;
  return p;
}

static void
munmap_chunk(mchunkptr p)
{
  INTERNAL_SIZE_T size = chunksize(p);
  int ret;

  assert (chunk_is_mmapped(p));
  assert(! ((char*)p >= sbrk_base && (char*)p < sbrk_base + sbrked_mem));
  assert((n_mmaps > 0));
  assert(((p->prev_size + size) & (malloc_getpagesize-1)) == 0);

  n_mmaps--;
  mmapped_mem -= (size + p->prev_size);

  ret = munmap((char *)p - p->prev_size, size + p->prev_size);

  /* munmap returns non-zero on failure */
  assert(ret == 0);
}

#endif /* HAVE_MMAP */

#if USE_ARENAS

static heap_info *
new_heap(size_t size)
{
  size_t page_mask = malloc_getpagesize - 1;
  char *p2;
  heap_info *h;

  if(size+top_pad < HEAP_MIN_SIZE)
    size = HEAP_MIN_SIZE;
  else if(size+top_pad <= HEAP_MAX_SIZE)
    size += top_pad;
  else if(size > HEAP_MAX_SIZE)
    return 0;
  else
    size = HEAP_MAX_SIZE;
  size = (size + page_mask) & ~page_mask;

  p2 = (char *)MemManager_UAllocUserPages((void*)0, HEAP_MAX_SIZE, HEAP_MAX_SIZE, MemManager_PROT_NONE);
  if(IS_PTR_ERROR(p2))
    return 0;
  if (MemManager_USetUserPagesProt(p2, size, MemManager_PROT_READ+MemManager_PROT_WRITE)) {
    munmap(p2, HEAP_MAX_SIZE);
    return 0;
  }
  h = (heap_info *)p2;
  h->size = size;
  THREAD_STAT(stat_n_heaps++);
  return h;
}

static int
grow_heap(heap_info *h, long diff)
{
  size_t page_mask = malloc_getpagesize - 1;
  long new_size;

  if(diff >= 0) {
    diff = (diff + page_mask) & ~page_mask;
    new_size = (long)h->size + diff;
    if(new_size > HEAP_MAX_SIZE)
      return -1;
    if(IS_PTR_ERROR(MemManager_UAllocUserPages((char *)h + h->size, diff, 0, MemManager_PROT_READ+MemManager_PROT_WRITE)))
      return -2;
  } else {
    new_size = (long)h->size + diff;
    if(new_size < (long)sizeof(*h))
      return -1;
    /* Try to re-map the extra heap space freshly to save memory, and
       make it inaccessible. */
    if(IS_PTR_ERROR(MemManager_UAllocUserPages((char *)h + new_size, -diff, 0, MemManager_PROT_NONE)))
      return -2;
  }
  h->size = new_size;
  return 0;
}

#define delete_heap(heap) munmap((char*)(heap), HEAP_MAX_SIZE)

#define arena_get(ptr, size) do { \
  Void_t *vptr = NULL; \
  ptr = (arena *)tsd_getspecific(arena_key, vptr); \
  if(ptr && !mutex_trylock(&ptr->mutex)) { \
    THREAD_STAT(++(ptr->stat_lock_direct)); \
  } else \
    ptr = arena_get2(ptr, (size)); \
} while(0)

static arena *
arena_get2(arena *a_tsd, size_t size)
{
  arena *a;
  heap_info *h;
  char *ptr;
  int i;
  unsigned long misalign;

  if(!a_tsd)
    a = a_tsd = &main_arena;
  else {
    a = a_tsd->next;
    if(!a) {
      /* This can only happen while initializing the new arena. */
      (void)mutex_lock(&main_arena.mutex);
      THREAD_STAT(++(main_arena.stat_lock_wait));
      return &main_arena;
    }
  }

  /* Check the global, circularly linked list for available arenas. */
  do {
    if(!mutex_trylock(&a->mutex)) {
      THREAD_STAT(++(a->stat_lock_loop));
      tsd_setspecific(arena_key, (Void_t *)a);
      return a;
    }
    a = a->next;
  } while(a != a_tsd);

  /* Nothing immediately available, so generate a new arena. */
  h = new_heap(size + (sizeof(*h) + sizeof(*a) + MALLOC_ALIGNMENT));
  if(!h) {
    /* Maybe size is too large to fit in a single heap.  So, just try
       to create a minimally-sized arena and let chunk_alloc() attempt
       to deal with the large request via mmap_chunk(). */
    h = new_heap(sizeof(*h) + sizeof(*a) + MALLOC_ALIGNMENT);
    if(!h)
      return 0;
  }
  a = h->ar_ptr = (arena *)(h+1);
  for(i=0; i<NAV; i++)
    init_bin(a, i);
  a->next = NULL;
  a->size = h->size;
  arena_mem += h->size;
  tsd_setspecific(arena_key, (Void_t *)a);
  mutex_init(&a->mutex);
  i = mutex_lock(&a->mutex); /* remember result */

  /* Set up the top chunk, with proper alignment. */
  ptr = (char *)(a + 1);
  misalign = (unsigned long)chunk2mem(ptr) & MALLOC_ALIGN_MASK;
  if (misalign > 0)
    ptr += MALLOC_ALIGNMENT - misalign;
  top(a) = (mchunkptr)ptr;
  set_head(top(a), (((char*)h + h->size) - ptr) | PREV_INUSE);

  /* Add the new arena to the list. */
  (void)mutex_lock(&list_lock);
  a->next = main_arena.next;
  main_arena.next = a;
  (void)mutex_unlock(&list_lock);

  if(i) /* locking failed; keep arena for further attempts later */
    return 0;

  THREAD_STAT(++(a->stat_lock_loop));
  return a;
}

#define heap_for_ptr(ptr) \
 ((heap_info *)((unsigned long)(ptr) & ~(HEAP_MAX_SIZE-1)))
#define arena_for_ptr(ptr) \
 (((mchunkptr)(ptr) < top(&main_arena) && (char *)(ptr) >= sbrk_base) ? \
  &main_arena : heap_for_ptr(ptr)->ar_ptr)

#else /* !USE_ARENAS */

/* There is only one arena, main_arena. */

#define arena_get(ptr, sz) (ptr = &main_arena)
#define arena_for_ptr(ptr) (&main_arena)

#endif /* USE_ARENAS */

#if MALLOC_DEBUG

static void do_check_chunk(arena *ar_ptr, mchunkptr p)
{
  INTERNAL_SIZE_T sz = p->size & ~PREV_INUSE;

  /* No checkable chunk is mmapped */
  assert(!chunk_is_mmapped(p));

#if USE_ARENAS
  if(ar_ptr != &main_arena) {
    heap_info *heap = heap_for_ptr(p);
    assert(heap->ar_ptr == ar_ptr);
    if(p != top(ar_ptr))
      assert((char *)p + sz <= (char *)heap + heap->size);
    else
      assert((char *)p + sz == (char *)heap + heap->size);
    return;
  }
#endif

  /* Check for legal address ... */
  assert((char*)p >= sbrk_base);
  if (p != top(ar_ptr))
    assert((char*)p + sz <= (char*)top(ar_ptr));
  else
    assert((char*)p + sz <= sbrk_base + sbrked_mem);

}

static void do_check_free_chunk(arena *ar_ptr, mchunkptr p)
{
  INTERNAL_SIZE_T sz = p->size & ~PREV_INUSE;
  mchunkptr next = chunk_at_offset(p, sz);

  do_check_chunk(ar_ptr, p);

  /* Check whether it claims to be free ... */
  assert(!inuse(p));

  /* Must have OK size and fields */
  assert((long)sz >= (long)MINSIZE);
  assert((sz & MALLOC_ALIGN_MASK) == 0);
  assert(aligned_OK(chunk2mem(p)));
  /* ... matching footer field */
  assert(next->prev_size == sz);
  /* ... and is fully consolidated */
  assert(prev_inuse(p));
  assert (next == top(ar_ptr) || inuse(next));

  /* ... and has minimally sane links */
  assert(p->fd->bk == p);
  assert(p->bk->fd == p);
}

static void do_check_inuse_chunk(arena *ar_ptr, mchunkptr p)
{
  mchunkptr next = next_chunk(p);
  do_check_chunk(ar_ptr, p);

  /* Check whether it claims to be in use ... */
  assert(inuse(p));

  /* ... whether its size is OK (it might be a fencepost) ... */
  assert(chunksize(p) >= MINSIZE || next->size == (0|PREV_INUSE));

  /* ... and is surrounded by OK chunks.
    Since more things can be checked with free chunks than inuse ones,
    if an inuse chunk borders them and debug is on, it's worth doing them.
  */
  if (!prev_inuse(p))
  {
    mchunkptr prv = prev_chunk(p);
    assert(next_chunk(prv) == p);
    do_check_free_chunk(ar_ptr, prv);
  }
  if (next == top(ar_ptr))
  {
    assert(prev_inuse(next));
    assert(chunksize(next) >= MINSIZE);
  }
  else if (!inuse(next))
    do_check_free_chunk(ar_ptr, next);

}

static void do_check_malloced_chunk(arena *ar_ptr,
                                    mchunkptr p, INTERNAL_SIZE_T s)
{
  INTERNAL_SIZE_T sz = p->size & ~PREV_INUSE;
  long room = sz - s;

  do_check_inuse_chunk(ar_ptr, p);

  /* Legal size ... */
  assert((long)sz >= (long)MINSIZE);
  assert((sz & MALLOC_ALIGN_MASK) == 0);
  assert(room >= 0);
  assert(room < (long)MINSIZE);

  /* ... and alignment */
  assert(aligned_OK(chunk2mem(p)));


  /* ... and was allocated at front of an available chunk */
  assert(prev_inuse(p));

}

#define check_free_chunk(A,P) do_check_free_chunk(A,P)
#define check_inuse_chunk(A,P) do_check_inuse_chunk(A,P)
#define check_chunk(A,P) do_check_chunk(A,P)
#define check_malloced_chunk(A,P,N) do_check_malloced_chunk(A,P,N)
#else
#define check_free_chunk(A,P)
#define check_inuse_chunk(A,P)
#define check_chunk(A,P)
#define check_malloced_chunk(A,P,N)
#endif

#define frontlink(A, P, S, IDX, BK, FD)                                       \
{                                                                             \
  if (S < MAX_SMALLBIN_SIZE)                                                  \
  {                                                                           \
    IDX = smallbin_index(S);                                                  \
    mark_binblock(A, IDX);                                                    \
    BK = bin_at(A, IDX);                                                      \
    FD = BK->fd;                                                              \
    P->bk = BK;                                                               \
    P->fd = FD;                                                               \
    FD->bk = BK->fd = P;                                                      \
  }                                                                           \
  else                                                                        \
  {                                                                           \
    IDX = bin_index(S);                                                       \
    BK = bin_at(A, IDX);                                                      \
    FD = BK->fd;                                                              \
    if (FD == BK) mark_binblock(A, IDX);                                      \
    else                                                                      \
    {                                                                         \
      while (FD != BK && S < chunksize(FD)) FD = FD->fd;                      \
      BK = FD->bk;                                                            \
    }                                                                         \
    P->bk = BK;                                                               \
    P->fd = FD;                                                               \
    FD->bk = BK->fd = P;                                                      \
  }                                                                           \
}

#define unlink(P, BK, FD)                                                     \
{                                                                             \
  BK = P->bk;                                                                 \
  FD = P->fd;                                                                 \
  FD->bk = BK;                                                                \
  BK->fd = FD;                                                                \
}                                                                             \

#define link_last_remainder(A, P)                                             \
{                                                                             \
  last_remainder(A)->fd = last_remainder(A)->bk = P;                          \
  P->fd = P->bk = last_remainder(A);                                          \
}

#define clear_last_remainder(A) \
  (last_remainder(A)->fd = last_remainder(A)->bk = last_remainder(A))

#if defined __GNUC__ && __GNUC__ >= 2
/* This function is called only from one place, inline it.  */
__inline__
#endif
static void
malloc_extend_top(arena *ar_ptr, INTERNAL_SIZE_T nb)
{
  unsigned long pagesz   = malloc_getpagesize;
  mchunkptr old_top      = top(ar_ptr);        /* Record state of old top */
  INTERNAL_SIZE_T old_top_size = chunksize(old_top);
  INTERNAL_SIZE_T top_size;                    /* new size of top chunk */

#if USE_ARENAS
  if(ar_ptr == &main_arena) {
#endif

    char*     brk;                  /* return value from sbrk */
    INTERNAL_SIZE_T front_misalign; /* unusable bytes at front of sbrked space */
    INTERNAL_SIZE_T correction;     /* bytes for 2nd sbrk call */
    char*     new_brk;              /* return of 2nd sbrk call */
    char*     old_end = (char*)(chunk_at_offset(old_top, old_top_size));

    /* Pad request with top_pad plus minimal overhead */
    INTERNAL_SIZE_T sbrk_size = nb + top_pad + MINSIZE;

    /* If not the first time through, round to preserve page boundary */
    /* Otherwise, we need to correct to a page size below anyway. */
    /* (We also correct below if an intervening foreign sbrk call.) */

    if (sbrk_base != (char*)(-1))
      sbrk_size = (sbrk_size + (pagesz - 1)) & ~(pagesz - 1);

    brk = (char*)(MORECORE (sbrk_size));

    /* Fail if sbrk failed or if a foreign sbrk call killed our space */
    if (brk == (char*)(MORECORE_FAILURE) ||
        (brk < old_end && old_top != initial_top(&main_arena)))
      return;

    sbrked_mem += sbrk_size;

    if (brk == old_end) { /* can just add bytes to current top */
      top_size = sbrk_size + old_top_size;
      set_head(old_top, top_size | PREV_INUSE);
      old_top = 0; /* don't free below */
    } else {
      if (sbrk_base == (char*)(-1)) /* First time through. Record base */
        sbrk_base = brk;
      else
        /* Someone else called sbrk().  Count those bytes as sbrked_mem. */
        sbrked_mem += brk - (char*)old_end;

      /* Guarantee alignment of first new chunk made from this space */
      front_misalign = (unsigned long)chunk2mem(brk) & MALLOC_ALIGN_MASK;
      if (front_misalign > 0) {
        correction = (MALLOC_ALIGNMENT) - front_misalign;
        brk += correction;
      } else
        correction = 0;

      /* Guarantee the next brk will be at a page boundary */
      correction += pagesz - ((unsigned long)(brk + sbrk_size) & (pagesz - 1));

      /* Allocate correction */
      new_brk = (char*)(MORECORE (correction));
      if (new_brk == (char*)(MORECORE_FAILURE)) return;

      sbrked_mem += correction;

      top(&main_arena) = chunk_at_offset(brk, 0);
      top_size = new_brk - brk + correction;
      set_head(top(&main_arena), top_size | PREV_INUSE);

      if (old_top == initial_top(&main_arena))
        old_top = 0; /* don't free below */
    }

    if ((unsigned long)sbrked_mem > (unsigned long)max_sbrked_mem)
      max_sbrked_mem = sbrked_mem;

#if USE_ARENAS
  } else { /* ar_ptr != &main_arena */
    heap_info *old_heap, *heap;
    size_t old_heap_size;

    if(old_top_size < MINSIZE) /* this should never happen */
      return;

    /* First try to extend the current heap. */
    if(MINSIZE + nb <= old_top_size)
      return;
    old_heap = heap_for_ptr(old_top);
    old_heap_size = old_heap->size;
    if(grow_heap(old_heap, MINSIZE + nb - old_top_size) == 0) {
      ar_ptr->size += old_heap->size - old_heap_size;
      arena_mem += old_heap->size - old_heap_size;
      top_size = ((char *)old_heap + old_heap->size) - (char *)old_top;
      set_head(old_top, top_size | PREV_INUSE);
      return;
    }

    /* A new heap must be created. */
    heap = new_heap(nb + (MINSIZE + sizeof(*heap)));
    if(!heap)
      return;
    heap->ar_ptr = ar_ptr;
    heap->prev = old_heap;
    ar_ptr->size += heap->size;
    arena_mem += heap->size;

    /* Set up the new top, so we can safely use chunk_free() below. */
    top(ar_ptr) = chunk_at_offset(heap, sizeof(*heap));
    top_size = heap->size - sizeof(*heap);
    set_head(top(ar_ptr), top_size | PREV_INUSE);
  }
#endif /* USE_ARENAS */

  /* We always land on a page boundary */
  assert(((unsigned long)((char*)top(ar_ptr) + top_size) & (pagesz-1)) == 0);

  /* Setup fencepost and free the old top chunk. */
  if(old_top) {
    /* The fencepost takes at least MINSIZE bytes, because it might
       become the top chunk again later.  Note that a footer is set
       up, too, although the chunk is marked in use. */
    old_top_size -= MINSIZE;
    set_head(chunk_at_offset(old_top, old_top_size + 2*SIZE_SZ), 0|PREV_INUSE);
    if(old_top_size >= MINSIZE) {
      set_head(chunk_at_offset(old_top, old_top_size), (2*SIZE_SZ)|PREV_INUSE);
      set_foot(chunk_at_offset(old_top, old_top_size), (2*SIZE_SZ));
      set_head_size(old_top, old_top_size);
      chunk_free(ar_ptr, old_top);
    } else {
      set_head(old_top, (old_top_size + 2*SIZE_SZ)|PREV_INUSE);
      set_foot(old_top, (old_top_size + 2*SIZE_SZ));
    }
  }
}

Void_t* mALLOc(size_t bytes)
{
  arena *ar_ptr;
  INTERNAL_SIZE_T nb; /* padded request size */
  mchunkptr victim;

  if(request2size(bytes, nb))
    return 0;
  arena_get(ar_ptr, nb);
  if(!ar_ptr)
    return 0;
  victim = chunk_alloc(ar_ptr, nb);
  if(!victim) {
    /* Maybe the failure is due to running out of mmapped areas. */
    if(ar_ptr != &main_arena) {
      (void)mutex_unlock(&ar_ptr->mutex);
      (void)mutex_lock(&main_arena.mutex);
      victim = chunk_alloc(&main_arena, nb);
      (void)mutex_unlock(&main_arena.mutex);
    } else {
#if USE_ARENAS
      /* ... or sbrk() has failed and there is still a chance to mmap() */
      ar_ptr = arena_get2(ar_ptr->next ? ar_ptr : 0, nb);
      (void)mutex_unlock(&main_arena.mutex);
      if(ar_ptr) {
        victim = chunk_alloc(ar_ptr, nb);
        (void)mutex_unlock(&ar_ptr->mutex);
      }
#endif
    }
    if(!victim) return 0;
  } else
    (void)mutex_unlock(&ar_ptr->mutex);
  return chunk2mem(victim);
}

static mchunkptr
chunk_alloc(arena *ar_ptr, INTERNAL_SIZE_T nb)
{
  mchunkptr victim;                  /* inspected/selected chunk */
  INTERNAL_SIZE_T victim_size;       /* its size */
  int       idx;                     /* index for bin traversal */
  mbinptr   bin;                     /* associated bin */
  mchunkptr remainder;               /* remainder from a split */
  long      remainder_size;          /* its size */
  int       remainder_index;         /* its bin index */
  unsigned long block;               /* block traverser bit */
  int       startidx;                /* first bin of a traversed block */
  mchunkptr fwd;                     /* misc temp for linking */
  mchunkptr bck;                     /* misc temp for linking */
  mbinptr q;                         /* misc temp */


  /* Check for exact match in a bin */

  if (is_small_request(nb))  /* Faster version for small requests */
  {
    idx = smallbin_index(nb);

    /* No traversal or size check necessary for small bins.  */

    q = _bin_at(ar_ptr, idx);
    victim = last(q);

    /* Also scan the next one, since it would have a remainder < MINSIZE */
    if (victim == q)
    {
      q = next_bin(q);
      victim = last(q);
    }
    if (victim != q)
    {
      victim_size = chunksize(victim);
      unlink(victim, bck, fwd);
      set_inuse_bit_at_offset(victim, victim_size);
      check_malloced_chunk(ar_ptr, victim, nb);
      return victim;
    }

    idx += 2; /* Set for bin scan below. We've already scanned 2 bins. */

  }
  else
  {
    idx = bin_index(nb);
    bin = bin_at(ar_ptr, idx);

    for (victim = last(bin); victim != bin; victim = victim->bk)
    {
      victim_size = chunksize(victim);
      remainder_size = victim_size - nb;

      if (remainder_size >= (long)MINSIZE) /* too big */
      {
        --idx; /* adjust to rescan below after checking last remainder */
        break;
      }

      else if (remainder_size >= 0) /* exact fit */
      {
        unlink(victim, bck, fwd);
        set_inuse_bit_at_offset(victim, victim_size);
        check_malloced_chunk(ar_ptr, victim, nb);
        return victim;
      }
    }

    ++idx;

  }

  /* Try to use the last split-off remainder */

  if ( (victim = last_remainder(ar_ptr)->fd) != last_remainder(ar_ptr))
  {
    victim_size = chunksize(victim);
    remainder_size = victim_size - nb;

    if (remainder_size >= (long)MINSIZE) /* re-split */
    {
      remainder = chunk_at_offset(victim, nb);
      set_head(victim, nb | PREV_INUSE);
      link_last_remainder(ar_ptr, remainder);
      set_head(remainder, remainder_size | PREV_INUSE);
      set_foot(remainder, remainder_size);
      check_malloced_chunk(ar_ptr, victim, nb);
      return victim;
    }

    clear_last_remainder(ar_ptr);

    if (remainder_size >= 0)  /* exhaust */
    {
      set_inuse_bit_at_offset(victim, victim_size);
      check_malloced_chunk(ar_ptr, victim, nb);
      return victim;
    }

    /* Else place in bin */

    frontlink(ar_ptr, victim, victim_size, remainder_index, bck, fwd);
  }

  /*
     If there are any possibly nonempty big-enough blocks,
     search for best fitting chunk by scanning bins in blockwidth units.
  */

  if ( (block = idx2binblock(idx)) <= binblocks(ar_ptr))
  {

    /* Get to the first marked block */

    if ( (block & binblocks(ar_ptr)) == 0)
    {
      /* force to an even block boundary */
      idx = (idx & ~(BINBLOCKWIDTH - 1)) + BINBLOCKWIDTH;
      block <<= 1;
      while ((block & binblocks(ar_ptr)) == 0)
      {
        idx += BINBLOCKWIDTH;
        block <<= 1;
      }
    }

    /* For each possibly nonempty block ... */
    for (;;)
    {
      startidx = idx;          /* (track incomplete blocks) */
      q = bin = _bin_at(ar_ptr, idx);

      /* For each bin in this block ... */
      do
      {
        /* Find and use first big enough chunk ... */

        for (victim = last(bin); victim != bin; victim = victim->bk)
        {
          victim_size = chunksize(victim);
          remainder_size = victim_size - nb;

          if (remainder_size >= (long)MINSIZE) /* split */
          {
            remainder = chunk_at_offset(victim, nb);
            set_head(victim, nb | PREV_INUSE);
            unlink(victim, bck, fwd);
            link_last_remainder(ar_ptr, remainder);
            set_head(remainder, remainder_size | PREV_INUSE);
            set_foot(remainder, remainder_size);
            check_malloced_chunk(ar_ptr, victim, nb);
            return victim;
          }

          else if (remainder_size >= 0)  /* take */
          {
            set_inuse_bit_at_offset(victim, victim_size);
            unlink(victim, bck, fwd);
            check_malloced_chunk(ar_ptr, victim, nb);
            return victim;
          }

        }

       bin = next_bin(bin);

      } while ((++idx & (BINBLOCKWIDTH - 1)) != 0);

      /* Clear out the block bit. */

      do   /* Possibly backtrack to try to clear a partial block */
      {
        if ((startidx & (BINBLOCKWIDTH - 1)) == 0)
        {
          binblocks(ar_ptr) &= ~block;
          break;
        }
        --startidx;
        q = prev_bin(q);
      } while (first(q) == q);

      /* Get to the next possibly nonempty block */

      if ( (block <<= 1) <= binblocks(ar_ptr) && (block != 0) )
      {
        while ((block & binblocks(ar_ptr)) == 0)
        {
          idx += BINBLOCKWIDTH;
          block <<= 1;
        }
      }
      else
        break;
    }
  }


  /* Try to use top chunk */

  /* Require that there be a remainder, ensuring top always exists  */
  if ( (remainder_size = chunksize(top(ar_ptr)) - nb) < (long)MINSIZE)
  {

#if HAVE_MMAP
    /* If the request is big and there are not yet too many regions,
       and we would otherwise need to extend, try to use mmap instead.  */
    if ((unsigned long)nb >= (unsigned long)mmap_threshold &&
        n_mmaps < n_mmaps_max &&
        (victim = mmap_chunk(nb)) != 0)
      return victim;
#endif

    /* Try to extend */
    malloc_extend_top(ar_ptr, nb);
    if ((remainder_size = chunksize(top(ar_ptr)) - nb) < (long)MINSIZE)
    {
#if HAVE_MMAP
      /* A last attempt: when we are out of address space in a
         non-main arena, try mmap anyway, as long as it is allowed at
         all.  */
      if (ar_ptr != &main_arena &&
          n_mmaps_max > 0 &&
          (victim = mmap_chunk(nb)) != 0)
        return victim;
#endif
      return 0; /* propagate failure */
    }
  }

  victim = top(ar_ptr);
  set_head(victim, nb | PREV_INUSE);
  top(ar_ptr) = chunk_at_offset(victim, nb);
  set_head(top(ar_ptr), remainder_size | PREV_INUSE);
  check_malloced_chunk(ar_ptr, victim, nb);
  return victim;

}

void fREe(Void_t* mem)
{
  arena *ar_ptr;
  mchunkptr p;                          /* chunk corresponding to mem */

  if (mem == 0)                              /* free(0) has no effect */
    return;

  p = mem2chunk(mem);

#if HAVE_MMAP
  if (chunk_is_mmapped(p))                       /* release mmapped memory. */
  {
    munmap_chunk(p);
    return;
  }
#endif

  ar_ptr = arena_for_ptr(p);
#if THREAD_STATS
  if(!mutex_trylock(&ar_ptr->mutex))
    ++(ar_ptr->stat_lock_direct);
  else {
    (void)mutex_lock(&ar_ptr->mutex);
    ++(ar_ptr->stat_lock_wait);
  }
#else
  (void)mutex_lock(&ar_ptr->mutex);
#endif
  chunk_free(ar_ptr, p);
  (void)mutex_unlock(&ar_ptr->mutex);
}

static void
chunk_free(arena *ar_ptr, mchunkptr p)
{
  INTERNAL_SIZE_T hd = p->size; /* its head field */
  INTERNAL_SIZE_T sz;  /* its size */
  int       idx;       /* its bin index */
  mchunkptr next;      /* next contiguous chunk */
  INTERNAL_SIZE_T nextsz; /* its size */
  INTERNAL_SIZE_T prevsz; /* size of previous contiguous chunk */
  mchunkptr bck;       /* misc temp for linking */
  mchunkptr fwd;       /* misc temp for linking */
  int       islr;      /* track whether merging with last_remainder */

  check_inuse_chunk(ar_ptr, p);

  sz = hd & ~PREV_INUSE;
  next = chunk_at_offset(p, sz);
  nextsz = chunksize(next);

  if (next == top(ar_ptr))                         /* merge with top */
  {
    sz += nextsz;

    if (!(hd & PREV_INUSE))                    /* consolidate backward */
    {
      prevsz = p->prev_size;
      p = chunk_at_offset(p, -(long)prevsz);
      sz += prevsz;
      unlink(p, bck, fwd);
    }

    set_head(p, sz | PREV_INUSE);
    top(ar_ptr) = p;

#if USE_ARENAS
    if(ar_ptr == &main_arena) {
#endif
      if ((unsigned long)(sz) >= (unsigned long)trim_threshold)
        main_trim(top_pad);
#if USE_ARENAS
    } else {
      heap_info *heap = heap_for_ptr(p);

      assert(heap->ar_ptr == ar_ptr);

      /* Try to get rid of completely empty heaps, if possible. */
      if((unsigned long)(sz) >= (unsigned long)trim_threshold ||
         p == chunk_at_offset(heap, sizeof(*heap)))
        heap_trim(heap, top_pad);
    }
#endif
    return;
  }

  islr = 0;

  if (!(hd & PREV_INUSE))                    /* consolidate backward */
  {
    prevsz = p->prev_size;
    p = chunk_at_offset(p, -(long)prevsz);
    sz += prevsz;

    if (p->fd == last_remainder(ar_ptr))     /* keep as last_remainder */
      islr = 1;
    else
      unlink(p, bck, fwd);
  }

  if (!(inuse_bit_at_offset(next, nextsz)))   /* consolidate forward */
  {
    sz += nextsz;

    if (!islr && next->fd == last_remainder(ar_ptr))
                                              /* re-insert last_remainder */
    {
      islr = 1;
      link_last_remainder(ar_ptr, p);
    }
    else
      unlink(next, bck, fwd);

    next = chunk_at_offset(p, sz);
  }
  else
    set_head(next, nextsz);                  /* clear inuse bit */

  set_head(p, sz | PREV_INUSE);
  next->prev_size = sz;
  if (!islr)
    frontlink(ar_ptr, p, sz, idx, bck, fwd);

#if USE_ARENAS
  /* Check whether the heap containing top can go away now. */
  if(next->size < MINSIZE &&
     (unsigned long)sz > trim_threshold &&
     ar_ptr != &main_arena) {                /* fencepost */
    heap_info *heap = heap_for_ptr(top(ar_ptr));

    if(top(ar_ptr) == chunk_at_offset(heap, sizeof(*heap)) &&
       heap->prev == heap_for_ptr(p))
      heap_trim(heap, top_pad);
  }
#endif
}

Void_t* rEALLOc(Void_t* oldmem, size_t bytes)
{
  arena *ar_ptr;
  INTERNAL_SIZE_T    nb;      /* padded request size */

  mchunkptr oldp;             /* chunk corresponding to oldmem */
  INTERNAL_SIZE_T    oldsize; /* its size */

  mchunkptr newp;             /* chunk to return */

#ifdef REALLOC_ZERO_BYTES_FREES
  if (bytes == 0 && oldmem != NULL) { fREe(oldmem); return 0; }
#endif

  /* realloc of null is supposed to be same as malloc */
  if (oldmem == 0) return mALLOc(bytes);

  oldp    = mem2chunk(oldmem);
  oldsize = chunksize(oldp);

  if(request2size(bytes, nb))
    return 0;

#if HAVE_MMAP
  if (chunk_is_mmapped(oldp))
  {
    Void_t* newmem;

    /* Note the extra SIZE_SZ overhead. */
    if(oldsize - SIZE_SZ >= nb) return oldmem; /* do nothing */
    /* Must alloc, copy, free. */
    newmem = mALLOc(bytes);
    if (newmem == 0) return 0; /* propagate failure */
    MALLOC_COPY(newmem, oldmem, oldsize - 2*SIZE_SZ, 0);
    munmap_chunk(oldp);
    return newmem;
  }
#endif

  ar_ptr = arena_for_ptr(oldp);
#if THREAD_STATS
  if(!mutex_trylock(&ar_ptr->mutex))
    ++(ar_ptr->stat_lock_direct);
  else {
    (void)mutex_lock(&ar_ptr->mutex);
    ++(ar_ptr->stat_lock_wait);
  }
#else
  (void)mutex_lock(&ar_ptr->mutex);
#endif

  /* As in malloc(), remember this arena for the next allocation. */
  tsd_setspecific(arena_key, (Void_t *)ar_ptr);

  newp = chunk_realloc(ar_ptr, oldp, oldsize, nb);

  (void)mutex_unlock(&ar_ptr->mutex);
  return newp ? chunk2mem(newp) : NULL;
}

static mchunkptr
chunk_realloc(arena* ar_ptr, mchunkptr oldp, INTERNAL_SIZE_T oldsize,
              INTERNAL_SIZE_T nb)
{
  mchunkptr newp = oldp;      /* chunk to return */
  INTERNAL_SIZE_T newsize = oldsize; /* its size */

  mchunkptr next;             /* next contiguous chunk after oldp */
  INTERNAL_SIZE_T  nextsize;  /* its size */

  mchunkptr prev;             /* previous contiguous chunk before oldp */
  INTERNAL_SIZE_T  prevsize;  /* its size */

  mchunkptr remainder;        /* holds split off extra space from newp */
  INTERNAL_SIZE_T  remainder_size;   /* its size */

  mchunkptr bck;              /* misc temp for linking */
  mchunkptr fwd;              /* misc temp for linking */

  check_inuse_chunk(ar_ptr, oldp);

  if ((long)(oldsize) < (long)(nb))
  {
    Void_t* oldmem = chunk2mem(oldp);

    /* Try expanding forward */

    next = chunk_at_offset(oldp, oldsize);
    if (next == top(ar_ptr) || !inuse(next))
    {
      nextsize = chunksize(next);

      /* Forward into top only if a remainder */
      if (next == top(ar_ptr))
      {
        if ((long)(nextsize + newsize) >= (long)(nb + MINSIZE))
        {
          newsize += nextsize;
          top(ar_ptr) = chunk_at_offset(oldp, nb);
          set_head(top(ar_ptr), (newsize - nb) | PREV_INUSE);
          set_head_size(oldp, nb);
          return oldp;
        }
      }

      /* Forward into next chunk */
      else if (((long)(nextsize + newsize) >= (long)(nb)))
      {
        unlink(next, bck, fwd);
        newsize  += nextsize;
        goto split;
      }
    }
    else
    {
      next = 0;
      nextsize = 0;
    }

    oldsize -= SIZE_SZ;

    /* Try shifting backwards. */

    if (!prev_inuse(oldp))
    {
      prev = prev_chunk(oldp);
      prevsize = chunksize(prev);

      /* try forward + backward first to save a later consolidation */

      if (next != 0)
      {
        /* into top */
        if (next == top(ar_ptr))
        {
          if ((long)(nextsize + prevsize + newsize) >= (long)(nb + MINSIZE))
          {
            unlink(prev, bck, fwd);
            newp = prev;
            newsize += prevsize + nextsize;
            MALLOC_COPY(chunk2mem(newp), oldmem, oldsize,
                        1);
            top(ar_ptr) = chunk_at_offset(newp, nb);
            set_head(top(ar_ptr), (newsize - nb) | PREV_INUSE);
            set_head_size(newp, nb);
            return newp;
          }
        }

        /* into next chunk */
        else if (((long)(nextsize + prevsize + newsize) >= (long)(nb)))
        {
          unlink(next, bck, fwd);
          unlink(prev, bck, fwd);
          newp = prev;
          newsize += nextsize + prevsize;
          MALLOC_COPY(chunk2mem(newp), oldmem, oldsize, 1);
          goto split;
        }
      }

      /* backward only */
      if (prev != 0 && (long)(prevsize + newsize) >= (long)nb)
      {
        unlink(prev, bck, fwd);
        newp = prev;
        newsize += prevsize;
        MALLOC_COPY(chunk2mem(newp), oldmem, oldsize, 1);
        goto split;
      }
    }

    /* Must allocate */

    newp = chunk_alloc (ar_ptr, nb);

    if (newp == 0) {
      /* Maybe the failure is due to running out of mmapped areas. */
      if (ar_ptr != &main_arena) {
        (void)mutex_lock(&main_arena.mutex);
        newp = chunk_alloc(&main_arena, nb);
        (void)mutex_unlock(&main_arena.mutex);
      } else {
#if USE_ARENAS
        /* ... or sbrk() has failed and there is still a chance to mmap() */
        arena* ar_ptr2 = arena_get2(ar_ptr->next ? ar_ptr : 0, nb);
        if(ar_ptr2) {
          newp = chunk_alloc(ar_ptr2, nb);
          (void)mutex_unlock(&ar_ptr2->mutex);
        }
#endif
      }
      if (newp == 0) /* propagate failure */
        return 0;
    }

    /* Avoid copy if newp is next chunk after oldp. */
    /* (This can only happen when new chunk is sbrk'ed.) */

    if ( newp == next_chunk(oldp))
    {
      newsize += chunksize(newp);
      newp = oldp;
      goto split;
    }

    /* Otherwise copy, free, and exit */
    MALLOC_COPY(chunk2mem(newp), oldmem, oldsize, 0);
    chunk_free(ar_ptr, oldp);
    return newp;
  }


 split:  /* split off extra room in old or expanded chunk */

  if (newsize - nb >= MINSIZE) /* split off remainder */
  {
    remainder = chunk_at_offset(newp, nb);
    remainder_size = newsize - nb;
    set_head_size(newp, nb);
    set_head(remainder, remainder_size | PREV_INUSE);
    set_inuse_bit_at_offset(remainder, remainder_size);
    chunk_free(ar_ptr, remainder);
  }
  else
  {
    set_head_size(newp, newsize);
    set_inuse_bit_at_offset(newp, newsize);
  }

  check_inuse_chunk(ar_ptr, newp);
  return newp;
}

Void_t* mEMALIGn(size_t alignment, size_t bytes)
{
  arena *ar_ptr;
  INTERNAL_SIZE_T    nb;      /* padded  request size */
  mchunkptr p;

  /* If need less alignment than we give anyway, just relay to malloc */

  if (alignment <= MALLOC_ALIGNMENT) return mALLOc(bytes);

  /* Otherwise, ensure that it is at least a minimum chunk size */

  if (alignment <  MINSIZE) alignment = MINSIZE;

  if(request2size(bytes, nb))
    return 0;
  arena_get(ar_ptr, nb + alignment + MINSIZE);
  if(!ar_ptr)
    return 0;
  p = chunk_align(ar_ptr, nb, alignment);
  (void)mutex_unlock(&ar_ptr->mutex);
  if(!p) {
    /* Maybe the failure is due to running out of mmapped areas. */
    if(ar_ptr != &main_arena) {
      (void)mutex_lock(&main_arena.mutex);
      p = chunk_align(&main_arena, nb, alignment);
      (void)mutex_unlock(&main_arena.mutex);
    } else {
#if USE_ARENAS
      /* ... or sbrk() has failed and there is still a chance to mmap() */
      ar_ptr = arena_get2(ar_ptr->next ? ar_ptr : 0, nb);
      if(ar_ptr) {
        p = chunk_align(ar_ptr, nb, alignment);
        (void)mutex_unlock(&ar_ptr->mutex);
      }
#endif
    }
    if(!p) return 0;
  }
  return chunk2mem(p);
}

static mchunkptr
chunk_align(arena* ar_ptr, INTERNAL_SIZE_T nb, size_t alignment)
{
  unsigned long m;            /* memory returned by malloc call */
  mchunkptr p;                /* corresponding chunk */
  char*     brk;              /* alignment point within p */
  mchunkptr newp;             /* chunk to return */
  INTERNAL_SIZE_T  newsize;   /* its size */
  INTERNAL_SIZE_T  leadsize;  /* leading space befor alignment point */
  mchunkptr remainder;        /* spare room at end to split off */
  long      remainder_size;   /* its size */

  /* Call chunk_alloc with worst case padding to hit alignment. */
  p = chunk_alloc(ar_ptr, nb + alignment + MINSIZE);
  if (p == 0)
    return 0; /* propagate failure */

  m = (unsigned long)chunk2mem(p);

  if ((m % alignment) == 0) /* aligned */
  {
#if HAVE_MMAP
    if(chunk_is_mmapped(p)) {
      return p; /* nothing more to do */
    }
#endif
  }
  else /* misaligned */
  {
    /*
      Find an aligned spot inside chunk.
      Since we need to give back leading space in a chunk of at
      least MINSIZE, if the first calculation places us at
      a spot with less than MINSIZE leader, we can move to the
      next aligned spot -- we've allocated enough total room so that
      this is always possible.
    */

    brk = (char*)mem2chunk(((m + alignment - 1)) & -(long)alignment);
    if ((long)(brk - (char*)(p)) < (long)MINSIZE) brk += alignment;

    newp = chunk_at_offset(brk, 0);
    leadsize = brk - (char*)(p);
    newsize = chunksize(p) - leadsize;

#if HAVE_MMAP
    if(chunk_is_mmapped(p))
    {
      newp->prev_size = p->prev_size + leadsize;
      set_head(newp, newsize|IS_MMAPPED);
      return newp;
    }
#endif

    /* give back leader, use the rest */

    set_head(newp, newsize | PREV_INUSE);
    set_inuse_bit_at_offset(newp, newsize);
    set_head_size(p, leadsize);
    chunk_free(ar_ptr, p);
    p = newp;

    assert (newsize>=nb && (((unsigned long)(chunk2mem(p))) % alignment) == 0);
  }

  /* Also give back spare room at the end */

  remainder_size = chunksize(p) - nb;

  if (remainder_size >= (long)MINSIZE)
  {
    remainder = chunk_at_offset(p, nb);
    set_head(remainder, remainder_size | PREV_INUSE);
    set_head_size(p, nb);
    chunk_free(ar_ptr, remainder);
  }

  check_inuse_chunk(ar_ptr, p);
  return p;
}

Void_t* vALLOc(size_t bytes)
{
  return mEMALIGn (malloc_getpagesize, bytes);
}

Void_t* pvALLOc(size_t bytes)
{
  size_t pagesize;
  pagesize = malloc_getpagesize;
  return mEMALIGn (pagesize, (bytes + pagesize - 1) & ~(pagesize - 1));
}

Void_t* cALLOc(size_t n, size_t elem_size)
{
  arena *ar_ptr;
  mchunkptr p, oldtop;
  INTERNAL_SIZE_T sz, csz, oldtopsize;
  Void_t* mem;

  if(request2size(n * elem_size, sz))
    return 0;
  arena_get(ar_ptr, sz);
  if(!ar_ptr)
    return 0;

  /* Check if expand_top called, in which case there may be
     no need to clear. */
#if MORECORE_CLEARS
  oldtop = top(ar_ptr);
  oldtopsize = chunksize(top(ar_ptr));
#if MORECORE_CLEARS < 2
  /* Only newly allocated memory is guaranteed to be cleared.  */
  if (ar_ptr == &main_arena &&
      oldtopsize < sbrk_base + max_sbrked_mem - (char *)oldtop)
    oldtopsize = (sbrk_base + max_sbrked_mem - (char *)oldtop);
#endif
#endif
  p = chunk_alloc (ar_ptr, sz);

  /* Only clearing follows, so we can unlock early. */
  (void)mutex_unlock(&ar_ptr->mutex);

  if (p == 0) {
    /* Maybe the failure is due to running out of mmapped areas. */
    if(ar_ptr != &main_arena) {
      (void)mutex_lock(&main_arena.mutex);
      p = chunk_alloc(&main_arena, sz);
      (void)mutex_unlock(&main_arena.mutex);
    } else {
#if USE_ARENAS
      /* ... or sbrk() has failed and there is still a chance to mmap() */
      (void)mutex_lock(&main_arena.mutex);
      ar_ptr = arena_get2(ar_ptr->next ? ar_ptr : 0, sz);
      (void)mutex_unlock(&main_arena.mutex);
      if(ar_ptr) {
        p = chunk_alloc(ar_ptr, sz);
        (void)mutex_unlock(&ar_ptr->mutex);
      }
#endif
    }
    if (p == 0) return 0;
  }
  mem = chunk2mem(p);

  /* Two optional cases in which clearing not necessary */

#if HAVE_MMAP
  if (chunk_is_mmapped(p)) return mem;
#endif

  csz = chunksize(p);

#if MORECORE_CLEARS
  if (p == oldtop && csz > oldtopsize) {
    /* clear only the bytes from non-freshly-sbrked memory */
    csz = oldtopsize;
  }
#endif

  csz -= SIZE_SZ;
  MALLOC_ZERO(chunk2mem(p), csz);
  return mem;
}

int mALLOC_TRIm(size_t pad)
{
  int res;

  (void)mutex_lock(&main_arena.mutex);
  res = main_trim(pad);
  (void)mutex_unlock(&main_arena.mutex);
  return res;
}

static int
main_trim(size_t pad)
{
  mchunkptr top_chunk;   /* The current top chunk */
  long  top_size;        /* Amount of top-most memory */
  long  extra;           /* Amount to release */
  char* current_brk;     /* address returned by pre-check sbrk call */
  char* new_brk;         /* address returned by negative sbrk call */

  unsigned long pagesz = malloc_getpagesize;

  top_chunk = top(&main_arena);
  top_size = chunksize(top_chunk);
  extra = ((top_size - pad - MINSIZE + (pagesz-1)) / pagesz - 1) * pagesz;

  if (extra < (long)pagesz) /* Not enough memory to release */
    return 0;

  /* Test to make sure no one else called sbrk */
  current_brk = (char*)(MORECORE (0));
  if (current_brk != (char*)(top_chunk) + top_size)
    return 0;     /* Apparently we don't own memory; must fail */

  new_brk = (char*)(MORECORE (-extra));

  if (new_brk == (char*)(MORECORE_FAILURE)) { /* sbrk failed? */
    /* Try to figure out what we have */
    current_brk = (char*)(MORECORE (0));
    top_size = current_brk - (char*)top_chunk;
    if (top_size >= (long)MINSIZE) /* if not, we are very very dead! */
    {
      sbrked_mem = current_brk - sbrk_base;
      set_head(top_chunk, top_size | PREV_INUSE);
    }
    check_chunk(&main_arena, top_chunk);
    return 0;
  }
  sbrked_mem -= extra;

  /* Success. Adjust top accordingly. */
  set_head(top_chunk, (top_size - extra) | PREV_INUSE);
  check_chunk(&main_arena, top_chunk);
  return 1;
}

#if USE_ARENAS

static int
heap_trim(heap_info *heap, size_t pad)
{
  unsigned long pagesz = malloc_getpagesize;
  arena *ar_ptr = heap->ar_ptr;
  mchunkptr top_chunk = top(ar_ptr), p, bck, fwd;
  heap_info *prev_heap;
  long new_size, top_size, extra;

  /* Can this heap go away completely ? */
  while(top_chunk == chunk_at_offset(heap, sizeof(*heap))) {
    prev_heap = heap->prev;
    p = chunk_at_offset(prev_heap, prev_heap->size - (MINSIZE-2*SIZE_SZ));
    assert(p->size == (0|PREV_INUSE)); /* must be fencepost */
    p = prev_chunk(p);
    new_size = chunksize(p) + (MINSIZE-2*SIZE_SZ);
    assert(new_size>0 && new_size<(long)(2*MINSIZE));
    if(!prev_inuse(p))
      new_size += p->prev_size;
    assert(new_size>0 && new_size<HEAP_MAX_SIZE);
    if(new_size + (HEAP_MAX_SIZE - prev_heap->size) < pad + MINSIZE + pagesz)
      break;
    ar_ptr->size -= heap->size;
    arena_mem -= heap->size;
    delete_heap(heap);
    heap = prev_heap;
    if(!prev_inuse(p)) { /* consolidate backward */
      p = prev_chunk(p);
      unlink(p, bck, fwd);
    }
    assert(((unsigned long)((char*)p + new_size) & (pagesz-1)) == 0);
    assert( ((char*)p + new_size) == ((char*)heap + heap->size) );
    top(ar_ptr) = top_chunk = p;
    set_head(top_chunk, new_size | PREV_INUSE);
    check_chunk(ar_ptr, top_chunk);
  }
  top_size = chunksize(top_chunk);
  extra = ((top_size - pad - MINSIZE + (pagesz-1))/pagesz - 1) * pagesz;
  if(extra < (long)pagesz)
    return 0;
  /* Try to shrink. */
  if(grow_heap(heap, -extra) != 0)
    return 0;
  ar_ptr->size -= extra;
  arena_mem -= extra;

  /* Success. Adjust top accordingly. */
  set_head(top_chunk, (top_size - extra) | PREV_INUSE);
  check_chunk(ar_ptr, top_chunk);
  return 1;
}

#endif /* USE_ARENAS */

size_t mALLOC_USABLE_SIZe(Void_t* mem)
{
  mchunkptr p;

  if (mem == 0)
    return 0;
  else
  {
    p = mem2chunk(mem);
    if(!chunk_is_mmapped(p))
    {
      if (!inuse(p)) return 0;
      check_inuse_chunk(arena_for_ptr(mem), p);
      return chunksize(p) - SIZE_SZ;
    }
    return chunksize(p) - 2*SIZE_SZ;
  }
}

static void
malloc_update_mallinfo(arena *ar_ptr, struct mallinfo *mi)
{
  int i;
  mbinptr b;
  mchunkptr p;
#if MALLOC_DEBUG
  mchunkptr q;
#endif
  INTERNAL_SIZE_T avail;

  (void)mutex_lock(&ar_ptr->mutex);
  avail = chunksize(top(ar_ptr));

  for (i = 1; i < NAV; ++i)
  {
    b = bin_at(ar_ptr, i);
    for (p = last(b); p != b; p = p->bk)
    {
#if MALLOC_DEBUG
      check_free_chunk(ar_ptr, p);
      for (q = next_chunk(p);
           q != top(ar_ptr) && inuse(q) && (long)chunksize(q) > 0;
           q = next_chunk(q))
        check_inuse_chunk(ar_ptr, q);
#endif
      avail += chunksize(p);
    }
  }

  mi->totalMem = ar_ptr->size;
  mi->totalAlloc = ar_ptr->size - avail;
  mi->totalFree = avail;
  mi->allocAreas = n_mmaps;
  mi->totalAreas = mmapped_mem;
  mi->trimArea = chunksize(top(ar_ptr));

  (void)mutex_unlock(&ar_ptr->mutex);
}

struct mallinfo mALLINFo()
{
  struct mallinfo mi;
  Void_t *vptr = NULL;

  tsd_getspecific(arena_key, vptr);
  malloc_update_mallinfo((vptr ? (arena*)vptr : &main_arena), &mi);
  return mi;
}

int mALLOPt(int param_number, int value)
{
  switch(param_number)
  {
    case OPTION_TRIM_THRESHOLD:
      trim_threshold = value; return 1;
    case OPTION_TOP_PAD:
      top_pad = value; return 1;
    case OPTION_MMAP_THRESHOLD:
#if USE_ARENAS
      /* Forbid setting the threshold too high. */
      if((unsigned long)value > HEAP_MAX_SIZE/2) return 0;
#endif
      mmap_threshold = value; return 1;
    case OPTION_MMAP_MAX:
#if HAVE_MMAP
      n_mmaps_max = value; return 1;
#else
      if (value != 0) return 0; else  n_mmaps_max = value; return 1;
#endif

    default:
      return 0;
  }
}
