/*
*****************************************
* This file implements the Mutex interface
*
* Version: 0.1
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 13/02/2002
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2002, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

/* THIS IMPLEMENTATION IS FOR A 386 OR BETTER. WITH A 486 OR BETTER WE HAVE OTHER OPTIONS.
   WITH A 486 OR BETTER WE DON'T NEED TO DO A SYSCALL IF THE MUTEX IS UNLOCK. */

#include <stddef.h>
#include "errormanager.h"
#include "lib/mutex.h"
#include "lib/thread.h"
#include "taskmanager.h"

typedef struct {
    TaskManager_Semaphore s;
    User_Thread_Thread *owner;
    unsigned int ownerCount;
} Mutex;

User_Thread_Thread *Thread_GetID(void);

int Fast_Destroy(Mutex *m) {
    return E_OK;
}

int Fast_Lock(Mutex *m) {
    return TaskManager_UP((TaskManager_Semaphore*)m);
}

int Fast_TryLock(Mutex *m) {
    return TaskManager_UTryP((TaskManager_Semaphore*)m);
}

int Fast_Unlock(Mutex *m) {
    return TaskManager_UV((TaskManager_Semaphore*)m);
}

int Rec_Init(Mutex *m, int prop) {
    m->s.count = 1;
    m->s.data = NULL;
    m->owner = NULL;
    m->ownerCount = 0;
    return E_OK;
}

int Rec_Lock(Mutex *m) {
    User_Thread_Thread *id;

    id = Thread_GetID();
    if (m->owner == id) {
        m->ownerCount++;
    } else {
        Fast_Lock(m);
        m->owner = id;
    }
    return E_OK;
}

int Rec_TryLock(Mutex *m) {
    int ret;
    User_Thread_Thread *id;

    ret = E_OK;
    id = Thread_GetID();
    if (m->owner == id) {
        m->ownerCount++;
    } else {
        ret = Fast_TryLock(m);
        if (!ret)
            m->owner = id;
    }
    return ret;
}

int Rec_Unlock(Mutex *m) {
    if (m->ownerCount)
        m->ownerCount--;
    else {
        m->owner = NULL;
        Fast_Unlock(m);
    }
    return E_OK;
}

int Error_Destroy(Mutex *m) {
    if (m->s.count)
        return E_BUSY;
    return E_OK;
}

int Error_Lock(Mutex *m) {
    User_Thread_Thread *id;

    id = Thread_GetID();
    if (m->owner == id)
        return E_BUSY;
    Fast_Lock(m);
    m->owner = id;
    return E_OK;
}

int Error_TryLock(Mutex *m) {
    int ret;

    ret = Fast_TryLock(m);
    if (!ret)
        m->owner = Thread_GetID();
    return ret;
}

int Error_Unlock(Mutex *m) {
    if (m->owner != Thread_GetID() || !m->s.count)
        return E_PERMISSION_DENIED;
    m->owner = NULL;
    Fast_Unlock(m);
    return E_OK;
}
